/***************************************************************\
*	cqt3.c			  				*
*   Make a constant-Q frequency transform of a file		*
*   cqt with fewer analysis frames (same sample dens per oct)	*
*   dpwe 26feb91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#include <dpwelib.h>
#ifdef __STDC__
#include <stdlib.h>
#include <string.h>
#else
#include <stdio.h>
#include <malloc.h>
#include <strings.h>
#endif
#include <math.h>
#ifndef PI
#define PI M_PI		/* for that alone */
#endif /* PI */
/* personal libs */
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include <genutils.h>	/* for Mymalloc() */
/* #include <gendsp.h> */
#include <resambuf.h>
#include <readaqt.h>
#include "cqlib.h"
#include "ds_fir_93.h"	/* special filter coeffs */

/* constants */
#define	DFLT_SLOP	1	/* boost HF by 3db/oct? */
#define DFLT_DC		1	/* cancel DC from cos kernels? */
#define DEF_EXT 	".aqt"	/* default output file extension */
#define READBLOKSIZE	4096	/* samples read off disk at a time */
#define LOBES		32	/* lobes for resampler */
#define GUARDBAND	.5	/* how close we try to get to Nyq frq */
#define MYINFOSIZE	32	/* PVinfo used for "Q=.." and "octs=.." */

/* static function prototype */
static void PerformCQTransform(char *inPath, char *outPath);
static void usage(void);
static void WriteFloatsAsIrc(float *flts, long len, char *name);

/* static data */
       char	*programName;
static int	verb = 0;		/* -v flag */
static int	lf_cut = DFLT_SLOP;
static int	dc_out = DFLT_DC;
static float	qFactor= DFLT_QFCT;
static float	guardband = GUARDBAND;
static int	nBins  = DFLT_NBIN;
static int	nOcts  = DFLT_8VES;
static int	graphs = 0;		/* use debugging graphs */

/* functions */

main(argc,argv)
    int 	argc;
    char	**argv;
    {
    int 	i = 0;
    int 	err = 0;
    char	*inPath = NULL, *outPath = NULL;

    programName = argv[0];
    /**** %% DPWE STD ARG HANDLER %% ****/
    i = 0;
    while(++i<argc && err == 0)
	{
	char c, *token, *arg, *nextArg;
	int  argUsed;

	token = argv[i];
	if(*token++ == '-')
	    {
	    if(i+1 < argc)	nextArg = argv[i+1];
	    else		nextArg = "";
	    argUsed = 0;
	    while(c = *token++)
		{
		if(NumericQ(token))	arg = token;
		else			arg = nextArg;
		switch(c)
		    {
	        case 'q': case 'Q':	qFactor = atof(arg);	argUsed = 1;
		    if(qFactor<1)
			{  fprintf(stderr,"%s: Q of %s = %.2f must be > 1\n",
				programName, arg, qFactor);
			   err = 1;  	}
		    break;
	        case 'g':	guardband = atof(arg);	argUsed = 1;
		    if(guardband < 0.0 || guardband > 1.0)
			{  fprintf(stderr,"%s: gb of %s = %.2f must be 0..1\n",
				programName, arg, guardband);
			   err = 1;  	}
		    break;
		case 'b':		nBins = atol(arg);	argUsed = 1;
		    if(nBins < 2 || nBins > 300)
			{  fprintf(stderr,"%s: Bins of %s = %d not 2-300\n",
				programName, arg, nBins);
			   err = 1;  	}
		    break;
		case 'o':		nOcts = atol(arg);	argUsed = 1;
		    if(nOcts < 1 || nOcts > MAXOCTS)
			{  fprintf(stderr,"%s: Octs of %s = %d must be 1-%d\n",
				programName, arg, nOcts, MAXOCTS);
			   err = 1;  	}
		    break;
		case 's':	lf_cut = !DFLT_SLOP;	break;
		case 'd':	dc_out = !DFLT_DC;	break;
		case 'v': 	verb = 1; break;
		default:	fprintf(stderr,"%s: unrec option %c\n",
					programName, c);
		                err = 1; break;
		    }
		if(argUsed)
		    {
		    if(arg == token)	token = "";   /* no more from token */
		    else		++i;	      /* skip arg we used */
		    arg = ""; argUsed = 0;
		    }
		}
	    }
	else{
	    if(inPath == NULL)		inPath = argv[i];
	    else if(outPath == NULL)	outPath = argv[i];
	    else{ fprintf(stderr,"%s: excess arg %s\n", programName, argv[i]);
		  err = 1;  }
	    }
	}
    /* other tests on args */
/*    if(data < 0)
 *	{ fprintf(stderr,"%s: bad data %d: must be >= 0\n", programName, data);
 *	  err = 1; }
 */
    if(err || inPath == NULL)
	usage();		/* never returns */

    if(outPath == NULL)
	{
	outPath=
	    (char *)Mymalloc(sizeof(char), (1+strlen(inPath)+strlen(DEF_EXT)), 
			   "cqt3:outpath");
	strcpy(outPath, inPath);
	SubstExtn(inPath, DEF_EXT, outPath);
	}

/*    fprintf(stderr,"%s: data = %d\n",programName,data);
 *    fprintf(stderr,"%s: verbose = %s\n",programName,verb?"yes":"no");
 *    fprintf(stderr,"%s: infile %s, outfile %s\n",programName,inPath,outPath);
 */
    /* args all dealt with.  On to the real business */
    PerformCQTransform(inPath, outPath);
    }

static void PerformCQTransform(char *inPath, char *outPath)
    {
    SFSTRUCT	*inSnd;
    PVSTRUCT    *outPv;
    CQTINFO 	*cqi;
    FILE	*inFile, *outFile;
    long	frameSize, hopSize, fbSize;
    float	*kernels;
    long	kernLen;
    int 	chans;
    int 	i, err, done=0, ndone=1;
    long	rdBlkSize, red;
    float	*rdBuf, *wrBuf;
    RsmpDescType   *rsDPtr0, *rsDPtr1;
    TrigRSBufType  *rsBufPtrs[MAXOCTS], *lastrsB;
    long	frame, trigSize;
    int 	lev;
    long	mask;
    long	rsbsize, lag, fill;
#ifdef DEBUGGRAPHS
    WINDAT	*wdptr[MAXOCTS];
#endif
    char	desc[DESCLEN];

/* code for a quick test of ds_f filter */
/*    SFRead(inPath, &inSnd);
/*    ConvertSoundFormat(&inSnd, SFMT_FLOAT);
/*    Convo((float *)SFDataLoc(inSnd), 1, inSnd->dataBsize/sizeof(float),
/*	  ds_fir, 1, DS_FIR_LEN);
/*    ConvertSoundFormat(&inSnd, SFMT_SHORT);
/*    SFWrite(outPath, inSnd);
/*    exit(0);
 */

    /* open input sound file */
    if( (inFile = SFOpenAllocRdHdr(inPath, &inSnd)) == NULL)
	SFDie(SFE_NOPEN, inPath);

    /* create & open output PV file */
    kernLen = (long)(qFactor*16);
    hopSize = qFactor/4;		/* advance after each kernel appl */
    cqi = NewCQTInfo("a cqi");
    cqi->nBins = nBins;	cqi->nOcts = nOcts; cqi->qFactor = qFactor;
    cqi->kernLen = kernLen; cqi->hopSize = hopSize; cqi->guardBand = guardband;
    chans = inSnd->channels;
    if(chans != 1)
	fprintf(stderr, "%s: using just 1st channel of %d channel sound\n",
		programName, chans);
    if(err = PVAlloc(&outPv, PV_UNK_LEN, PVFLOAT, inSnd->samplingRate,
		     chans, 0L,0L,0L,0L,0.0,0.0,0L, MYINFOSIZE))
	PVDie(err,NULL);
    EncodeCQTPV(cqi, outPv);
    if((outFile = PVOpenWrHdr(outPath, outPv))==NULL)
	PVDie(PVE_NOPEN, outPath);

    /* allocate buffers */
    rdBlkSize = READBLOKSIZE;
    rdBuf = (float *)Mymalloc(sizeof(float), rdBlkSize*chans, "cqt3:rdBuf");
    wrBuf = (float *)Mymalloc(sizeof(cpxpolar), nBins, "cqt3:wrBuf");
                            /* one 8ve of polar */
    kernels = (float *)Mymalloc(sizeof(float), kernLen*2*nBins, "cqt3:kern");
    NewSetupKernels((long)nBins, (FLOATARG)qFactor, 
		    (FLOATARG)cqi->guardBand*PI/2, (int)kernLen, kernels,
		    1,(FLOATARG)1.0);
#ifdef DEBUGGRAPHS
    if(graphs)
	{
	wdptr[0] = NewWin("cqt3",0);
	DoDisp(wdptr[0], kernels, kernLen+1);
	}
#endif
/*    rsDPtr1 = NewResampKernel(2, 1, LOBES, (float)1);	/* 2:1 downsample */
    rsDPtr1 = MakeResampKernel(2, 1, DS_FIR_LEN, ds_fir);
    /* setup a resampling kernel using our special filter */
/*    WriteFloatsAsIrc(ds_fir, DS_FIR_LEN, "ds_fir.irc");	*/
/*    WriteFloatsAsIrc(rsDPtr1->kernCtr - rsDPtr1->kernhalflen,
		     (rsDPtr1->kernhalflen * 2)+1, "kern.irc");	*/

    /* setup chained resampler objects */
    lastrsB = NULL;
    trigSize = kernLen;
    lag = trigSize/1 + 2*LOBES;	/* should be trigSize/2 - but that's slightly
			   too short.  Never did find right answer */
    fill = trigSize/2;
    for(i = nOcts-1; i>=0; --i)
	{
	sprintf(desc,"%d",i);
#ifdef DEBUGGRAPHS
	if(graphs)
	    wdptr[i] = NewWin(desc,0);
#endif
	rsDPtr0 = (i==0)?NULL:rsDPtr1;
	rsbsize = MinResampleBufLen(rsDPtr0,
				    trigSize/2+lag,
	/* by the argument that worst case, you might wait for a lowest
	   octave whose centre is aligned to yours, an effective delay
	   in your terms of half a trig len scaled up n octaves */
				    rdBlkSize>>((i>1)?i-1:0));
	lastrsB = rsBufPtrs[i] = NewTrigRSBuf(1L,trigSize,rsbsize,fill,
				    (lastrsB==NULL)?NULL:FeedTrigRSBuf,
				    (char *)lastrsB, NULL, (char *)i, 
				    rsDPtr0, desc);
	lag <<= 1;		/* lower buffers have to wait 2ce as long */
	}

    /* read sound into buffers, convert, write output file */
    if(inSnd->dataBsize > 0)
    fprintf(stderr,"Est frames = %ld\n", 
	2*(inSnd->dataBsize/(chans*sizeof(short)))
	       /hopSize);
    frame = 0;
    ndone = (1<<nOcts)-1;	/* preset all buffers as right on */
    while(ndone)
	{
	if((frame%100)==0)
	    fprintf(stderr,"%d ",frame);
	lev = OctOfFileFrame(frame);	/* which octave we're going to write */
	if(lev >= nOcts)
	    {
/*	    fprintf(stderr,"%d-OTT", lev);	/* an octave we don't want */
	    /* WriteANullBuffer(); */
	    ClearSampls(wrBuf, 1, (long)2*nBins);
	    fwrite(wrBuf, sizeof(float), 2*nBins, outFile);
	    }
	else{
	    while(!done && RBValidSize(rsBufPtrs[lev]->rsmpBufPtr)<kernLen)
		{	/* keep reading samples until enough filter down */
		red = ReadShortSampls(rdBuf, rdBlkSize*chans,inFile);
		if(chans > 1)
		    CopyFloats(rdBuf, chans, rdBuf, 1, rdBlkSize, REV);
		if(red < chans*rdBlkSize)
		    done = 1;
		FeedTrigRSBuf(rsBufPtrs[0], rdBuf, red/chans, 0);
		}
	    /* apply kernels & write result to file */
#ifdef DEBUGGRAPHS
	    if(graphs)
		DoDisp(wdptr[lev], RBBufPtr(rsBufPtrs[lev]->rsmpBufPtr,0,
					    kernLen),
		       rsBufPtrs[lev]->rsmpBufPtr->bufSize);
#endif
	    ApplyKernBlock(
		      RBBufPtr(rsBufPtrs[lev]->rsmpBufPtr,0,kernLen)+kernLen/2,
			   -kernLen/2, kernLen-kernLen/2, 
			   kernels, kernLen, 2*nBins, wrBuf);
	    Rect2Polar(wrBuf, nBins);
	    fwrite(wrBuf, sizeof(float), (long)2*nBins, outFile);
	    /* advance buffer by dropping oldest samples */
	    DropFromTrigRSBuf(rsBufPtrs[lev], hopSize);
	    /* stop when all buffers are under half full of valid stuff */
	    mask = (1L<<lev);
	    if(RBValidSize(rsBufPtrs[lev]->rsmpBufPtr)>=kernLen/2)
	        ndone |= mask;
	    else
		ndone &= ~mask;
	    }
/*	fprintf(stderr,"\n");	*/
	++frame;
	}
    fprintf(stderr, "%d\n",frame);
    PVCloseWrHdr(outFile, outPv);
    }

static void usage(void)		/* print syntax & exit */
    {
    fprintf(stderr,
 "Usage: %s [-q Qfact][-b bins/oct][-o octs][-s][-d][-v] inSnd [outFT]\n",
	    programName);
    fprintf(stderr,"where  (defaults in brackets)\n");
    fprintf(stderr,
 " -q Qfact   is the q-factor of the bins       (%6.1f)\n", (float)DFLT_QFCT);
    fprintf(stderr,
 " -b bins    is the number of filter per 8ve   (%6d)\n", DFLT_NBIN);
    fprintf(stderr,
 " -o octs    is the # of 8ves, starting at top (%6d)\n", DFLT_8VES);
    fprintf(stderr,
 " -g guard   is the proportion of the spectrum (%6.1f)\n", (float)GUARDBAND);
    fprintf(stderr,
 " -s         switches %3s 3dB/oct HF boost     (%3s)\n",
	    (!DFLT_SLOP) ? "in ":"out", DFLT_SLOP ? "in ":"out");
    fprintf(stderr,
 " -d         switches %3s kernel dc canc'l'n   (%3s)\n",
	    (!DFLT_DC) ? "in ":"out", DFLT_DC ? "in ":"out");
    fprintf(stderr,
 " -v         verbose mode\n");
    fprintf(stderr,
 " inSnd      is a SNDF soundfile for input\n");
    fprintf(stderr,
 " outFT      is a PVOC-type file,              (inStem+'%s')\n", DEF_EXT);
    exit(1);
    }

static void WriteFloatsAsIrc(flts, len, name)
    /* dump a vector of floats out to file as a soundfile */
    float	*flts;
    long 	len;
    char 	*name;		/* filename to write out as */
    {
    SFSTRUCT	*snd;
    long	l;
    short	*psh;
    int 	err = 0;

    if(err = SFAlloc(&snd, len*sizeof(short), SFMT_SHORT, (float)10000,
		  1, 0) )
	SFDie(err, NULL);
    psh = (short *)SFDataLoc(snd);
    for(l=0; l<len; ++l)
	{
	*psh++ = SHORTMAX * *flts++;
	}
    if(err = SFWrite(name, snd))
	SFDie(err, name);
    }
