/***********************************************************************\
*   db.groups.c								*
*   Functions to deal with track groups generated by other rules	*
*   generically								*
*   1993may05 dpwe							*
\***********************************************************************/

/********* switches ********/
#define DEBUGGRAPHS 1	/* display tracks in little windows */

/********* includes ********/
#include <dpwelib.h>
#include <genutils.h>		/* for Convo ??? (in DSPbench.h anyway) */
#include <vecops.h>
#include <math.h>

#include "DSPbench.h"
#ifdef TRAX			/* only do anything if in TRAX mode */
#ifdef CQFILE			/* AND CQ mode */

#include "db.tracks.h"

/********* defines *********/

/********* macros ***********/

/********* types ***********/

/***** static globals ******/

/******** prototypes ********/

int tg_is_subset(TKGROUP *tg, int type, TKGROUP *bigtg)
{	/* returns 1 if <tg->parts> is a subset of <bigtg->parts> */
    assert(type == T_TKGROUP);
    return is_subset((LIST)tg->parts, T_LIST, (LIST)bigtg->parts);
}

static int tg_prune_subsets(TKGROUP *tg, int type, TGLIST *tgl)
{	/* NodeFunc identifies subsets and removes them */
    TGLIST new;

    assert(type == T_TKGROUP);
    new = (TGLIST)remove_if(*tgl, (NodeFunc)tg_is_subset, tg);
    /* that will remove ourselves of course, so ..?  add ourselves back in? */
    new = (TGLIST)new_node(new, (LIST)tg, T_TKGROUP);
    free_list(*tgl);
    *tgl = new;
    return 1;
}

TGLIST tglRemoveSubsets(TGLIST tgl)
{	/* go through a list of tracks removing subsets */
    TGLIST rslt;

/*    tglPrint(tgl, "all");  */
/*    fprintf(stderr, "tglRmvSbs: size in = %d.. ", top_size((LIST)tgl));  */
    fflush(stderr);
    rslt = copy_top(tgl);
    maparg(tgl, (NodeFunc)tg_prune_subsets, &rslt);
/*    fprintf(stderr, "out = %d\n", top_size((LIST)rslt));	*/
/*    tglPrint(rslt, "w/o subs");  */
    return rslt;
}

double tkWeight(TRACK *tk)
{	/* `weight' of a track is its total energy */
    return LINOFDB(tk->st.mavg) * (tk->st.tmax - tk->st.tmin);
}

static int tk_calc_weight(TRACK *tk, int type, void *arg)
{	/* MAPpable fn to calc `weight' of a list of tracks.
	   <arg> points to a double which accumulates. */
    double *psum = (double *)arg;

    assert(type == T_TRACK);
    (*psum) += tkWeight(tk);
    return 1;
}

double tklWeight(TKLIST tkl)
{	/* calculate the `weight' of a list of tracks */
    double weight = 0.0;

    maparg(tkl, (NodeFunc)tk_calc_weight, &weight);
    if(weight > 0.0) 
	return DBOFLIN(weight);
    else 
	return MINUSALOT;
}

double tgWeight(TKGROUP *tg)
{	/* calculate the weight of an actual TKGROUP */
    return tklWeight(tg->parts);
}

int tg_weight_cmp(TKGROUP **ptg1, TKGROUP **ptg2)
{	/* for qsort of TKGROUPS into descending weight, return 1 if *ptg2 
	   points to a TKGROUP of lesser weight than *ptg1 */
    double w1 = tgWeight(*ptg1);
    double w2 = tgWeight(*ptg2);
    
    if(w2 > w1)		return 1;
    else if (w1 > w2)	return -1;
    else		return 0;
}

int tk_weight_cmp(TRACK **ptk1, TRACK **ptk2)
{	/* for qsort of TRACKS into descending weight, return 1 if *ptg2 
	   points to a TRACK of lesser weight than *ptg1 */
    double w1 = tkWeight(*ptk1);
    double w2 = tkWeight(*ptk2);
    
    if(w2 > w1)		return 1;
    else if (w1 > w2)	return -1;
    else		return 0;
}

double tgOvlpWeigh(TKGROUP *subject, TKGROUP *context)
{	/* figure the weight of the overlap between the (smaller) <subject>
	   track group and the larger <context>.  Return dB difference 
	   btwn overlap weight and total <subject> weight */
    double subWt = tgWeight(subject);
    double conWt = tgWeight(context);
    /* parts in <subject> not in <context> */
    TKLIST diff  = list_sub(subject->parts, context->parts);
    /* parts in <subject> that *are* in <context> */
    TKLIST ovlp  = list_sub(subject->parts, diff);
    double ovlWt = tklWeight(ovlp);
    double ovlMarg;

    if(ovlWt > MINUSALOT)		/* there *was* some overlap */
	ovlMarg = ovlWt - MAX(subWt, conWt);
    else
	ovlMarg = MINUSALOT;
/*    fprintf(stderr, "tkOvWt: sub %.2f  con %.2f  ovm %.2f\n",
	    subWt, conWt, ovlMarg);   */
    return ovlMarg;
}

#define WT_OVL_DB_THRESH	(10.0)	
/* if difference is 10dB below total weight, count track as subset */

static int tg_weight_ovl_pred(TKGROUP *tg, int type, TKGROUP *larger)
{	/* predicate returns 1 if <tg> is `close' to being a subset 
	   of <larger> i.e. if the weight of the difference is 
	   only a small proportion of the total weight */
    double totWt = tgWeight(tg);
    /* find tracks in this tg *not* in larger */
    TKLIST diff  = list_sub(tg->parts, larger->parts);
    double difWt = tklWeight(diff);

    /* wts are in dB, so proportion is just subtractive difference */
/*    fprintf(stderr, "tgWOvl: tot %.2fdB, dif %.2fdB\n", totWt, difWt);  */
    if(difWt < (totWt - WT_OVL_DB_THRESH))
	return 1;	/* dif was much lighter than tk: a (near-)subset */
    else {
	if(difWt < (totWt - 6.0))	/* dif is less than half of wt */
	    if(difWt < (tgWeight(larger) - 20.0))   /* dif vsmall c/w par */
		return 1;
	return 0;
    }
}

TGLIST tglRemoveNearDuplicates(TGLIST tgl)
{ 	/* given a list of track groups, will attempt to prune away 
	   subsets and near-subsets by following scheme:
	   (1) Sort all elements by `weight' so heaviest is first 
	   (2) Work down list of groups.  For each:
	       (2a) Compare with each element below it in the list.
	            If the weight of the intersect of that element's support 
		    with this group is above some threshold proportion 
		    of that element's total weight, the element is 
		    a subset or near-subset of the current track.  KILL
	       (2b) transfer this element to the output.
	   (3) continue with the rest of the list, if any */
    TKLIST diff;
    double totWeight, partWeight;
    TGLIST rslt = NULL, *rsTail, tgl2, this;

/*    fprintf(stderr, "tglRmvDup: size in = %d.. ", top_size((LIST)tgl));  */
    fflush(stderr);
    /* first sort by weight */
    tgl = copy_top(tgl);	/* make our own copy we can mess with */
    tgl = (TGLIST)qsort_list((LIST)tgl, (NodeCmpFn)tg_weight_cmp);
    /* now do the loop thing */
    rsTail = &rslt;
    while(tgl)  {
	/* unlink 1st element - the current fave (heaviest remaining) */
	this = tgl;
	tgl = (TGLIST)CDR((LIST)tgl);
	set_next((LIST)this, NULL);
	assert(CARTYPE(this) == T_TKGROUP);
	/* strip out lighter elements that are pretty much subsets */
	tgl2 = (TGLIST)remove_if(tgl, (NodeFunc)tg_weight_ovl_pred, 
				 CAR((LIST)this));
	free_list(tgl);
	tgl = tgl2;
	/* copy the current one to the result list */
	*rsTail = this;
	advance_ptr(&rsTail);
    }
/*    tglPrint(rslt, "tglRmv");	*/
/*    fprintf(stderr, "out = %d\n", top_size((LIST)rslt));	*/
    return rslt;
}

int tg_nest_addg(TKGROUP *toadd, int type, TKGROUP *nesttg)
{	/* MAPpable fn to add tkgroups into a nested tkgroup.
	   does NOT update support field (for now) */
    assert(type == T_TKGROUP);
    tkgrAddTracks(nesttg, toadd->parts);
    return 1;
}

TKGROUP *tgNewNested(TGLIST tgl)
{	/* make a `nested' TKGROUP out of the TKGROUPs in the list.
	   A nested TKGROUP has its own parts TKLIST and a special 
	   TGLIST of TKGROUPs upon which it is based. */
    TKGROUP *tg = tkgrNew(NULL, TG_NEST);

    maparg(tgl, (NodeFunc)tg_nest_addg, tg);
    /* may as well sort the parts list into canonical order */
    tg->parts = (TKLIST)qsort_list((LIST)tg->parts, tkCmpTrxByT);
    /* set up `nest' info field to list all supporting groups */
    tg->v.nest.support = (TGLIST)copy_top((LIST)tgl);
    return tg;
}

int tgNestedMerge(TKGROUP *ntg, TKGROUP *toadd)
{	/* add another tkgroup to a nested tkgroup.
	   Merge all the parts into the parent. */
    tkgrAddTracks(ntg, toadd->parts);
    /* add this tkgroup to the nested's support list */
    *(find_last_ptr((LIST *)&(ntg->v.nest.support))) 
	= new_node(NULL, (LIST)toadd, T_TKGROUP);
    return top_size(ntg->v.nest.support);
    /* return total number of supporting gps */
}

#define TGS_OVL_MARG	(-9.0)		/* overlap abv this, close thing */

static TGLIST *stSupGpPtr = NULL;	/* static for gathering supergroups */

int tg_cmp_grp_grp(TKGROUP *gp1, int type, TKGROUP *gp2)
{	/* MAPpable fn compares two groups (for outer-prod 
	   comparison of two tglists) */
    char s1[32], s2[32], s3[80];
    double margin;
    
    s1[31] = '\0'; 
    s2[31] = '\0'; 
    tkgrTitle(gp1, s3);
    strncpy(s1, s3, 31);
    tkgrTitle(gp2, s3);
    strncpy(s2, s3, 31);
/*    fprintf(stderr, "'%s'<>'%s'\n", s1, s2);	*/
    margin = tgOvlpWeigh(gp1, gp2);
    if(margin >= TGS_OVL_MARG)  {
	TKGROUP *new;

/*	fprintf(stderr, "******************\n");	*/
	new = tgNewNested(NULL);
	tgNestedMerge(new, gp1);
	tgNestedMerge(new, gp2);
	/* add onto a static ptr for now */
	*stSupGpPtr = (TGLIST)new_node(NULL, (LIST)new, T_TKGROUP);
	advance_ptr(&stSupGpPtr);
    }
    return 1;
}

int tg_cmp_grp_list(TKGROUP *gp1, int type, TGLIST *others)
{	/* MAPpable fn to compare two lists of tkgroups - 
	   i.e. compare each member of first list against 
	   all members of 2nd list in turn */
/*    fprintf(stderr, "tk_cmp_grps:\n");  */
    maparg((LIST)others, (NodeFunc)tg_cmp_grp_grp, gp1);
    return 1;
}

#define SEED_MTHRESH	(-5.0)	/* *must* have mmax > -5dB to be seed */
#define SEED_PROPORTION	(0.10)	/* at most 10% of tracks are seeds */

TGLIST tgSupGpsForFocus(double tmin, double tmax, TKLIST allTx)
{	/* generate all the supergroups over the specified focus 
	   from the passed trax universe */
    TGLIST	hGps, oGps, nGps, t2, supGps = NULL;
    TKLIST 	focus, ofocus, seeds, s2;
    long 	focussize, seedsize;
static SELPNL_STR 	*spnl1 =  NULL;

/*    if(spnl1 == NULL)  {
	spnl1 = spnlNew("Noises");
    } else {
	spnlReset(spnl1);
	spnlUnhide(spnl1);
    }
*/
    focus = tkGetIntersect(allTx, tmin, tmax, 0.0, 44100.0);	/* all frq */
    focussize = top_size(focus);
    /* actual focus must (?) be in ascending tmin order */

    /* now find the `seed' tracks.
       Currently, just highest `weight' (total energy) - 
       take ? up to 10% if they are above some min magnitude */
    seeds = SelTrxAbvMag(focus, SEED_MTHRESH);
    seeds = (TKLIST)qsort_list((LIST)seeds, (NodeCmpFn)tk_weight_cmp);
    seedsize = SEED_PROPORTION * focussize;
    if(top_size(seeds) > seedsize)  {
	/* too many - throw away lighter */
	s2 = (TKLIST)nth((LIST)seeds, seedsize - 1);
	s2 = (TKLIST)break_after((LIST)s2);
	free_list(s2);		/* dispose of excess */
    }
/*    RedrawTrax(focus, TKM_SELECT);  */
    RedrawTrax(seeds, TKM_SPECIAL); 
    seedsize = top_size(seeds);
    fprintf(stderr, "focus %.6f..%.6f sec, fsz %d, ssz %d\n", 
	    tmin, tmax, focussize, seedsize);

    /* make all harmonic groups */
    hGps = lmappend(seeds, (LNodeFunc)tgl_harm_tk, focus);
    t2   = tglRemoveNearDuplicates(hGps);
    free_list((LIST)hGps);	hGps = t2;

    /* now all onset groups */
    /* (set up by clearing prominence fields) */
    ofocus = tklOnsetsPrepare(focus);	/* onset has pruned focus */
    oGps = lmappend(seeds, (LNodeFunc)tgl_onset_tk, ofocus);
    free_list(ofocus);
    t2   = tglRemoveNearDuplicates(oGps);
    free_list((LIST)oGps);	oGps = t2;
/*    t2	 = tgl_cont_tk(CAR(oGps), CARTYPE(oGps), focus);   */

    /* grow each onset into a noise region, maybe */
    nGps = lmappend(oGps, (LNodeFunc)tgl_noise_tg, focus);
    t2   = tglRemoveNearDuplicates(nGps);
    fprintf(stderr, "o->nGps %d->%d->%d\n", top_size(oGps), 
	    top_size(nGps), top_size(t2));
    free_list((LIST)nGps);	nGps = t2;

    stSupGpPtr = &supGps;	/* global gathers supergroups in 
				   tg_cmp_grp_list */
    maparg((LIST)hGps, (NodeFunc)tg_cmp_grp_list, (void *)oGps);


    supGps = append(supGps, nGps);	/* attach noises too */

    return supGps;
}

#define FOCUS_TIME	(0.5)	/* consider tracks over 0.5 s span */
    
TGLIST tgGenAll(WINRCD **wrcd)
{	/* automatically generate track groups */
    double 	time, tstt, tend, tmin, tmax, dur;
    TGLIST	gps, allgps = NULL, t2;
static SELPNL_STR 	*spnl1 =  NULL;
static SELPNL_STR 	*spnl2 =  NULL;
static SELPNL_STR 	*spnl3 =  NULL;
    extern int SynthFromTracks(WINRCD **wrcd);


    noisyInitStatics(wrcd, 0);

    if(spnl3 == NULL)  {
/*	spnl1 = spnlNew("Focus Hs");
	spnl2 = spnlNew("Focus Os");
 */
 	spnl3 = spnlNew("SuperGroups");
/*	spnlSetHackFn(spnl3, SynthFromTracks, wrcd, "play");	*/
	/* hack fns set in spnlFromTkgList now */
    } else {
	/* should dispose of all last time's stuff here */
/*	spnlReset(spnl1);
	spnlUnhide(spnl1);
	spnlReset(spnl2);
	spnlUnhide(spnl2);
 */
	spnlReset(spnl3);
	spnlUnhide(spnl3);
    }
    /* select half-second in middle of current view */
    VTimeLims(wrcd, &time, &tend, &dur);
/*    time = 0.5 * (time+tend);	/* middle of view */
/*    tstt = time - (FOCUS_TIME/2);
    tend = time + (FOCUS_TIME/2);
 */
    tmin = time;
    while(tmin + (0.75*FOCUS_TIME) < tend)  {	/* scan focus over whole lot */
	tmax = tmin + FOCUS_TIME;
	gps = tgSupGpsForFocus(tmin, tmax, allTrax);
	allgps = append(allgps, gps);
	tmin += FOCUS_TIME/2.0;
    }

    t2   = tglRemoveNearDuplicates(allgps);
    free_list((LIST)allgps);	allgps = t2;
    
    spnl3 = spnlFromTkgList(spnl3, allgps, wrcd);

    return gps;
}    


#else /* !CQFILE */   

    /* dummy functions here */

#endif /* CQFILE */   

#endif /* TRAX */   

