/* 
 * gendsp.c
 * random processing-type functions I wanted to take out of genutils.
 * dpwe 1994mar29
 */

#include "dpwelib.h"
#include <math.h>
#include <gendsp.h>

#ifdef MIPS	/* #if (MFG == MIPS)	*/
#include <mips/fpu.h>	/* for TrapFPEs() */
#endif

#define LOGTABSZ 32
/* 32 log-spaced steps between POSMAX and POSMAX/2 for 8-bit res logmaker */
static long logTable[] = 
    { 16743, 17109, 17484, 17867, 18258, 18658, 19066, 19484, 
      19911, 20347, 20792, 21247, 21713, 22188, 22674, 23170,
      23678, 24196, 24726, 25268, 25821, 26386, 26964, 27554, 
      28158, 28774, 29405, 30048, 30706, 31379, 32066, 32767 };

#define PMAXBITS 7
#define PMAX     0x7FFF
#define PMAXD2   (PMAX>>1)
#define LOGNEGMAX (LOGTABSZ*(PMAXBITS-15))

long intLog(n)	/* an integer algorithm for log scale conversion */
    /* intLog(32767) = 0xFF
       intLog(16384) = 0xE0
       intLog( 8192) = 0xC0
       intLog( 4096) = 0xA0
       intLog( 2048) = 0x80
       intLog( 1024) = 0x60
       intLog(  512) = 0x40
       intLog(  256) = 0x20
       intLog(  128) = 0x00
     */
    long    n;
    {
    long    log = PMAXBITS;
    long    i;

    if(n < 0)
	n = -n;
    if(n == 0)
	return(LOGNEGMAX);
        /* or at least the most  negative number I can think of */
    if(n<=PMAXD2)
	{
	while(n <= PMAXD2)
	    {
	    n <<= 1;
	    --log;
	    }
	}
    else if(n>PMAX)
	{
	while(n>PMAX)
	    {
	    n >>= 1;
	    ++log;
	    }
	}
    /* n now in range pmaxd2 <= n < pmax */
    for(i = 0; i<LOGTABSZ; ++i)
	{
	if(n <= logTable[i])
	    break;
	}
    return (LOGTABSZ * log + i);
    }

#define PRIMETABSIZE	10
/* little prime bootstrap table */
static int primeTab[] = { 2,3,5,7,11,13,17,19,23,29 };
#define PRIMEMAX 10000	/* refuse to search for primes above here */

int NextPrime(n)	/* returns smallest prime bigger than n */
    int n;
    {
    int i = PRIMETABSIZE-1;
    int m;

    if(n<primeTab[i])
	{
	while( primeTab[i] > n && i>=0 )
	    --i;
	return(primeTab[++i]);
	}
    else{
	m = n + ((n&1)?2:1);	/* next odd number */
	while(!TestPrime(m) && m<PRIMEMAX)
	    m += 2;
	if(m >= PRIMEMAX)
	    {
	    fprintf(stderr, "NextPrime: overflow for %d\n", n);
	    abort();
	    }
	return m;
	}
    }
    
int TestPrime(m)	/* test m as a prime number */
    int m;
    {
    int lim;
    int root;
    int tabIndex = 0;

    if(m<2)	return 0;
    lim = (int)sqrt((float)m);	/* largest potential root we need check */
    while(tabIndex<PRIMETABSIZE)
	{
	root = primeTab[tabIndex++];
	if(root>lim)		return 1;	/* found no roots */
	if( (m % root) == 0)	return 0;	/* root is a root */
	}
    while(root < lim)
	{
	root = NextPrime(root);		/* recursiv.. */
	if( (m%root) == 0)	return 0;	/* found root */
	}
    return 1;			/* no root found below sqrt(m) => pass */
    }

long dumbGCF(n1, n2)	/* return largest common factor of n1 & n2 */
    long n1, n2;	
    {	/* approach is to find prime roots of n1 & n2 */
    long gcf = 1;	/* fall back */
    long tmp, root, lim;

    if(n1>n2)	{ tmp = n2; n2 = n1; n1 = tmp; }	/* make n1 smallest */
    if(n1<1)	return 1;		/* what's going on? */
    root = primeTab[0];			/* seed with first prime */
    lim  = (int)sqrt((float)n1);	/* no common prime factors bigger */
    while(root<=lim)
	{
	if( (n1%root)==0 && (n2%root)==0)
	    {
	    gcf *= root;
	    n1 /= root;
	    n2 /= root;
	    }
	else
	    root = NextPrime(root);
	}
    return gcf;
    }

long GCF(n1,n2)		/* use the proper algorithm ! */
    long n1,n2;
    {
    long tmp;

    if(n1>n2)	{ tmp = n2; n2 = n1; n1 = tmp; }	/* make n1 smallest */
    if(n1 < 1)		return n2;
    if(n1 == 1)		return 1;		/* not necessary */
    return GCF( (n2%n1), n1);
    }

void TrapFPEs()
    {
#ifdef MIPS	/* #if (MFG == MIPS)	*/
    union fpc_csr fpcr;

    fpcr.fc_word = get_fpc_csr();
/*    fpcr.fc_word |= FPCSR_ENABLES;	/* all exceptions passed to Unix */
    fpcr.fc_struct.en_invalid = 1;	/* just NaN exceptions rejected */
    fpcr.fc_struct.en_divide0 = 1;	/* catch div0 as well */
    set_fpc_csr(fpcr.fc_word);
#endif
    }

float InnerProd(ptr1, stp1, ptr2, stp2, len)	/* compute inner product */
    float	*ptr1;
    int 	stp1;
    float	*ptr2;
    int 	stp2;
    long	len;
    {
    float	rslt = 0.0;

    while(len--)
	{
	rslt += *ptr1 * *ptr2;
	ptr1 += stp1;
	ptr2 += stp2;
	}
    return rslt;
    }

#define MAXKLEN 4096
static float *dumstor = NULL;

void Convo(src,sinc,slen,kern,kinc,klen)  /* convolve a signal with a kernel */
    float	*src;
    long	sinc;
    long	slen;
    float	*kern;
    long	kinc;
    long	klen;
    {
    long	i;
    long	j, k, l, m, u;
    long	koffs;
    double	a, t;

    if(!dumstor)
	dumstor = (float *)Mymalloc((long)MAXKLEN, sizeof(float), "genu:conv");
    koffs = (klen-1) / 2;		/* align to 'centre' of kernel */
    for(i = 0; i < slen; ++i)
	{
	t = 0.0;
	l = 0;	u = klen;
	if( (j = koffs-i) > 0)	{
	    /* underrun of kernel and signal - replicate lowest sig bin */
	    l = j;
	    a = src[sinc*0];
	    for(m = 0; m<l; ++m)  {
		t += a * kern[kinc*(klen - 1 - m)];
	    }
	}
	if( (j = i + klen - koffs - slen) > 0)	{
	    /* kernel runs off end of signal - replicate last point */
	    a = src[sinc*(slen - 1)];
	    for(m = 0; m<j; ++m)  {
		t += a * kern[kinc*(klen - 1 - (--u))];
	    }
	    /* i.e. the effect is u -= j;  */
	}
	k = i + l - koffs;
	for(j = l; j < u; ++j, ++k)
		t += kern[kinc*(klen - 1 - j)] * src[sinc*k];
	if(i<koffs)
	    dumstor[i] = t;
	else
	    src[sinc*(i-koffs)] = t;
	}
    for(i = slen-1; i >= 0; --i)
	{
	if(i < koffs)
	    src[sinc*i] = dumstor[i];
        else
	    src[sinc*i] = src[sinc*(i - koffs)];
	}
    }

void Overlay(ptr1, stp1, ptr2, stp2, len, coef)	/* superpose scaled kernel */
    float	*ptr1;
    int 	stp1;
    float	*ptr2;
    int 	stp2;
    long	len;
    FLOATARG	coef;
    {
    if(len)
	do  {
	    *ptr2 += coef * *ptr1;
	    ptr1 += stp1;
	    ptr2 += stp2;
	    } while(--len);
    }

double maskPhs(phs)	/* do it inline instead! */
    FLOATARG   phs;
    {
    while(phs > PI)
	{
	phs -= 2*PI;
	}
    while(phs < -PI)
	{
	phs += 2*PI;
	}
    return(phs);
    }
