/***************************************************************\
* idrawop.c							*
* Provide graphics primitives that write an idraw file op	*
* Based on stripk		 				*
* Generalized dpwe 1992oct24					*
* 1994mar15 modified to write files for current idraw (SGI)	*
\***************************************************************/

#include "idrawop.h"
#include <time.h>

#define PSMAXY 447		/* 'top' of PostScript pixel range */
#define PSMINX 8		/* left of PS pixels */
#define TEXTFIXX (-4)		/* shift text left - better */
#define TEXTFIXY (2)		/* shift text upward - better */

#define ID_DFLT_DPS	0	/* default number of decimal places */

/* postscript context */
static char *psfont = NULL;
static char *idwfont = NULL;
static int  idNdps = 0;
static char *idFmt = NULL;

/* offsets from ps global scaling:
       [ 0.8 0 0 0.8 20 10 ] concat
       /originalCTM matrix currentmatrix def
   to allow bounding box mapping in true co-ordinates 
*/
static double scaleoffsx = 20;
static double scaleoffsy = 10;
static double xscale = 0.8;
static double yscale = 0.8;

void idMap(int x, int y, int *mx, int *my)
{	/* translate co-ordiantes for bounding box calcs */
    *mx = scaleoffsx + xscale*(PSMINX+x);
    *my = scaleoffsy + yscale*(PSMAXY-y);
}

static void idInitStatics(void)
{   /* allocate the static data */
    if(psfont==NULL)  {
	psfont = TMMALLOC(char, 64, "idInitS:psf");
	idwfont = TMMALLOC(char, 64, "idInitS:idw");
        idFmt = TMMALLOC(char, 16, "idInitS:idf");
    }
}

void idSetDPs(WIND *wind, int ndps)
{	/* Make it such that all future calls to idraw output routines 
	   specify co-ordinates to this many decimal places of accuracy. 
	   Must be zero if the output file is to be read by Idraw. */
    idNdps = ndps;
    sprintf(idFmt, "%%3.%df", idNdps);
}

#define TPLTFNAME "tplt.ps"	/* name of template file */

FILE *idTemplateFile(void)
{   /* find the idraw template file */
    char fname[256];
    char home[256];
    char *env;
    FILE *file;

    if( (env = getenv("PSLIB")) == NULL)  {
	if( (env = getenv("HOME")) == NULL)  {
	    env = "/usr/local";
	}
	sprintf(home, "%s/lib/ps", env);
    } else {
	strcpy(home, env);
    }
    sprintf(fname, "%s/%s", home, TPLTFNAME);
    if( (file = fopen(fname, "r")) == NULL)  {
	fprintf(stderr, "idOpen: could not find template file %s in \n", 
		TPLTFNAME);
	fprintf(stderr, "        $PSLIB $HOME/lib/ps /usr/local/lib/ps\n");
    }
    return file;
}

#define UNREASONABLE_LINE_LENGTH	256
static int GetLine(FILE *input, char *buffer)
{	/* fill a buffer until, and not including, the next '\n'.
	   Return the number of chars read, negative if hit EOF.
	   Just pray we won't overrun the buffer.  */
    int     c;
    int     n = 0;

    while( (c = getc(input))!=EOF && c != '\n')  {
	*buffer++ = c;
	++n;
	if(n == UNREASONABLE_LINE_LENGTH)  {
	    fprintf(stderr, "GetLine: unreasonable line length %d\n", n);
	    break;
	}
    }
    if( c == EOF)
	n = -MAX(1,n);	/* (make sure it's at least -1 even if n == 0) */
    *buffer = '\0';	/* add Cstr terminator */
    return n;
}

static int CopyFilesUntil(FILE *srcf, FILE *dstf, char *termstr)
{	/* copy from one text file into another, stopping 
	   directly before we copy the string matching <termstr>.
	   Return 1 if termstr found, 0 if stopped at EOF. */
    char linebuf[UNREASONABLE_LINE_LENGTH];
    int  done = FALSE;
    char *where;

    if(dstf == NULL)  {		/* passing dst as NULL means skip/discard */
	dstf = fopen("/dev/null", "w");		/* .. so write to devnull */
    }
    while(!done)  {
	done = (GetLine(srcf, linebuf)<0);
	if( (where = strstr(linebuf, termstr)) != NULL)  {
	    /* copy across up to that point ..
	       but lose rest of the line?? */
	    fwrite(linebuf, 1, (where-linebuf), dstf);
	    return 1;
	} else {	/* termstr not in this line, copy the lot */
	    fprintf(dstf, "%s\n", linebuf);
	}
    }
    fprintf(stderr, "CopyUntil: never found '%s'\n", termstr);
    return 0;		/* never found termstr */
}

void idOpen(WIND *wind, char *fname, int x, int y, int w, int h)
{   /* open a file to capture draws to the WIND as postscript.
       x,y,w,h are the bounding box, if known.  Set w & h to zero if not */
    FILE *tpltFile, *psFile;
    char timestr[64];
    int  mx, my;
    time_t tm;

    if( (wind->opf = fopen(fname, "w")) == NULL)  {
	fprintf(stderr, "idOpen: couldn't create idraw output file '%s'\n", 
		fname);
    }
    /* initialize the file - copy the template */
    psFile = wind->opf;
    if(tpltFile=idTemplateFile())  {
	/* copy across some of template */
	if(CopyFilesUntil(tpltFile, psFile, "*TITLE*") == 0)  {
	    /* never found the token */
	    fprintf(stderr, "PSout: internal error reading template\n");
	    abort();
	}
	fprintf(psFile, "%s\n", fname);
	CopyFilesUntil(tpltFile, psFile, "*DATE*");
	time(&tm);
	strcpy(timestr, ctime(&tm));
	timestr[strlen(timestr)-1] = '\0';	/* kill terminating CR */
	fprintf(psFile, "%s\n", timestr);
	CopyFilesUntil(tpltFile, psFile, "*BBOX*");
	if( w == 0 || h == 0)  {	/* no real bbox */
	    fprintf(psFile, " ? ? ? ?\n%%PageBoundingBox: 30 31 582 761\n");
	    /* (words for PrintMonitor!) */
	} else {
	    idMap(x,y+h, &mx, &my);
	    fprintf(psFile, " %d %d", mx, my-2);
	    idMap(x+w,y, &mx, &my);
	    fprintf(psFile, " %d %d\n", mx, my+2);
	}
	CopyFilesUntil(tpltFile, psFile, "*BODY*");
	/* that's all for now */
	fclose(tpltFile);
    }
    idInitStatics();
    idSetDPs(wind, ID_DFLT_DPS);
}

void idClose(WIND *wind)
{
    FILE *idf = wind->opf;
    FILE *tpltFile;

    if(idf)  {
	if(tpltFile = idTemplateFile())  {
	    /* read trailer from template file */
	    CopyFilesUntil(tpltFile, NULL, "*BODY*");
	    CopyFilesUntil(tpltFile, idf, "*END*");
	}
	fclose(idf);
	wind->opf = NULL;
    }
}

static idwFgBg(WIND *wind, int bgtoo)
{	/* output foreground and (if bgtoo nonzero) background defs */
    FILE *idf = wind->opf;

    if(idf)  {
	if(wind->fg == WGMD_BLACK)  {
	    fprintf(idf,"%%I cfg Black\n0 0 0 SetCFg\n");
	    if(bgtoo)	fprintf(idf,"%%I cbg White\n1 1 1 SetCBg\n");
	} else if(wind->fg == WGMD_BLUE)  {
	    fprintf(idf,"%%I cfg Blue\n0.3 0.3 1.0 SetCFg\n");
	    if(bgtoo)	fprintf(idf,"%%I cbg White\n1 1 1 SetCBg\n");
	} else {	/* assume reverse video */
	    fprintf(idf,"%%I cfg White\n1 1 1 SetCFg\n");
	    if(bgtoo)	fprintf(idf,"%%I cbg Black\n0 0 0 SetCBg\n");
	}
    }
}

static void idwConcat(WIND *wind)
{	/* output oblig context line (each element has a CTM?) */
    FILE *idf = wind->opf;

    if(idf) {
	fprintf(idf,"[ 1 0 0 1 0 0 ] concat\n");
    }
}

void idLineTo(WIND *wind, int x, int y)
{
    FILE *idf = wind->opf;

    if(idf)  {
	fprintf(idf,"Begin %%I Line\n%%I b 65535\n%1.0f 0 0 [] 0 SetB\n",
		wind->linew);
	idwFgBg(wind,1);
	fprintf(idf,"none SetP %%I p n\n%%I t\n");
	idwConcat(wind);
	fprintf(idf,"%%I\n%d %d %d %d Line\n%%I 1\nEnd\n\n", 
		PSMINX+wind->cur_x, PSMAXY-wind->cur_y, PSMINX+x, PSMAXY-y);
	fflush(idf);
    }
}

#define PSLSMAX	200	/* max segments in line */

void idPolyLine(WIND *wind, int buf[][2], long cnt)
{	/* poly line independent of line from/to */
    int 	i;
    float	pslinew = 1.0;
    FILE *idf = wind->opf;

    if(idf)  {
	while(cnt>PSLSMAX)  {
	    idPolyLine(wind, buf, PSLSMAX);
	    buf += PSLSMAX - 1;
	    cnt -= PSLSMAX - 1;
	}

	fprintf(idf,"Begin %%I MLine\n%%I b 65535\n%1.0f 0 0 [] 0 SetB\n", 
		wind->linew);
	idwFgBg(wind,1);
	fprintf(idf,"none SetP %%I p n\n%%I t\n");
	idwConcat(wind);
	fprintf(idf, "%%I %d\n", cnt);
	for(i = 0; i<cnt; ++i)
	    fprintf(idf, "%d %d\n", PSMINX+buf[i][0], PSMAXY-buf[i][1]);
	fprintf(idf, "%d MLine\n%%I 1\nEnd\n\n", cnt);
	fflush(idf);
    }
}

void idFrame(WIND *wind, RECT *pr)
{
    FILE *idf = wind->opf;

    if(idf)  {
	fprintf(idf,"Begin %%I Rect\n%%I b 65535\n%1.0f 0 0 [] 0 SetB\n", 
		wind->linew);
	idwFgBg(wind,1);
	fprintf(idf,"none SetP %%I p n\n%%I t\n");
	idwConcat(wind);
	fprintf(idf,"%%I\n%d %d %d %d Rect\nEnd\n\n", 
		PSMINX+pr->x, PSMAXY-pr->y,
		PSMINX+pr->x+pr->width, PSMAXY-(pr->y+pr->height));
	fflush(idf);
    }
}

void idFill(WIND *wind, RECT *pr)
{
    FILE *idf = wind->opf;

    if(idf)  {
	fprintf(idf,"Begin %%I Rect\n%%I b 65535\n%1.0f 0 0 [] 0 SetB\n", 
		wind->linew);
	idwFgBg(wind,1);
	fprintf(idf,"%%I p\n0 SetP\n%%I t\n");
	idwConcat(wind);
	fprintf(idf,"%%I\n%d %d %d %d Rect\nEnd\n\n", 
		PSMINX+pr->x, PSMAXY-pr->y,
		PSMINX+pr->x+pr->width, PSMAXY-(pr->y+pr->height));
	fflush(idf);
    }
}

#define IDGRY_BYTESPLINE	30	/* 30 bytes each line of output */
static int idgryColumn = 0;	/* state variable aligns columns for gry */
#define COLSPERBYTE		2	/* each written byte occupies 2 chrs */

long idgryWriteBytes(FILE *file, MyByte *data, long nbytes)
{	/* write out some bytes into a hex gray region, wrapping lines 
	   according to idryColumn and IDGRY_BYTESPLINE */
    int 	leftThisLine;
    int		thisTime;
    int 	i;
    long	done = 0;

    while(nbytes)  {
	leftThisLine = IDGRY_BYTESPLINE - idgryColumn/COLSPERBYTE;
	thisTime = MIN(leftThisLine, nbytes);
	for(i = 0; i < thisTime; ++i)  {
	    fprintf(file, "%02x", *data++);
	}
	idgryColumn += COLSPERBYTE * thisTime;
	if(idgryColumn >= IDGRY_BYTESPLINE * COLSPERBYTE) { /* line ended */
	    fprintf(file, "\n"); /* move to next line */
	    idgryColumn = 0;
	}
	nbytes -= thisTime;
	done += thisTime;
    }
    return done;
}

#define RLE_MAX_BLOCK	128	/* RLE uses 7 bits to hold block length */
/* format is this:  positive byte <n> => <n+1> unrepeated bytes follow 
                    negative byte <-n> => <n> repetitions of next byte */
#define RLE_MIN_RPTS	3	/* same byte three times kicks in RLE */

long rleWriteNonChunk(FILE *file, MyByte *data, long remBytes)
{	/* write out rle data that isn't actually runlengthed */
    /* find how far we have to go before finding repeated bytes */
    long 	searchRange = RLE_MAX_BLOCK;
    int		i, j, found = 0;
    long	done = 0;
    MyByte	rleNum, d;

    while(!found && remBytes)  {
	searchRange = MIN(RLE_MAX_BLOCK, remBytes - RLE_MIN_RPTS + 1);
	for(found = 0, i = 0; !found && i < searchRange; ++i)  {
	    d = data[i];
	    found = 1;
	    for(j = 1; found && j < RLE_MIN_RPTS; ++j)  {
		if(d != data[i+j])	found = 0;
	    }
	}
	if(found)	--i;	/* found means last byte should NOT be sent */
	/* We couldn't search right up to the end of the data since 
	   we can't search for repeats over the end of the data.  If 
	   we got really close to the end, add in that data to 
	   unrepeated block */
	if( (remBytes - i) < RLE_MIN_RPTS )
	    i = MIN(RLE_MAX_BLOCK, remBytes);
	if(i>0)  {		/* some unrepeated data */
	    rleNum = i - 1;	/* write out byte <= 127 as num follow */
	    idgryWriteBytes(file, &rleNum, 1);
	    /* add following bytes */
	    idgryWriteBytes(file, data, i);
	}
	done += i;
	data += i;
	remBytes -= i;
    }
    return done;
}

long rleWriteRleChunk(FILE *file, MyByte *data, long rembytes)
{	/* write out a series of repeated bytes as RLE data */
    long	repeats = 0;
    MyByte	rleNum, d;
    int		thisTime;
    long	done = 0;

    d = *data;
    /* count how many of these we get */
    while(repeats < rembytes && d == *data++)		++repeats;
    /* write out the blocks */
    while(repeats >= RLE_MIN_RPTS)  {
	thisTime = MIN(RLE_MAX_BLOCK, repeats);
	rleNum = 256-(thisTime);	/* -ve indicates repeat */
	idgryWriteBytes(file, &rleNum, 1);
	idgryWriteBytes(file, &d, 1);
	repeats -= thisTime;
	done += thisTime;
    }
    return done;
}

long rleWriteData(FILE *file, MyByte *data, long rembytes)
{	/* Write out some asciihex data, RLE coded */
    long 	done = 0;
    long 	thistime;

    /* reset output column alignment static */
    idgryColumn = 0;
    
    while(rembytes)  {
	thistime = rleWriteNonChunk(file, data, rembytes);
	rembytes -= thistime;
	data += thistime;
	if(rembytes) {
	    thistime = rleWriteRleChunk(file, data, rembytes);
	    rembytes -= thistime;
	    data += thistime;
	}
    }
    if(idgryColumn)  {
	fprintf(file, "\n");
	idgryColumn = 0;
    }
    return done;
}

void idRectRleGraymap(WIND *wind, RECT *pr, ByArray *array)
{	/* render an array of bytes as a postscript graymap 
	   into a specified rectangle (with scaling)
	   but use run-length encoding copied from pnmtops output. */
    FILE *idf = wind->opf;
    int  rows, cols;
    long rembytes, thistime, i;
    MyByte *data;
    int x = pr->x;
    int y = pr->y;
    int w = pr->width;
    int h = pr->height;

    if(idf)  {
	rows = array->rows;
	cols = array->cols;
	fprintf(idf, "Begin %%I RleGraymap\n/picstr %d string def\n", cols);
	fprintf(idf, "%d %d translate\n", PSMINX+x, PSMAXY-y-h);
	fprintf(idf, "%d %d scale\n%d %d 8\n", w, h, cols, rows);
	fprintf(idf, "[ %d 0 0 -%d 0 %d ]\n", cols, rows, rows);
	fprintf(idf, "{ picstr readstring }\nimage\n");

	rembytes = rows*cols;
	data = array->data;

	for(i = 0; i < rembytes; ++i)
	    data[i] = BYTEMAX - data[i];	/* flip sense */

	/* rleWriteData(idf, data, rembytes); */
	/* rle code within rows only ? */
	for(i = 0; i < rows; ++i)  {
	    rleWriteData(idf, data + i * cols, cols);
	}

	for(i = 0; i < rembytes; ++i)
	    data[i] = BYTEMAX - data[i];	/* flip back */

	fprintf(idf, "End\n\n");
    }
}

void idRleGraymap(WIND *wind, int x, int y, ByArray *array)
{	/* render an array of bytes as a postscript graymap 
	   but use run-length encoding copied from pnmtops output. */
    RECT r;

    r.x = x;
    r.y = y;
    r.width = array->cols;
    r.height = array->rows;
    idRectRleGraymap(wind, &r, array);
}

void idGraymap(WIND *wind, int x, int y, ByArray *array)
{	/* render an array of bytes as a postscript graymap */
    FILE *idf = wind->opf;
    int  rows, cols;
    long rembytes, nowbytes, i;
    MyByte *data;

    if(idf)  {
	rows = array->rows;
	cols = array->cols;
	fprintf(idf, "Begin %%I Graymap\n/picstr %d string def\n", cols);
	fprintf(idf, "%d %d translate\n", PSMINX+x, PSMAXY-y-rows);
	fprintf(idf, "%d %d scale\n%d %d 8\n", cols, rows, cols, rows);
	fprintf(idf, "[ %d 0 0 -%d 0 %d ]\n", cols, rows, rows);
	fprintf(idf, "{ currentfile picstr readhexstring pop }\nimage\n");
	/* Now write out image bytes in the correct way: by row, 
	   from top left (the same order as in array->data).
	   Postscript format is two-character asciihex per byte;
	   all white is FF.  Sixty chars per line, say */
	rembytes = rows*cols;
	data = array->data;
	while(rembytes)  {
	    nowbytes = MIN(rembytes, 30);
	    for(i = 0; i < nowbytes; ++i)  {
		fprintf(idf, "%02x", BYTEMAX - *data++);
	    }
	    fprintf(idf, "\n");
	    rembytes -= nowbytes;
	}
	fprintf(idf, "End\n\n");
    }
}

void idSetFont(WIND *wind, int fontindex)
{
    int size;
    char *basePsFont, *baseIdFont;

    idInitStatics();
    if(fontindex == IDF_HELV_SMALL
       || fontindex == IDF_HELV_MEDIUM
       || fontindex == IDF_HELV_BIG)  {
	basePsFont = "Helvetica %d";
	baseIdFont = "*-helvetica-medium-r-normal-*-%d-*";
    } else {
	basePsFont = "Courier %d";
	baseIdFont = "-*-courier-medium-r-normal-*-%d-*";
    }
    if(fontindex == IDF_HELV_SMALL
       || fontindex == IDF_COUR_SMALL)  {
	size = 9;
    } else if(fontindex == IDF_HELV_MEDIUM 
	      || fontindex == IDF_COUR_MEDIUM)  {
	size = 12;
    } else {
	size = 18;
    }
    sprintf(psfont, basePsFont, size);
    sprintf(idwfont, baseIdFont, size);
}

void idWriteAt(WIND *wind, int x, int y, char *str)
{
    FILE *idf = wind->opf;

    if(idf)  {
	if(psfont==NULL)  {
	    idSetFont(wind, IDF_COUR_MEDIUM);
	}
	fprintf(idf,"Begin %%I Text\n");
	idwFgBg(wind,0);	/* no bg for text */
	fprintf(idf,"%%I f %s\n%s SetF\n", idwfont, psfont);
	fprintf(idf,"%%I t\n[ 1 0 0 1 %d %d ] concat\n%%I\n", 
		PSMINX+TEXTFIXX+x, PSMAXY+TEXTFIXY-y);
	fprintf(idf,"[\n(%s)\n] Text\nEnd\n\n", str);
	fflush(idf);
    }
}

#ifdef MAIN	/* test code below here */

#define ALEN	256
main()
{
    MyByte byte[ALEN];
    int i;
    long len;

    printf("idrawop: testing RLE functions\n");

    for(i = 0; i< ALEN; ++i)  {
	byte[i] = i;
    }

    for(i = 64; i<64+131; ++i) {
	byte[i] = 0xde;
    }

    rleWriteData(stdout, byte, ALEN);
}
#endif /* MAIN */
