/*******************************************************\
*	llists.c					*
*	lisp-like nested loop functions			*
*	dpwe 16jan91					*
\*******************************************************/

#include <dpwelib.h>	/* for stderr msgs only */
#include <genutils.h>	/* for Mymalloc() */
#include <llists.h>
#if (MIPSEL && ultrix)   /* MIPS compiler can't handle ANSI assert */
#ifndef __GNUC__
#undef __STDC__
#endif /* __GNUC__ */
#endif /* (MIPSEL && ultrix) */
#include <assert.h>

NODE *new_node(next, data, type)
    NODE *next;
    NODE *data;
    int type;
    {
    NODE 	*newNode;

    if( (newNode = (NODE *)Mymalloc(1L, sizeof(NODE),
    						"new_node")) != NULL)
	{
	newNode->next = next;
	newNode->data = data;
	newNode->type = type;
	}
#ifdef TRACEALLOC
    fprintf(stderr, "NN(%lx,t%d) ", newNode, type);	fflush(stderr);
#endif /* TRACEALLOC */
    return newNode;
    }

void free_node(node)
    NODE *node;
    {
#ifdef TRACEALLOC
    fprintf(stderr, "FN(%lx,t%d) ", node, node->type);	fflush(stderr);
#endif /* TRACEALLOC */
    if(node != NULL)
	free(node);
    }

void free_list(list)
    LIST list;		/* free all top level nodes in list */
    {
    LIST	next;

    while(list)
	{
	next = cdr(list);
	free_node(list);
	list = next;
	}
    }

void free_any_lists(LIST list, int ntypes, int types[], NodeFunc free_fns[])
{   /* recursively free all nodes from <list>, using the apprpopriate 
       function from <free_fns> as determined by a cartype matching the 
       corresponding element of <types> (there are <ntypes> of them) */
    int 	t, i;
    LIST	next;

    while(list)
	{
	if( (t = cartype(list)) == (int)T_LIST )
	    free_any_lists(car(list), ntypes, types, free_fns);	
	/* above used to be free_list 1992oct28 */
	else{
	    for(i = 0; i<ntypes; ++i)
		if(t == types[i])
		    (*free_fns[i])(car(list), cartype(list), NULL);
	}
	next = cdr(list);
	free_node(list);
	list = next;
    }
}

void free_lists(list)
    LIST list;		/* recursively free all nodes from list */
    {
    LIST	next;
    int 	t;

    while(list)
	{
	if( (t = cartype(list)) == (int)T_LIST )
	    free_list(car(list));
	next = cdr(list);
	free_node(list);
	list = next;
	}
    }

void free_homo_list(LIST list, NodeFunc freeFn)
{	/* <list> is a list of homogeneous nodes.  Each car() will
	   have <freeFn> applied to it (to free it), then will be 
	   freed itself */
    map(list, freeFn);
    free_list(list);
}

NODE *add_after(par, data, type)
    NODE *par;
    NODE *data;
    int type;
    /* add a new node after given list element */
    {
    NODE	*new;

    new = new_node(par->next, data, type);
    par->next = new;		/* insert into list */
    return(new);
    }

NODE *add_before(next, data, type)
    NODE *next;
    NODE *data;
    int type;
    /* insert node as predecessor to one passed */
    {
    NODE	*new;

    new = new_node(next, data, type);
    return(new);
    }			/* identical to new_node?? */

void map(LIST head, NodeFunc func)
{    /* apply node func to each element of list top level */
    while(head)	{
	(*func)(car(head), cartype(head), NULL);
	head = cdr(head);
    }
}

void maparg(LIST head, NodeFunc func, void *arg)
{    /* as map; supply argument to called function */
    while(head)	{
	(*func)(car(head), cartype(head), arg);
	head = cdr(head);
    }
}

LIST lmaparg(LIST head, LNodeFunc func, void *arg)
{   /* apply <func> to each top level item in <head>
       return a new list cons'd up from the returns of <func> */
    LIST rslt = NULL, *rsTail, node;
    int  type;
    
    rsTail = &rslt;
    while(head) {
	type = CARTYPE(head);
	node = (*func)(CAR(head), &type, arg);
	head = CDR(head);
	*rsTail = new_node(NULL, node, type);
	advance_ptr(&rsTail);
    }
    return rslt;
}

LIST lmappend(LIST head, LNodeFunc func, void *arg)
{   /* apply <func> to each top level item in <head>
       return a new list appended-up from the returns of <func> */
    LIST rslt = NULL, *rsTail, node;
    int  type;
    
    rsTail = &rslt;
    while(head) {
	type = CARTYPE(head);
	node = (*func)(CAR(head), &type, arg);
	assert(type == T_LIST);	/* MUST return lists if we're to append them */
	*rsTail = node;
	rsTail = find_last_ptr(rsTail);
	head = CDR(head);
    }
    return rslt;
}

void map_all(LIST head, NodeFunc func)
{   /* apply node func to every NON-T_LIST element of list recursively */
    while(head)	{
	if(cartype(head) == (int)T_LIST)
	    map_all(car(head), func);
	else	/* ? */
	    (*func)(car(head), cartype(head), NULL);
	head = cdr(head);
    }
}

NODE *find_last(head)
    LIST head;
    /* return the last node in a list's top level */
    {
    NODE *last;

    last = head;
    while(last->next)
	last = last->next;
    return(last);
    }

LIST *find_last_ptr(phead)
    LIST *phead;
    /* return the last node in a list's top level */
    {
    LIST *plast;

    plast = phead;
    while(*plast)
	plast = &((*plast)->next);
    return(plast);
    }

LIST find_prec(LIST list, NODE *node)
{ /* returns the node IMMEDIATELY BEFORE node, or NULL if not found */ 
    while(list && cdr(list)!=node) 	
	list = cdr(list); 
    return list; 
}

LIST cons(data, type, rest)
    NODE *data;
    int type;
    LIST rest;
    /* create a new node with the given contents at the front of the list */
    {
    return add_before(rest, data, type);	/* identical */
    }

LIST append(l1, l2)
    LIST l1;
    LIST l2;
    /* stick l2 onto the end of l1 */
    {
    NODE *last;

    if(l1 == NULL)	return l2;
    last = find_last(l1);
    last->next = l2;
    return(l1);
    }

LIST *append_ptr(pl1, l2)
    LIST *pl1;
    LIST l2;
    /* stick l2 onto the end of the list pointed to by pl1 */
    {
    LIST *plast;

    plast  = find_last_ptr(pl1);
    *plast = l2;
    return(plast);
    }

LIST snoc(start, data, type)
    LIST start;
    NODE *data;
    int type;
    /* upside-down cons : creates a new node at the end of a list */
    {
    NODE *last;

    last = find_last(start);
    add_after(last, data, type);
    return(last->next);
    }

LIST break_after(par)
    NODE *par;
    /* separate two segments of a list by cutting after the parent */
    {
    LIST	rest;

    rest = par->next;
    par->next = NULL;
    return(rest);
    }

LIST insert_after(par, new)
    NODE *par;
    LIST new;
    /* splice the whole of a list into a new one */
    {
    LIST l2;

    l2 = break_after(par);
    append(par, new);
    append(new, l2);
    return par;
    }

NODE *extract_next(par)
    NODE *par;
    /* remove & return one element, close up list */
    {
    NODE *node;

    node = par->next;
    par->next = node->next;
    node->next = NULL;
    return node;
    }

void free_next(par)
    NODE *par;		/* remove succeeding member from list */
    {
    NODE	*node;

    node = extract_next(par);
    free_node(node);		/* get rid of it */
    }

NODE *extract_next_ptr(ppar)
    LIST *ppar;
    /* remove & return one element, close up list passed as reference */
    {
    NODE *node;

    node = *ppar;
    *ppar = node->next;
    node->next = NULL;
    return node;
    }

void free_next_ptr(ppar)
    LIST *ppar;		/* remove succeeding member from list as ref */
    {
    NODE	*node;

    node = extract_next_ptr(ppar);
    free_node(node);		/* get rid of it */
    }

NODE *find_el(where, data, type)
    LIST where;
    NODE *data;
    int type;
    /* search for an exact match to data in top level of where */
    {
    while(where && where->data != data)
	where = where->next;
    return(where);	/* NULL if not found */
    }

NODE *find_car(LIST where, NODE *data)
{   /* searches top level items of <where> for CAR == data, 
       as find_el, but no <type> nonsense */
    while(where && where->data != data)
	where = where->next;
    return where;
}

NODE **find_par_ptr(proot, node)
    LIST *proot;
    NODE *node;			/* return ptr to node */
    {
#ifdef DEBUG
    fprintf(stderr, "fpp:");	fflush(stderr);
#endif
    while(proot && *proot != node)
	{
#ifdef DEBUG
	fprintf(stderr, "*%lx=%lx/", proot, *proot); fflush(stderr);
#endif
	if(*proot)
	    proot = cdr_ptr(*proot);
	else
	    proot = NULL;
	}
#ifdef DEBUG
    fprintf(stderr,"\n");	fflush(stderr);
#endif
    return proot;
    }

NODE *car(par)
    LIST par;
    /* return contents of data */
    {
    return(par->data);
    }

int cartype(par)
    LIST par;
    /* return type of data register */
    {
    return(par->type);
    }

NODE *set_car(node, data, type)
    NODE *node;
    NODE *data;
    int type;
    /* rewrite the value of a node */
    {
    node->data = data;
    node->type = type;
    return(node);
    }

LIST cdr(par)
    LIST par;
    /* returns list without first element */
    {
    return(par->next);
    }

NODE **next_ptr(node)
    NODE *node;
    /* return pointer to where to write next element of list */
    {
    return(&(node->next));
    }

NODE **cdr_ptr(node)
    NODE *node;
    /* return pointer to where to write next element of list */
    {
    return(&(node->next));
    }

NODE *set_next(node, next)
    NODE *node;
    LIST next;
    /* just set the next element of a list node */
    {
    node->next = next;
    return node;	/* ? or next ? */
    }

NODE *close_up_ptr(pNodePtr)
    NODE **pNodePtr;
    /* given the address of a ptr to a node, unlink that to which it points */
    {
    NODE	*node;

    node = *pNodePtr;
    *pNodePtr = cdr(node);
    set_next(node,NULL);
    return node;
    }

LIST advance_next(listp)
    NODE ***listp;
    /* advance ptr to next field to next element in list */
    {
    NODE	*pnode;

    pnode = **listp;
    *listp = next_ptr(pnode);
    return cdr(pnode);	/* == *(*listp) */
    }

LIST advance_ptr(listp)
    NODE ***listp;
    /* same as advance_next */
{
    return advance_next(listp);
}

NODE *insert_before(pNodePtr, insertNode)
    NODE **pNodePtr;
    LIST insertNode;
    /* insert a node OR LIST ahead of an address of a next-pointer */
    {
    NODE *next;

    next = *pNodePtr;
    *pNodePtr = insertNode;
    /* had better just be one node ? */
    insertNode = find_last(insertNode);
    set_next(insertNode, next);
    return *pNodePtr;
    }

LIST insert_by_comp(LIST *pListRoot, NODE *node, NodeCmpFn compFn)
{   /* insert a node before the first node which <node> is LESS THAN.
       compFn is invoked (*compFn)(&node-to-insert, &next-node-from-list)
       and insert happens whenever this returns < 0 */
    NODE	**pThisNode;
    float	nodeMetric;
 
    pThisNode = pListRoot;
    while(*pThisNode && (*compFn)(&node, pThisNode) >= 0 )
	advance_next(&pThisNode);
    /* need to insert node before *pThisNode */
    insert_before(pThisNode, node);
    return *pListRoot;
}

typedef int (*qcmpfn)(const void *a, const void *b);
/* type of fn arg to qsort(), to please compiler with coercion */

LIST qsort_list(LIST list, NodeCmpFn compar)
{        /* re-order a list with qsort().  Return head of new list.
	    comparison function must accept two DOUBLY-INDIRECTED 
	    nodes (NODE **pcar1,2) and return -1, 0 or 1 (or similar) */
    NODE **nodelist;
    NODE *ptr;
    long nnodes, i;

    if( (nnodes = top_size(list)) < 2)  return list;
    /* copy top elements of list into a linear array to allow 
       qsorting */
    nodelist = TMMALLOC(NODE *, nnodes, "qsort_list:tmp");
    ptr = list;
    i = 0;
    while(ptr)  {
        nodelist[i++] = car(ptr);
	ptr = cdr(ptr);
    }
    assert(i == nnodes);
    qsort(nodelist, nnodes, sizeof(NODE *), (qcmpfn)compar);
    /* rewrite new contents into the top-level list */
    ptr = list;
    i = 0;
    while(ptr)  {
        set_car(ptr, nodelist[i++], cartype(ptr));
	ptr = cdr(ptr);
    }
    /* release temporary array */
    free(nodelist);
    return list;
}

LIST qsort_list2(LIST list, NodeCmpFn compar)
{   /* as qsort list, except <compar> is passed pointers to nodes 
       in top level (so it can inspect their cartypes), not just 
       the cars.  Fixup then rewrites CDRs of list, not CARs */
    NODE **nodelist;
    NODE *ptr;
    long nnodes, i;

    if( (nnodes = top_size(list)) < 2)  return list;
    /* copy top elements of list into a linear array to allow 
       qsorting */
    nodelist = TMMALLOC(NODE *, nnodes, "qsort_list:tmp");
    ptr = list;
    i = 0;
    while(ptr)  {
        nodelist[i++] = ptr;
	ptr = cdr(ptr);
    }
    assert(i == nnodes);
    qsort(nodelist, nnodes, sizeof(NODE *), (qcmpfn)compar);
    /* rewrite new contents into the top-level list */
    ptr = NULL;
    for(i = nnodes-1; i >=0 ; --i)  {
	set_next(nodelist[i], ptr);
	ptr = nodelist[i];
    }
    /* release temporary array */
    free(nodelist);
    return ptr;
}

LIST drop_first(list)
    LIST list;
    /* free first element of list & return remainder */
    {
    LIST ret;

    ret = cdr(list);
    set_next(list, NULL);
    free_node(list);
    return ret;
    }

LIST copy_top(src)
    LIST src;		/* make new nodes for top level, same cars */
    {
/* recursive version is mind-blowing for mega-lists */
/*    if(src == NULL)
	return NULL;
    else
	return cons(car(src), cartype(src), copy_top(cdr(src)));
 */
    /* icky nonrecursive version */
    LIST *thisNode;
    LIST result = NULL;

    thisNode = &result;
    while(src)
	{
	*thisNode = new_node(NULL, car(src), cartype(src));
	src = cdr(src);
	thisNode = cdr_ptr(*thisNode);
	}
    return result;
    }

int top_size(list)	/* count top-level elements in a list */
    LIST list;
    {
    int  n = 0;

    while(list)
	{
	++n;
	list = cdr(list);
	}
    return n;
    }

void map_with_abort(LIST head, NodeFunc func)
    {    /* apply node func to each element of list top level.
	    Stop immediately if function returns -1 */
    int started = 0, done = 0;

/*    while(head && !started)
	{
	started = ( (*func)(car(head), cartype(head), NULL)  0);
	head = cdr(head);
	}
*/
    while(head && !done)
	{
	done = ( (*func)(car(head), cartype(head), NULL) < 0);
	head = cdr(head);
	}
    }

void maparg_with_abort(LIST head, NodeFunc func, void *arg)
    {    /* apply node func to each element of list top level.
	    Once one func returns non zero, will only keep applying until 0 */
    int done = 0;

    while(head && !done)  {
	done = ( (*func)(car(head), cartype(head), arg) < 0);
	head = cdr(head);
    }
}

int car_match(node, type, car)	/* for SearchList, check for equal car's */
    NODE *node;
    int type;
    void *car;
    {
    return(node==(NODE *)car);
    }

LIST search_by_fn(list, fn, arg)	
    LIST	list;
    int 	(*fn)(LIST, int, void *);	/* test function */
    void	*arg;		/* argument to fn */
    /* apply fn to top level cars of list, return tail with 1st that passes */
    {
    int 	found = 0;

    while(list && found == 0)
	{
	if( (*fn)(car(list), cartype(list), arg)>0 )
	    found = 1;
	else
	    list = cdr(list);
	}
    return list;
    }

LIST search_with_abort(list, fn, arg)	
    LIST	list;
    int 	(*fn)(LIST, int, void *);	/* test function */
    void	*arg;		/* argument to fn */
    /* apply fn to top level cars of list, return tail if fn +, abort if fn- */
    {
    int 	f, found = 0, done = 0;

    while(list && found == 0 && !done)
	{
	if( (f = (*fn)(car(list), cartype(list), arg))>0 )
	    found = 1;
	else if(f<0)
	    done = 1;
	else
	    list = cdr(list);
	}
    if(found == 0)
	list = NULL;
    return list;
    }

typedef struct  {	/* state structure for filter structure */
    LIST	*rslt;	/* passing tracks are added to this list */
    NodeFunc	predFn;	/* this is applied to nodes; if 1, they are included */
    void	*predArg;	/* the argument to the predicate function */
} LFILT_STR;

static int filter_fn(NODE *data, int type, void *arg)
{	/* aux function to be mapped to list to effect filtering */
    LFILT_STR 	*lfs = (LFILT_STR *)arg;	/* recover state */
    int 	rc;

    if((rc = (*lfs->predFn)(data, type, lfs->predArg))>0)  { 
	/* passes predicate */
	*(lfs->rslt) = new_node(NULL, data, type);	/* add result */
	lfs->rslt = cdr_ptr(*lfs->rslt);	/* advance list */
    }
    return rc;	/* rc of one means 'keep applying me'; -1 means 'stop now' */
}

LIST keep_if(LIST list, NodeFunc predFn, void *predArg)
{	/* return a new list of the elements of <list> for which 
	   <predFn> (called with <predArg>) returns > 0  */
    LFILT_STR	lfs;
    LIST	rslt = NULL;

    lfs.rslt = &rslt;
    lfs.predFn = predFn;
    lfs.predArg = predArg;

    maparg(list, filter_fn, (void *)&lfs);
    /* filter_fn writes into <rslt> as held in <lfs>, builds new list */

    return rslt;
}

LIST keep_if_abort(LIST list, NodeFunc predFn, void *predArg)
{	/* return a new list of the elements of <list> for which 
	   <predFn> (called with <predArg>) returns > 0.
	   If if ever returns < 0, stop looking immediately. */
    LFILT_STR	lfs;
    LIST	rslt = NULL;

    lfs.rslt = &rslt;
    lfs.predFn = predFn;
    lfs.predArg = predArg;

    maparg_with_abort(list, filter_fn, (void *)&lfs);
    /* filter_fn writes into <rslt> as held in <lfs>, builds new list */

    return rslt;
}

int inv_node_fn(LIST node, int type, void *arg)
{	/* wrapper to invert node functions - treats arg as InvNFStr */
    InvNFStr	*infs = (InvNFStr *)arg;

    return !((*(infs->fn))(node, type, infs->arg) > 0);
}

LIST remove_if(LIST list, NodeFunc predFn, void *predArg)
{	/* return a new list of the elements of <list> for which 
	   <predFn> (called with <predArg>) returns FALSE */
    InvNFStr	infs;

    infs.fn  = predFn;
    infs.arg = predArg;
    return keep_if(list, inv_node_fn, &infs);
}

int equalp_nf(NODE *node, int type, void *arg)
{	/* return 1 if cars match, else 0 */
    NODE *with = (NODE *)arg;
    if(node == with)
	return 1;
    else
	return 0;
}

int member_p(NODE *node, int type, void *arg)
{	/* node function returns true if <node> is the same as 
	   a CAR of one of <list>'s elements */
    NODE *list = (NODE *)arg;
    LIST within;

    within = keep_if(list, equalp_nf, node);
    if(within == NULL)  
	return 0;
    else {
	free_list(within);
	return 1;
    }
}

int nonmember_p(NODE *node, int type, void *arg)
{	/* node function returns true if <node> is NOT in list */
    return !member_p(node, type, arg);
}

LIST remove_list(LIST list, LIST withouts)
{	/* remove all car's matching those in <withouts> from <list> */
    LIST rslt = keep_if(list, nonmember_p, withouts);
    return rslt;
}

LIST keep_list(LIST list, LIST withs)
{	/* remove all car's from <list> that *don't* occur in <withs> */
    LIST rslt = keep_if(list, member_p, withs);
    return rslt;
}

long list_index_ptr(LIST *plist, void *data) 
{	/* Find the node in <*plist> whose CAR matches <data>.
	   If found, return its position in the list, and modify 
	   <*plist> to point to it.   If not found, leave <*plist>
	   at last element of list, but return -1.
	   First element of list has index *0* (zero).  */
    long 	rslt = 0;
    int 	found = 0;

    while(*plist && !found)  {
	if( ((void *)CAR(*plist)) != data)  {
	    *plist = CDR(*plist);
	    ++rslt;
	}
	else
	    found = 1;
    }
    if(!found) 	return -1;	/* *plist left at end */
    else	return rslt;	/* *plist points to found */
}

long list_index(LIST list, void *data)
{	/* return the index of the top-level item in <list> whose CAR 
	   is identical to <data>, or (-1) if not found. 
	   First element of list has index *0* (zero). */
    LIST root = list;

    return list_index_ptr(&root, data);
}

LIST nth(LIST list, long n)
{	/* return result of applying CDR <n> times to <list> */
    while(list && n--)  {
	list = CDR(list);
    }
    return list;	/* if n >= top_size(list), return NULL */
}

int nth_ptr(LIST *plist, long n)
{	/* attempt to step plist over n elements;
	   return 0 on success, or -1 if run off end 
	   (i.e. n >= top_size(list) ) */
    while(*plist && n > 0)  {
	*plist = cdr(*plist);
	--n;
    }
    if(n > 0)
	return -1;	/* not long enough */
    else
	return 0;
}

LIST list_sub(LIST l1, LIST l2)
{	/* returns a list of the CARs of l1 that are not in l2 */
    return remove_list(l1, l2);		/* (allocates a new list) */
}

LIST list_diff(LIST l1, LIST l2)
{	/* returns a list of the elements from l1 and l2 that do 
	   not occur in both.  Order of elements in return is arbitrary. */
    LIST l3 = list_sub(l1, l2);		/* items in l1 not in l2 */
    LIST l4 = list_sub(l2, l1);		/* items in l2 not in l1 */
    /* there cannot be any elements in both l3 and l4.  To get into 
       l3, an element must be NOT in l2;  but all elements in l4 originated 
       from l2.  Therefore, there is no issue of removing duplicates */
    return append(l3, l4);		/* no longer ordered well */
}

LIST list_merge(LIST l1, LIST l2)
{	/* returns a new list which is the union of l1 and l2 with 
	   diplicates removed */
    LIST l2minusl1 = list_sub(l2, l1);
    LIST merge = copy_top(l2);	/* make a copy */

    merge = append(merge, l2minusl1);
    return merge;
}

int is_subset(LIST little, int type, LIST bigger)
{	/* NodeFunc tests the first argument as a subset of the arg */
    LIST diff;

    assert(type == T_LIST);
    /* find all elements of little not in bigger */
    diff = list_sub(little, bigger);
    if(diff == NULL)
	return 1;	/* IS a subset */
    else {
	free_list(diff);
	return 0;	/* difference is nonzero - NOT a subset */
    }
}

#ifdef MAIN /* TEST CODE hereafter ------------------------------- */

int cmpStrs(NODE **ppn1, NODE **ppn2)
{    /* comparison function for a list of strings */
    char a, b;    /* just compare on first char */

    /* *ppn1 is the car of an element of the list we are searching.
       In general, that would be the address of another list;  in 
       this case, we know them to be pointers to C strings */

    a = **(char **)ppn1;
    b = **(char **)ppn2;

    if( a == b )    return 0;
    else if (a > b) return 1;
    else return -1;
}

main()  {
    LIST list = NULL;
    NODE *ptr;

    printf("Testing qsort_list ... \n");
  
    list = new_node(list, (LIST)"This", T_STR);
    list = new_node(list, (LIST)"is", T_STR);
    list = new_node(list, (LIST)"a", T_STR);
    list = new_node(list, (LIST)"test", T_STR);
    list = new_node(list, (LIST)"of", T_STR);
    list = new_node(list, (LIST)"qsort", T_STR);
    list = new_node(list, (LIST)"_list", T_STR);

    list = qsort_list(list, cmpStrs);
    ptr = list;
    while(ptr)  {
        printf("out:'%s'\n", (char *)car(ptr));
	ptr = cdr(ptr);
    }
}
    
#endif /* MAIN */
