/*******************************************************\
*	llists.h					*
*	lisp-like nested loop functions			*
*	dpwe 16jan91					*
\*******************************************************/

#ifndef _LLISTS_H_
#define _LLISTS_H_

/* misc */
#ifndef NULL
#define NULL 0L
#endif

/* data types */
typedef struct node {
	struct node *next;
	struct node *data;
	int    type;		/* flag type of data - node or other data */
} NODE, *LIST;
/* IF YOU CHANGE THIS, be sure to update conformal PKNODE data structure 
   in 'peaks.h' if you care about the peak/track forming code still working
   1992oct27 dpwe */

/* symbols */
enum {
     T_LIST,
     T_CHAR,
     T_STR,
     T_INT,
     T_ALLOC,			/* i.e. free(data) on destroying the node */
     T_0,
     T_1,
     T_2,
     T_3,			/* "user-defined" */
     T_OTHER
     };

/* macros */
#define CAR(list)	(list)->data
#define CDR(list)	(list)->next
#define CARTYPE(list)	(list)->type
#define CDR_PTR(list)	&((list)->next)

/* define generic function of a type applied to one node */
typedef int (*NodeFunc)(NODE *node, int type, void *data);
/* <type> is the cartype from the node;  data is (optionally) 
   data passed in by calling fn */

/* as NodeFunc, but returns a list e.g. for lmaparg(), lmappend() */
typedef LIST (*LNodeFunc)(NODE *node, int *ptype, void *data);
/* ptype passes *in* type, returns *back* type of result node */

/* define generic comparison function.  Arguments are both 
   addresses of pointers to <car>s, to be compatible with qsort function */
typedef int (*NodeCmpFn)(NODE **pcar1, NODE **pcar2);

typedef struct  {	/* function to hold args for inv_node_fn */
    NodeFunc	fn;	/* the original node function which is to be inv'd */
    void 	*arg;	/* args to original node function */
} InvNFStr;

/* function prototypes */

NODE *new_node(NODE *next, NODE *data, int type);
void free_node(NODE *node);
void free_list(LIST list);	/* free all top level nodes in list */
void free_lists(LIST list);	/* recursively free all nodes from list */
void free_any_lists(LIST list, int ntypes, int types[], NodeFunc free_fns[]);
    /* recursively free nodes, apply free_fns to cartypes in types[] */
void free_homo_list(LIST list, NodeFunc freeFn);
	/* <list> is a list of homogeneous nodes.  Each car() will
	   have <freeFn> applied to it (to free it), then will be 
	   freed itself */
NODE *add_after(NODE *par, NODE *data, int type);
    /* add a new node after given list element */
NODE *add_before(NODE *next, NODE *data, int type);
    /* insert node as predecessor to one passed */
void map(NODE *head, NodeFunc func);
    /* apply func(data, type) to car of each element of list top level */
void maparg(NODE *head, NodeFunc func, void *arg);
    /* as map, except <arg> is passed as third arg to <func> */
LIST lmaparg(LIST head, LNodeFunc func, void *arg);
    /* apply <func> to each top level item in <head>
       return a new list cons'd up from the returns of <func> */
LIST lmappend(LIST head, LNodeFunc func, void *arg);
    /* apply <func> to each top level item in <head>
       return a new list appended-up from the returns of <func> */
void map_with_abort(LIST head, NodeFunc func);	/* map until fn returns -1 */
void maparg_with_abort(NODE *head, NodeFunc func, void *arg);
    /* as map_with_abort, except <arg> is passed as third arg to <func> */
void map_all(LIST head, NodeFunc func);
    /* apply node func to every NON-T_LIST element of list recursively */
NODE *find_last(LIST head);
    /* return the last node in a list's top level */
LIST *find_last_ptr(LIST *phead);
    /* return address of last node pointer in a list's top level */
LIST find_prec(LIST list, NODE *node);
    /* returns the node IMMEDIATELY BEFORE node, or NULL if not found */ 
LIST cons(NODE *data, int type, LIST rest);
    /* create a new node with the given contents at the front of the list */
LIST append(LIST l1, LIST l2);
    /* stick l2 onto the end of l1 */
LIST *append_ptr(LIST *pl1, LIST l2);
    /* stick l2 onto end of list chain pointed to by pl1. Return new pl2 */
LIST snoc(LIST start, NODE *data, int type);
    /* upside-down cons : creates a new node at the end of a list */
LIST break_after(NODE *par);
    /* separate two segments of a list by cutting after the parent */
LIST insert_after(NODE *par, LIST new);
    /* splice the whole of a list into a new one */
NODE *extract_next(NODE *par);
    /* remove & return one element, close up list */
void free_next(NODE *par);	/* remove succeeding member from list */
NODE *extract_next_ptr(LIST *ppar);
    /* remove & return one element, close up list passed as reference */
void free_next_ptr(LIST *ppar);	/* remove succeeding member from list as ref */
NODE *find_el(LIST where, NODE *data, int type);
    /* search for an exact match to data in top level of where */
NODE *find_car(LIST where, NODE *data);
    /* searches top level items of <where> for CAR == data, 
       as find_el, but no <type> nonsense */
NODE **find_par_ptr(LIST *proot, NODE *node);	/* return ptr to node */
NODE *car(LIST par);
    /* return contents of data register */
int cartype(LIST par);
    /* return type of data register */
LIST cdr(LIST par);
    /* returns list without first element */
NODE **next_ptr(NODE *node); /* return ptr to where to write next el of list */
NODE **cdr_ptr(NODE *node);  /* return ptr to where to write next el of list */
NODE *set_car(NODE *node, NODE *data, int type); /* rewrite value of a node */
NODE *set_next(NODE *node, LIST next);  /* just set the next el of list node */
NODE *close_up_ptr(NODE **pNodePtr);
    /* given the address of a ptr to a node, unlink that to which it points */
LIST advance_next(LIST **listp);    /* ? advance to next element in list */
LIST advance_ptr(LIST **listp);     /* another name for advance_next */
NODE *insert_before(NODE **pNodePtr, LIST insertNode);
    /* insert a node OR LIST ahead of an address of a next-pointer */
LIST insert_by_comp(LIST *pListRoot, NODE *node, NodeCmpFn compFn);
    /* insert node before first node against which compFn compares it 
       as `-1'.  compFn called as compFn(&node, &(*pListRoot)->car); etc */
LIST qsort_list(LIST list, NodeCmpFn compar);
    /* re-order a list with qsort().  Return head of new list.
       comparison function must accept two DOUBLY-INDIRECTED 
       nodes (NODE **pcar1,2) and return -1, 0 or 1 (or similar) 
       rewrites CARs of top level */
LIST qsort_list2(LIST list, NodeCmpFn compar);
    /* sort a list via qsort;  compar returns 1/0/-1 for rel of 2 NODES
       i.e. pointers to elements of toplevel, then rewrites NEXTs*/
LIST drop_first(LIST list);
    /* free first element of list & return remainder */
LIST copy_top(LIST src);	/* make new nodes for top level, same cars */
int top_size(LIST list);	/* count top-level elements in a list */
int car_match(NODE *node, int cartype, void *car);  /* check for equal car's */
LIST search_by_fn(LIST list, NodeFunc fn, void *arg);
    /* apply fn to top level cars of list, return tail with 1st that passes */
LIST search_with_abort(LIST list, NodeFunc fn, void *arg);
    /* apply fn to top level cars, return tail that returns +, abort if fn - */
LIST remove_if(LIST list, NodeFunc predFn, void *predArg);
    /* return a new list of the elements of <list> for which 
       <predFn> (called with car, cartype, <predArg>) returns <= 0 */
LIST keep_if(LIST list, NodeFunc predFn, void *predArg);
    /* return a new list of the elements of <list> for which 
       <predFn> (called with car, cartype, <predArg>) returns > 0 */
int member_p(NODE *node, int type, void *arg);
   /* node function returns true if <node> is the same as 
int nonmember_p(NODE *node, int type, void *arg);
   /* node function returns true if <node> is NOT in list */
LIST remove_list(LIST list, LIST withouts);
   /* remove all car's matching those in <withouts> from <list> */
LIST keep_list(LIST list, LIST withs);
    /* remove all car's from <list> that *don't* occur in <withs> */
int inv_node_fn(LIST node, int type, void *arg);
    /* wrapper to invert node functions - treats arg as InvNFStr */
int equalp_nf(NODE *node, int type, void *arg);
    /* NodeFunc returns 1 if cars match, else 0 */
long list_index_ptr(LIST *plist, void *data);
    /* Find the node in <*plist> whose CAR matches <data>.
       If found, return its position in the list, and modify 
       <*plist> to point to it.   If not found, leave <*plist>
       at last element of list, but return -1.
       First element of list has index *0* (zero).  */
long list_index(LIST list, void *data);
    /* return the index of the top-level item in <list> whose CAR 
       is identical to <data>, or (-1) if not found. 
       First element of list has index *0* (zero). */
LIST nth(LIST list, long n);
    /* return result of applying CDR <n> times to <list> */
int nth_ptr(LIST *plist, long n);
    /* attempt to step plist over n elements; 
       return 0 on success, or -1 if run off end 
       (i.e. n >= top_size(list) ) */
LIST list_sub(LIST l1, LIST l2);
    /* returns a list of the CARs of l1 that are not in l2 */
LIST list_diff(LIST l1, LIST l2);
    /* returns a list of the elements from l1 and l2 that do 
       not occur in both.  Order of elements in return is arbitrary. */
LIST list_merge(LIST l1, LIST l2);
    /* returns a new list which is the union of l1 and l2 with 
       duplicates removed */
int is_subset(LIST little, int type, LIST bigger);
    /* NodeFunc tests the first argument as a subset of the arg */


#endif /* _LLISTS_H_ */

