/*******************************************************************\
*	playmac.c														*
*	real-time sound output via mac sound manager (including stereo)	*
*	compatible with 'abstract' play layer							*
*	dpwe 1993mar06  after xmpeg.c									*
\*******************************************************************/

#include <dpwelib.h>
#include <play.h>
#include <IEEE80.h>		/* for double_to_ieee_80(), for extSH->sampRate */
#include <assert.h>
/* Mac include .. */
#include <Sound.h>
#define rate44khz ((unsigned long)(44100.0*65536))

#define PLAYBFRSIZE		4096	/* sample frames per sound manager call */

#define NPLAYBUFS		8		/* total # of buffers we allocate (slack) */
#define STARTUPBUFS		6		/* fill up this many before sending 1st */

#define TIMEOUTLIM 1000

#define STEREO					/* try and play stereo if sound is 2 chan */
#define USEEXTSH				/* use the Extended SH (necess for stereo) */

#ifdef USEEXTSH
#define SOUNDHEADER ExtSoundHeader
#else /* !USEEXTSH i.e. use stdSH */
#undef STEREO   /* cannot have stereo with stdSH */
#define SOUNDHEADER SoundHeader
#endif /* USEEXTSH */

typedef unsigned char uchar;
static int playBufIx = 0;      /* index of buffer currently setting up */
static int framesThisBuf = 0;  /* frames so far put into current buffer */
static int framesPerBuf = 0;   /* allocated size of every buffer */
static int sentBufIx = -1;     /* index of buffer last sent to sound mgr */
static int doneBufIx = -1;     /* index of buffer last completed by snd mgr */
static int sendStarted = 0;    /* if we haven't send *any* blocks yet, wait */
static int origChans = 1;	   /* for mono, how many chans in original? */
static SOUNDHEADER   *playBufs[NPLAYBUFS];      /* all our buffers */
static SndChannelPtr schn;
static SndCommand    scmd;

static void playCallback(void)
{	/* function called back by sound manager indicates buffer free */
    ++doneBufIx;
}

int play_set(int chans, int dsize, double srate, int scale)
{               /* setup the sound manager */
    long            frames = PLAYBFRSIZE;
    long            i;
    SOUNDHEADER     *tmpH;
    OSErr           oerr;
    unsigned long   fxSr;
    
	playBufIx = 0;      /* index of buffer currently setting up */
	framesThisBuf = 0;  /* frames so far put into current buffer */
	sentBufIx = -1;     /* index of buffer last sent to sound mgr */
	doneBufIx = -1;     /* index of buffer last completed by snd mgr */
	sendStarted = 0;    /* if we haven't send *any* blocks yet, wait */
	origChans = chans;
#ifndef STEREO
    chans = 1;              /* force mono - only way it works for now */
#endif /* !STEREO */
	
    /* setup sound output hardware */
    schn = NULL;
    if(chans == 1)  
        oerr = SndNewChannel(&schn, sampledSynth, initMono, 
                             (SndCallBackProcPtr)playCallback);     
    else /* chans == 2 */
        oerr = SndNewChannel(&schn, sampledSynth, initStereo, 
                             (SndCallBackProcPtr)playCallback);     
	if(oerr)  {
		return -1;		/* fail */
	}

    /* allocate all the play buffers */
    framesPerBuf = frames;
    for(i = 0; i < NPLAYBUFS; ++i) {
        playBufs[i] = (SOUNDHEADER *)NewPtr(sizeof(SOUNDHEADER)
                                            + chans * frames * sizeof(uchar));
        if(playBufs[i] == NULL)  {
            D_prints("Mem alloc fail in playInit\n");
            return 0;
        }
    }
    fxSr = (unsigned long)(srate * (1L<<16));
    
    /* 'pull in' to hardware supported sample rates to avoid poor 
       interpolation (no 'frac' continuity across blocks) */
    if(42000 < srate && srate < 46000)       fxSr = rate44khz;
    else if(21000 < srate && srate < 23000)  fxSr = rate22khz;
    else if(10500 < srate && srate < 11500)  fxSr = rate11khz;  

    tmpH = playBufs[0];
    tmpH->samplePtr   = tmpH->sampleArea;
    tmpH->sampleRate  = fxSr;
    tmpH->loopStart   = 0L;   
    tmpH->loopEnd     = 0L;
    tmpH->baseFrequency = 60; /* encode & middle C */
#ifdef USEEXTSH
    tmpH->encode      = extSH;      /* was stdSH */
    tmpH->numChannels = chans;      /* */
    tmpH->numFrames   = frames;
    double_to_ieee_80(srate, (uchar *)&(tmpH->AIFFSampleRate));
    tmpH->markerChunk = NULL;
    tmpH->instrumentChunks = NULL;
    tmpH->AESRecording = NULL;
    tmpH->sampleSize   = 8;
#else /* !USEEXTSH - stdSH instead */
    tmpH->encode      = stdSH;
    tmpH->length      = chans * frames * sizeof(uchar);
#endif /* USEEXTSH */
        
    for(i = 1; i < NPLAYBUFS; ++i) {
        *(playBufs[i]) = *tmpH;         /* copy config to other bufs */
        (playBufs[i])->samplePtr = (playBufs[i])->sampleArea;
    }
    return frames;
}

static void playSendBuf(int idx)
{       /* actually issue Snd commands for indexed buffer */
    scmd.cmd    = bufferCmd;
    scmd.param1 = 0;
    scmd.param2 = (long)playBufs[idx];
    SndDoCommand(schn, &scmd, FALSE);
    scmd.cmd    = callBackCmd;           /* so that buffer is released */
    SndDoCommand(schn, &scmd, FALSE);
    ++sentBufIx;
}

void play_rls(void)
{       /* unsetup everything */
    long i;
    int  timeout;

	if(framesThisBuf)  {	/* part-finished buffer : send it */
#ifdef USEEXTSH
		(playBufs[playBufIx % NPLAYBUFS])->numFrames = framesThisBuf;
#else /* !USEEXTSH */
		(playBufs[playBufIx % NPLAYBUFS])->length = framesThisBuf * sizeof(uchar);
#endif /* USEEXTSH */
	}	/* playBufIx already reflects this frame */
		
    while(sentBufIx < playBufIx)  {
        playSendBuf( (sentBufIx+1) % NPLAYBUFS);
    }
    scmd.cmd = nullCmd;
    timeout = 0;
    while(doneBufIx < playBufIx && ++timeout < TIMEOUTLIM)  {
        SndDoCommand(schn,&scmd,FALSE);
    }
    SndDisposeChannel(schn, TRUE); /* waits for commands to end */
    for(i = 0; i < NPLAYBUFS; ++i)  {
        DisposPtr((Ptr)playBufs[i]);
    }
}

void play_on(short *sptr, long frames)
{       /* actually play some samples .. must agree with init */
    uchar	*uptr;
    long	i, samples, rem, todo;
    int		pbIdx;
    int 	chans;

	while(frames)  {
	    /* On entry, playBufIx is the index of the buffer we are
	       currently filling.  However, if it looks full, we already sent
	       it, but we don't step on until we have frames to put in the 
	       next buffer.
	       We check that against the number that have completed;  if 
	       there are NPLAYBUFS `in progress', we block until 
	       another completes, thereby freeing one of our buffers */
	    if(framesThisBuf == framesPerBuf) {  /* this buf full: want another */
	        ++playBufIx;
	        framesThisBuf = 0;
	        scmd.cmd = nullCmd;
	        while((playBufIx - doneBufIx) > NPLAYBUFS)  {
	            SndDoCommand(schn, &scmd, FALSE); /* null cmd allows callbk */
	        }
	    }
	
		/* find out which buffer we are filling and # remaining spaces in it */
	    pbIdx = (playBufIx) % NPLAYBUFS;
	    rem = framesPerBuf - framesThisBuf;
#ifdef USEEXTSH
	    chans = (playBufs[pbIdx])->numChannels;
#else  /* !USEEXTSH => stdSH => mono */
		chans = 1;
#endif /* USEEXTSH */
	    
	    /* convert 16 bit signed to 8 bit unsigned */
		todo = MIN(frames, rem);
#ifdef STEREO   /* implies USEEXTSH */
	    uptr = (uchar *)((playBufs[pbIdx])->sampleArea)
	                       + framesThisBuf*(origChans*sizeof(uchar));
	    samples = todo*origChans;
	    for(i = 0; i<samples; ++i)  {  
	        *uptr++ = 0x80 + ((*sptr)>>8);
	        sptr += 1;
	    }
#else /* !STEREO */
	    uptr = (uchar *)((playBufs[pbIdx])->sampleArea)
	                      + framesThisBuf* 1 /* mono */ *sizeof(uchar);
	    for(i = 0; i<todo; ++i)  {
	        *uptr++ = 0x80 + ((*sptr)>>8);
	        sptr += origChans;
	    }
#endif /* STEREO */
		framesThisBuf += todo;
		frames -= todo;
	
		if(framesThisBuf == framesPerBuf)  {	/* just FINISHED a block to play */
			/* record the number of frames in this buf */
#ifdef USEEXTSH
		    (playBufs[pbIdx])->numFrames = framesThisBuf;
#else /* !USEEXTSH */
	   		(playBufs[pbIdx])->length = framesThisBuf * sizeof(uchar);
#endif /* USEEXTSH */
	        if(sendStarted)  {              /* we are sending frames .. */
	            playSendBuf(pbIdx);
	        } else {
	            if(playBufIx == STARTUPBUFS - 1)  {  /* compltd this many frames */
	                for(i = 0; i <= pbIdx; ++i)  {  /* send all the bufs at 1ce */
	                    playSendBuf(i);
	                }
	                sendStarted = 1;
	            }       /* else not started - just accumulate frames */
	        }
	    }
    }
}

long play_bufremains(void)
{		/* return number of unplayed sample frames */
	return (playBufIx - doneBufIx)*framesPerBuf + framesThisBuf;
}

	

