/***********************************************************************\
*	readaft.c							*
*  functions to extract useful values from aft files (STFT anal)	*
*  dpwe 02sep91 after readaqt.c, 11mar91				*
\***********************************************************************/

#include <dpwelib.h>	/* for FILE, stderr */
#include <genutils.h>	/* for Mymalloc() */
#include <ctype.h>
#include <math.h>
#include <readaqt.h>
#include <readaft.h>
#include <samplops.h>   /* e.g. CopyFloats */
#ifdef DEBUGGRAPHS
#include <window.h>	/* for test code only */
#endif
#include <arrayops.h>	/* ditto   */

/************** switches *****************/
#define SINC 1
#define GAUSS 0
#define DDIFF -1
#define INTERPTYPE SINC
/* if INTERPTYPE == 1, then SpecOfAFFile uses sinc interpolator, else gauss */

/* stat fn protos */
#ifdef __STDC__
static long ReadAFTFrames(AFTINFO *afi, cpxpolar *dstBase, long dstFrm, 
		long rdFrame, long framesToRead);
static void CopyAFTFrames(cpxpolar *srcBase, long srcFrm, cpxpolar *dstBase, 
		       long dstFrm, long nFrms, int dir, int frCpxs);
static int ReadAFTFileFrame(AFTINFO *afi, long frame, cpxpolar *dest);
#else
static long ReadAFTFrames();
static void CopyAFTFrames();
static int ReadAFTFileFrame();
#endif

AFTINFO *NewAFTInfo(desc)
    char *desc;
    /* allocate a new struct, copy in label */
    {
    AFTINFO *rslt;

    rslt = (AFTINFO *)Mymalloc(1L, sizeof(AFTINFO), "NewAFTi");
    if(rslt == NULL)
	{  fprintf(stderr, "NewAFTInfo: malloc fail\n");
	   abort();	}
    /* maybe fill in with defaults? */
    rslt->totalFrames = PV_UNK_LEN;	rslt->duration = 0.0;
    rslt->flags = 0;	rslt->pvh = NULL;	rslt->pvFile = NULL;
    if(desc != NULL)
	strncpy(rslt->desc, desc, DESCLEN);
    else
	rslt->desc[0] = '\0';
    return rslt;
    }

void FreeAFTInfo(afi)
    AFTINFO *afi;
    /* destroy AFT & all it stands for */
    {
    if(afi->pvFile != NULL)
	fclose(afi->pvFile);
    if(afi->pvh != NULL)
	PVFree(afi->pvh);
    free(afi);
    }

FILE *OpenAFTPVfile(path, pafi)
    char *path;
    AFTINFO **pafi;
    /* attempt to open the file; check it's type; allocate & fill new
       headers if it works */
    {
    PVSTRUCT	*pvh;
    FILE	*file;

    if( (file = PVOpenAllocRdHdr(path, &pvh))==NULL )
	{
	/* PVDie(PVE_NOPEN, path);	*/
	return NULL;
	}
    *pafi = DecodeAFTPVH(pvh, file, "noname");
    if(*pafi == NULL)	/* e.g. not right type of file */
	{
	fclose(file);
	PVFree(pvh);
	return NULL;
	}
    else
	return file;
    }

void CloseAFTPVfile(afi)
    AFTINFO *afi;
    {
    FreeAFTInfo(afi);	/* also closes file & frees PVH */
    }

AFTINFO *DecodeAFTPVH(pvh, file, desc)
    PVSTRUCT *pvh;
    FILE *file;
    char *desc;
    /* allocate & fill a new AFTINFO struct from the passed PVSTRUCT */
    /* doesn't access <file>, just puts it in the struct */
    {
    AFTINFO	*rslt;
    char	*info;
    float	tmpflt;
    long	fbSize;
    int 	chans;

    if( pvh->dataFormat != PVFLOAT || pvh->frameFormat != (PVPOLAR) )
	{ fprintf(stderr, "DecodeAFT: PV file is not float,polar(%ld,%ld)\n",
		  pvh->dataFormat, pvh->frameFormat);
	  return NULL;	}
    chans = pvh->channels;
    if(chans != 1)
	fprintf(stderr,"DecodeAFTPV: warning: chans = %d != 1\n", chans);
    /* haven't even thought about how to deal with such a situation */
    rslt = NewAFTInfo(desc);
    info = pvh->info;
    fbSize = pvh->frameBsize;
    if(pvh->dataBsize != PV_UNK_LEN)
	{
	rslt->totalFrames = (pvh->dataBsize/fbSize/* /chans */);
	rslt->duration = (rslt->totalFrames)*pvh->frameIncr /
				pvh->samplingRate;
	/* each frame counts for frameIncr samples */
	}
    else{
	rslt->totalFrames = PV_UNK_LEN;
	rslt->duration = 0.0;
	}
    rslt->pvh    = pvh;
    rslt->pvFile = file;
    return rslt;
    }

static long ReadAFTFrames(afi, dstBase, dstFrm, rdFrame, framesToRead)
    AFTINFO *afi;
    cpxpolar *dstBase;
    long dstFrm; 
    long rdFrame;
    long framesToRead;
    {
    cpxpolar *rdto;
    long     frameChars, frameCpxs;

    frameChars = afi->pvh->frameBsize;
    frameCpxs  = frameChars/sizeof(cpxpolar);
    rdto = dstBase + dstFrm*frameCpxs;
    fseek(afi->pvFile, afi->pvh->headBsize + frameChars*rdFrame, SEEK_SET);
    return fread(rdto, frameChars, framesToRead, afi->pvFile);
    }

static void CopyAFTFrames(srcBase, srcFrm, dstBase, dstFrm, nFrms, dir, cxSize)
    cpxpolar *srcBase;
    long srcFrm;
    cpxpolar *dstBase;
    long dstFrm;
    long nFrms;
    int dir;
    int cxSize;
    {
    long flPerFrm = cxSize*sizeof(cpxpolar) / sizeof(float);
    CopyFloats(((float *)srcBase) + flPerFrm*srcFrm, 1, 
	       ((float *)dstBase) + flPerFrm*dstFrm, 1, flPerFrm*nFrms, dir);
    }

#define CACHEFRAMES	64	/* how many frames to cache (maybe) */
    /* CACHEFRAMES too small -> disk thrash on interp. Too big->slow shifts */
#define FRAMESKIP	8	/* USED : how far to hop the cache fwd */

static int ReadAFTFileFrame(afi, frame, dest)
    AFTINFO *afi;
    long frame;
    cpxpolar *dest;
    /* read a frame of coefs from an aft file.  May do caching */
    {
    static FILE 	*cacheFile = NULL;
    static cpxpolar 	*cacheBuf  = NULL;
    static long		cacheBufSize = 0L;
    static long 	baseFrame  = -1;	/* cache invalid */
    static long	 	nFrames    = 0;
    static int		cacheAtEof = 0;
    int frCpxs = afi->pvh->frameBsize / sizeof(cpxpolar);

    if(cacheBuf == NULL || frame < 0 || afi->pvFile != cacheFile)
	{	/* negative frame resets the cache */
    	baseFrame = -1;
	nFrames = 0;	cacheAtEof = 0;
	}
    if(cacheBuf == NULL)
	{
	cacheBufSize = frCpxs*(long)CACHEFRAMES;
	cacheBuf = (cpxpolar *)Mymalloc(cacheBufSize, sizeof(cpxpolar), 
					"ReadAFT:cacheBuf");
	if(cacheBuf == NULL)	
	    {  fprintf(stderr, 
		       "ReadAFTFileFrame: cannot alloc buffer of %ld cpx\n",
		       cacheBufSize);
	       abort(); }
	}
    if(frame < 0)	return 0;		/* reset cache : all done */
    if(afi->pvFile == cacheFile && baseFrame >= 0)	/* cache is valid */
	{
	long cacheFrames, newBase, skipFrames;

	if(frame>=baseFrame)
	    {
	    if(frame < baseFrame+nFrames)
		{	/* desired frame is in cache */
		CopyAFTFrames(cacheBuf,frame-baseFrame,dest,0,1,FWD,frCpxs);
		return 1;
		}
	    else if(cacheAtEof)	/* attempt to read past EOF */
		{
		ClearFloats((float *)dest,1,2*frCpxs);
		return 0;
		}
	    }
	/* here if ( frame < baseFrame ) or ( frame >= baseFrame + nFrames ) */
	/* i.e. cache miss */
	cacheFrames = cacheBufSize / frCpxs;
	if(frame < baseFrame)
	    {
	    newBase = FRAMESKIP*(int)(frame/FRAMESKIP);
	    /* always in steps of FRAMESKIP */
	    skipFrames = baseFrame - newBase;
	    if(skipFrames < cacheFrames /* nFrames */ ) /* some overlap */
		{
		fprintf(stderr, "CACHE: prepend %ld, start %ld, len %ld\n",
			skipFrames, newBase, nFrames);	/* */
		CopyAFTFrames(cacheBuf, 0, cacheBuf, skipFrames,
			      MIN(cacheFrames - skipFrames, nFrames), 
			      FWD, frCpxs);
		ReadAFTFrames(afi, cacheBuf, 0, newBase, skipFrames);
		baseFrame = newBase;
		if(nFrames + skipFrames < cacheFrames)
		    nFrames += skipFrames;
		else{
		    nFrames = cacheFrames;
		    cacheAtEof = 0;	/* definitely moved back from EOF */
		    }
		return ReadAFTFileFrame(afi, frame, dest);
		}
	    else{	/* fill whole cache */
		ReadAFTFileFrame(afi, -1, NULL);	/* reset cache */
		return ReadAFTFileFrame(afi, frame, dest);
		}
	    }
	else{	/* frame past last in buffer, not at EOF */
	    newBase = FRAMESKIP*
		(int)((frame - cacheFrames + FRAMESKIP)/FRAMESKIP);
	    /* frame will be last, modulo alignment */
	    if(newBase<0)	newBase = 0;	/* "impossible" */
	    skipFrames = newBase - baseFrame;
	    if(skipFrames < nFrames)
		{
		fprintf(stderr, "CACHE: append %ld, start %ld, len %ld\n",
			skipFrames, newBase, nFrames);	/* */
		CopyAFTFrames(cacheBuf, skipFrames, cacheBuf, 0,
			      nFrames-skipFrames, REV, frCpxs);
		nFrames -= skipFrames;
		nFrames += ReadAFTFrames(afi, cacheBuf, nFrames,
					 newBase+nFrames, cacheFrames-nFrames);
		cacheAtEof = (nFrames < cacheFrames);
		baseFrame = newBase;
		return ReadAFTFileFrame(afi, frame, dest);
		}
	    else{ 	/* fill whole cache */
		ReadAFTFileFrame(afi, -1, NULL);	/* reset cache */
		return ReadAFTFileFrame(afi, frame, dest);
		}
	    }
	}
    else{	/* read whole new cache */
	long cacheFrames = cacheBufSize / frCpxs;
	long newBase = FRAMESKIP*(int)(frame/FRAMESKIP);
/*	long newBase = frame;	*/

	cacheFile = afi->pvFile;
	nFrames = ReadAFTFrames(afi, cacheBuf, 0, newBase, cacheFrames);
	cacheAtEof = (nFrames < cacheFrames);
	baseFrame = newBase;
	fprintf(stderr,"CACHE: reread %ld from %ld : read %ld, EOF = %d\n", 
		cacheFrames,newBase,nFrames,cacheAtEof);	/* */
	return ReadAFTFileFrame(afi, frame, dest);
	}
    }

#define LOBOS	8	/* how many points to read for each interpolation */
#define MAXUNDERSAMPL 4 /* undersampling factor allowed for <prd> */

void SpecOfAFTFile(afi, time, prd, ans, dphasq)
    AFTINFO *afi;
    FLOATARG time;
    FLOATARG prd;
    cpxpolar *ans;
    int dphasq;		/* if nz, return phi-dot, not phase */
    /* using the AFT file described by <afi>
       interpolate frames to obtain a spectrum at the exact time
       specified by <time>.  Low pass filter the result where necessary to
       avoid aliasing if the sampling increment (in seconds) is <prd>.
       Put this complex, polar vector into the space
       pointed to by <ans>, assumed large enough.
       Differentiate phases if dphasq != 0 */
    {
    long	zeroFrameIndex, frmIndexPerFrame, firstFrame, lastFrame;
    long	numFrames, frame, i;
    float	frameLocation;
    float	timePerFrame;
    cpxpolar	*rsltPtr;
    cpxpolar	*rdPtr;
static cpxpolar	*rdBuffer = NULL;
static long     rdBufPts = 0;
    int 	lobes, maxlobes;
    float	lpfRatio;
    float	octFactor, binFactor, lowestFrq, frqInHz, twoPiDt, edp;
    int 	frCpxs = afi->pvh->frameBsize / sizeof(cpxpolar);

#ifdef DEBUGGRAPHS
static    WINDAT	*wdptr;
static 	  WINDAT	*reswin;
    char	dummy[16];
    float	dumbuf[MAXFCPX];
#endif

    if(rdBuffer == NULL)
	{
	rdBufPts = MAXFCPX * LOBOS * MAXUNDERSAMPL;
	rdBuffer = (cpxpolar *)Mymalloc(rdBufPts, sizeof(cpxpolar), "SpecOfAFT");
	
#ifdef DEBUGGRAPHS
	wdptr = NULL;
#endif
	}

/*    fprintf(stderr, "SpecOfAFTFile(%lx, %f, %f, %lx, %d)\n",
	    afi, time, prd, ans, dphasq); /* */
    timePerFrame = afi->pvh->frameIncr/afi->pvh->samplingRate;
    twoPiDt   = 2.0*PI*timePerFrame;
#ifdef DEBUGGRAPHS
    sprintf(dummy,"as read");
    if(wdptr == NULL)
	wdptr = NewWin(dummy, 0);	
#endif
    zeroFrameIndex = 0;
    frmIndexPerFrame = 1;
    rsltPtr = ans;
    numFrames      = afi->totalFrames;
    maxlobes = rdBufPts / frCpxs;
    if(timePerFrame < prd)	/* undersampled : need to LPF frames on file */
	{
	lobes = (LOBOS*prd)/timePerFrame;
	lpfRatio = prd/(float)timePerFrame;
	}
    else{
	lobes = LOBOS;	/* oversampled : use min. num of frames */
	lpfRatio = (float)1;
	}
    if(lobes > maxlobes)	lobes = maxlobes;	/* clamp to buf size */
    
    frameLocation = time/timePerFrame;
    firstFrame = (int)frameLocation - (lobes/2-1);
    lastFrame  = firstFrame + lobes;	/* one after that read */
	
    if(firstFrame < 0)	firstFrame = 0;
    else if(firstFrame>=numFrames)  firstFrame = numFrames-1;
    if(lastFrame <= 0)	lastFrame = 1;
    else if(lastFrame>numFrames)    lastFrame = numFrames;
/*    fprintf(stderr,
            "SAFT: zFI %ld, tPF %6.5f, frs %ld, 1st %ld, lst %ld\n",
            zeroFrameIndex, timePerFrame,
            numFrames, firstFrame, lastFrame); /*	*/
    rdPtr = rdBuffer;
    for(frame = firstFrame; frame < lastFrame; ++frame)
	{
	if(!ReadAFTFileFrame(afi, zeroFrameIndex+frmIndexPerFrame*frame, 
			     rdPtr))
	    lastFrame = frame;
	else{
#ifdef DEBUGGRAPHS
	    CopyFloats((float *)rdPtr, 2, dumbuf, 1, frCpxs, 1);
	    DoDisp(wdptr, dumbuf, frCpxs);	
#endif
	    rdPtr += frCpxs;
	    }
	}
    rdPtr = rdBuffer;
#ifdef DEBUGGRAPHS
    if(reswin==NULL)
	reswin = NewWin("result",1);	
#endif	

/*    fprintf(stderr, "SpFTF: frames %ld, pos %6.3f, ratio %6.3f\n",
            lastFrame-firstFrame, frameLocation-firstFrame,lpfRatio);
  /* */
    SetupInterpVal(lastFrame-firstFrame, frameLocation-firstFrame,
		   lpfRatio, timePerFrame, INTERPTYPE);	
                   /* last arg: 1=sinc, 0=noneg */
    for(i = 0; i < frCpxs; ++i)
	{
	DoInterpVal((float *)(rdPtr+i), 2*frCpxs, (float *)(rsltPtr+i)); 
	}
    SetupInterpVal(lastFrame-firstFrame, frameLocation-firstFrame,
		   (float)1.0 * lpfRatio, timePerFrame, 
		   dphasq?DDIFF:INTERPTYPE);
    /* make interpolator do differentiation too for phase if requested */
    for(i = 0; i < frCpxs; ++i)
	{
	frqInHz = i * afi->pvh->samplingRate / (float)afi->pvh->frameSize;
	edp = twoPiDt*frqInHz;
	UnwrapPhases(((float *)(rdPtr+i))+1, 2*frCpxs, 
		     ((float *)(rdPtr+i))+1, 2*frCpxs,
		     lastFrame - firstFrame, edp);
	DoInterpVal((float *)(rdPtr+i)+1, 2*frCpxs,(float *)(rsltPtr+i)+1); 
	if(0/*debug*/)
	    {
	    float	dphi, afrq;
	    dphi = rdPtr[i+frCpxs*(int)(1+frameLocation-firstFrame)].pha 
	           - rdPtr[i+frCpxs*(int)(frameLocation-firstFrame)].pha;
	    dphi += 2.0*PI*(int)((edp-dphi+PI)/(2.0*PI));
	    afrq = (dphi/timePerFrame) / (2*PI);
	    fprintf(stderr,"frq %d = (%.2f) %.2f (%.2f) -> %.2f\n", 
		    i,frqInHz,afrq,edp,rsltPtr[i].pha/(2*PI));
#ifdef DEBUGGRAPHS
	    CopyFloats(((float *)(rdPtr+i))+1, 2*frCpxs, dumbuf, 1, 
		       lastFrame-firstFrame, 1);
	    DoDisp(reswin, dumbuf, lastFrame-firstFrame);	
#endif
	    }

	}
    if(!dphasq)		/* only mask actual phases, not derivatives */
	MaskPhases(((float *)rsltPtr)+1, 2, ((float *)rsltPtr)+1, 2, frCpxs);
    
#ifdef DEBUGGRAPHS
    CopyFloats((float *)rsltPtr, 2, dumbuf, 1, frCpxs, 1);
    DoDisp(reswin, dumbuf, frCpxs);	
#endif

    }

/* SimpleSpecOfAFTFile
   As a more predictable alternative to SpecOfAFTFile with its 
   discrete differentiator and extravagent range of reference, this 
   is a return pretty much to the linear-interp-between-two-frames 
   kind of phase vocoder.  This call, with the same arguments as 
   SpecOfAFTFile, grabs four frames from the datafile/cache, two 
   on each side of the point of interpolation.  (these four frames 
   are loaded into a private, dynamically allocated data buffer - 
   problems if afi->frameSize changes! - and they are *not* reread on 
   successive calls if the base frame is the same (although the 
   exact interpolation point can change, e.g. frame 32.1 vs frame 32.9).  
   The returned magnitudes are linear interpolation between the nearest 
   two values.  The phase is unwrapped along the four points; if phase 
   alone is requested, the linearly interpolated phase along this 
   slope segment is returned (masked to principle value before return).
   If phase slope (dphi/dt) is requested, the value returned is the 
   linear interpolation between the two simple slopes by taking phase 
   values *either side* of each boundary point.  Thus the phase slope 
   at frame 32.1 is (0.9)*(phase(33)-phase(31)+(0.1)*(phase(34)-phase(32)) 
   (all divided by the appropriate dt i.e. twice the frame sampling 
   period since the differences are two frames apart).
   In the case where the frames are interpolated against the edges of 
   the available frame range, these slopes will drop back to 
   phase(n+1)-phase(n) rather than phase(n+1)-phase(n-1)  etc.
   
   dpwe 1992sep19
 */ 

void SimpleSpecOfAFTFile(afi, time, prd, ans, dphasq)
    AFTINFO *afi;
    FLOATARG time;
    FLOATARG prd;
    cpxpolar *ans;
    int dphasq;		/* if nz, return phi-dot, not phase */
    {
    long i, f, totFrms;
    long  frameFloats = afi->pvh->frameSize+2;	/* # of floats in data frame */
    float frameNum = time / (afi->pvh->frameIncr / afi->pvh->samplingRate);
    float frac, d1, d2, timePerFrame;
static float *iaftbuf = NULL;
static long  baseFrame;

    if(iaftbuf == NULL) {
	iaftbuf = (float *)Mymalloc(4*frameFloats, sizeof(float), "SimpOfAFT");
	baseFrame = -1;	/* won't match anything */
	}
    f = (long)frameNum;
    frac = frameNum - f;	/* the fractional part */
    totFrms = afi->totalFrames;	/* 1+largest legal requestable index */
    timePerFrame = afi->pvh->frameIncr/afi->pvh->samplingRate;
    if(f!=baseFrame)  {	/* don't reload if same as last time */
	float twoPiDt;

	/* load the frames */
	for(i = 0; i<4; ++i)  {
	    long actfrm = f+i-1;	/* load frames f-1, f, f+1, f+2 */
	                                /* since interpolating f & f+1 */
	    if(actfrm>=0 && actfrm<totFrms)
		ReadAFTFileFrame(afi, actfrm, 
				 (cpxpolar *)(iaftbuf+i*frameFloats));
	    }
	/* unwrap the phases */
	twoPiDt   = 2.0*PI*timePerFrame;
	for(i = 0; i < frameFloats/2; ++i)
	    {
	    float frqInHz, edp;

	    frqInHz = i * afi->pvh->samplingRate / (float)afi->pvh->frameSize;
	    edp = twoPiDt*frqInHz;
	    UnwrapPhases(iaftbuf+2*i+1, frameFloats, iaftbuf+2*i+1, 
			 frameFloats, 4, edp);
	    }
	baseFrame = f;
	}
    /* linear interpolation of magnitude */
    if(frac == 0.0)
	CopyFloats(iaftbuf+frameFloats, 2, (float *)ans, 2, frameFloats/2, 
		   FWD);
    else
	for(i = 0; i<frameFloats/2; ++i)  {
	    d1 = *(iaftbuf + frameFloats + 2*i);
	    ans[i].mag = d1 + frac * ( *(iaftbuf + 2*frameFloats + 2*i) - d1 );
	    }
    /* either linearly interpolate unwrapped phase, or take slope between 
       points either side and interpolate between them */
    if(dphasq) {
	int a,b,c,d;
	float g,h;

	if(frac == 0.0)  {
	    if(f == 0) {	/* first frame; iaftb[0..frameFloat-1]inval */
		a = 1; b = 2; g = 1.0;
		}
	    else if(f == totFrms - 1) { /* last frame; iab[>2*fF] inval */
		a = 0; b = 1; g = 1.0;
		}
	    else {	/* OK to read frames on both sides */
		a = 0; b = 2; g = 2.0;
		}
	    for(i = 0; i<frameFloats/2; ++i)  {
		ans[i].pha = (*(iaftbuf + b*frameFloats + 2*i + 1)
			      - *(iaftbuf + a*frameFloats + 2*i + 1))/g
				  /timePerFrame;
		}
	    }
	else {	/* have to check which of 4 frame is valid like above */
	    if(f == 0) {
		a = 1; b = 2; g = 1.0;
		c = 1; d = 3; h = 2.0;
		}
	    else if(f == totFrms - 2) {	/* can't have frac >0 with f = tF-1 */ 
		a = 0; b = 2; g = 2.0;
		c = 1; d = 2; h = 1.0;
		}
	    else {	/* ok to take two 2-spaced slopes */
		a = 0; b = 2; g = 2.0;
		c = 1; d = 3; h = 2.0;
		}
	    for(i = 0; i<frameFloats/2; ++i)  {
		d1 = (*(iaftbuf + b*frameFloats + 2*i + 1) 
		     - *(iaftbuf + a*frameFloats + 2*i + 1))/g;
		d2 = (*(iaftbuf + d*frameFloats + 2*i + 1) 
		 - *(iaftbuf + c*frameFloats + 2*i + 1))/h;
		ans[i].pha = (d1 + frac * (d2 - d1))/timePerFrame;
		}
	    }
	}
    else {
	if(frac == 0.0)
	    CopyFloats(iaftbuf+frameFloats+1, 2, ((float *)ans)+1, 2, 
		       frameFloats/2, FWD);
	else
	    for(i = 0; i<frameFloats/2; ++i)  {
		d1 = *(iaftbuf + 1*frameFloats + 2*i + 1);
		d2 = *(iaftbuf + 2*frameFloats + 2*i + 1) ;
		ans[i].pha = (d1 + frac * (d2 - d1));
		}
	}
    if(!dphasq)		/* only mask actual phases, not derivatives */
	MaskPhases(((float *)ans)+1, 2, ((float *)ans)+1, 2, frameFloats/2);
    if(0)  {
	int a = 45;
	fprintf(stderr, "frmnm %8f pha[%d] = %8f\n", frameNum, a, ans[a].pha);
	}
    }

/********************* test code only below here ***************************/
#ifdef DEBUG

#define VECLEN 32L
#define BIGVEC 400L
static test_phase()
    {
    float vec[VECLEN];
    float r[BIGVEC];
    long i,j;
    float f;
#ifdef DEBUGGRAPHS
    WINDAT *wdptr;
#endif

    for(i=0; i<VECLEN/2; ++i)
	vec[i] = (float)i/2.0;
    for(i = VECLEN/2; i<VECLEN; ++i)
	vec[i] = (float)VECLEN/4.0; /* - (i-VECLEN/2.0);	*/
    print_vec(vec, VECLEN, "start");
    MaskPhases(vec,1,vec,1,VECLEN);
    print_vec(vec, VECLEN, "masked");
    UnwrapPhases(vec,1,vec,1,VECLEN,(float)0);
    print_vec(vec, VECLEN, "unwrapd");

    for(i = 0; i<BIGVEC; ++i)
	{
	SetupInterpVal(VECLEN, VECLEN/2.0-10.0+(float)i/20.0, 2.0, 1.0,
		       INTERPTYPE);
	DoInterpVal(vec,1,&r[i]);

/*	InterpolateValue(vec,1,VECLEN,VECLEN/2.0-10.0+(float)i/20.0,&r[i],
			 1.0,0);	*/
	}
#ifdef DEBUGGRAPHS
    wdptr = NewWin("test interp",1);
    DoDisp(wdptr, r, BIGVEC);
#endif
    for(i = 0; i<BIGVEC; ++i)
	{
	SetupInterpVal(VECLEN, VECLEN/2.0-10.0+(float)i/20.0, 1.0, 0.1, 
		       INTERPTYPE);
	DoInterpVal(vec,1,&r[i]);
/*	InterpolateValue(vec,1,VECLEN,VECLEN/2.0-10.0+(float)i/20.0,&r[i],
			 1.0,1);	*/
	}
#ifdef DEBUGGRAPHS
    wdptr = NewWin("test interp 2",1);
    DoDisp(wdptr, r, BIGVEC);
#endif
    }

#define BUFLEN (257)	/* number of bins ? */

static test_afint()
    {
    FILE *file;
    AFTINFO *afi;
    cpxpolar	rslt[BUFLEN];
    float	res2[BUFLEN];
#ifdef DEBUGGRAPHS
    WINDAT	*wdptr;
#endif
    int		i,j;

    ClearFloats((float *)rslt,1,2*BUFLEN);
    ClearFloats(res2,1,BUFLEN);
#ifdef DEBUGGRAPHS
    wdptr = NewWin("afint",1);
#endif
    file = OpenAFTPVfile("/usr/local/sound/test/sins.aft", &afi);
    for(i = 2; i>=0; --i)
	{
	SpecOfAFTFile(afi, .4 + .04*(float)i, .04, rslt, 1);
	for(j=0; j<BUFLEN; ++j)
	    {
	    fprintf(stderr, "j %d phidot %f frq %f\n", j, rslt[j].pha,
		    rslt[j].pha/(2*PI));
	    }
	CopyFloats((float *)rslt,2,res2,1,BUFLEN,1);
#ifdef DEBUGGRAPHS
	DoDisp(wdptr, res2, BUFLEN);
#endif
	}
    }

char *programName;
main(argc, argv)
    int argc;
    char **argv;
    {
    programName = argv[0];
    test_afint();
    ExitGraph();
    }


#define BUFLEN2 (6*24)

void TestRead(filename, ppArray, pTmin, pTmax, pFmin, pFmax)
    char	*filename;
    SampArray	**ppArray;
    float	*pTmin;
    float 	*pTmax;
    float	*pFmin;
    float 	*pFmax;
    {
    FILE *file;
    AFTINFO *afi;
    int rows, cols, i;
    float *fPtr, tmpBuf[2*BUFLEN2];

    if( (file=OpenAFTPVfile(filename, &afi))==NULL)
	exit(1);
    rows = afi->pvh->frameBsize/sizeof(cpxpolar);
    cols = 40;

    (*pTmin) = 0.0;			(*pTmax) = afi->duration;
    *pTmin = .40 * afi->duration; 	*pTmax = .60*afi->duration;	/* */
    (*pFmin) = 0;	(*pFmax) = afi->pvh->samplingRate/2.0;
    *ppArray = NewSampA(rows, cols);
    fPtr = (*ppArray)->data;
    for(i=0; i<cols; ++i)
	{
/*	fprintf(stderr,"%8d",i);	fflush(stderr);	*/
	SpecOfAFTFile(afi, *pTmin + (float)i/(float)cols*(*pTmax-*pTmin), 
		      1/(float)cols*(*pTmax-*pTmin), (cpxpolar *)tmpBuf, 1);
	CopyFloats(tmpBuf, 2, fPtr+cols*(rows-1), -cols, rows, 1);
	++fPtr;
	}
    }
/** end of test code **/

#endif /* def DEBUG */
