/***************************************************************\
*	twindow.c						*
*  time windowing operations	       				*
*  dpwe 06nov91							*
\***************************************************************/

#include <math.h>
#include <dpwelib.h>
#include <twindow.h>

void FreeTwindow(twdw)
    TWINDOW	*twdw;
    {
    free(twdw);
    }

void FillWindowBuffer(buf, size, fsize, gain, type)
    float	*buf;	/* array */
    long 	size;	/* actual size of array */
    FLOATARG	fsize;	/* size of window, centered in array */
    FLOATARG	gain;	/* scaling away from unity */
    TWINTYPE	type;	/* which window */
    {
    long l, hsize;
    float pi = PI;
    float hfsize;
    float *buf2;

    /* preset whole thing to zero */
    buf2 = buf;
    for(l = 0; l<size; ++l)	*buf2++ = 0.0;

    hsize = (size-1)/2;		/* index of array middle point */
    buf2  = buf+hsize;		/* point to middle sample */
    if(fsize == 0.0)	fsize = (float)size;
    hfsize = (fsize-1)/2.0;
    /* hsize describes range of cells to fill away from centre */
    /* if fsize is telling us only to do a bit in the middle, restrict */
    if( ((long)hfsize) < hsize )	hsize = (long)hfsize;

    switch(type)
	{
    case TT_NONE: case TT_ZERO: case TT_OTHR:
	/* leave it as zero */
	break;
    case TT_HAMM:
	for(l = 0; l<=hsize; ++l)
	    *(buf2+l) = *(buf2-l) = gain*(0.5+0.5*cos(pi*((float)l)/hfsize));
	break;
    case TT_HANN:
	for(l = 0; l<=hsize; ++l)
	    *(buf2+l) = *(buf2-l) = gain*(0.54+0.46*cos(pi*((float)l)/hfsize));
	break;
    case TT_FRXP: case TT_RRXP:
	fprintf(stderr, "FillWinBuf: roexp not yet implemented\n");
	abort();
	break;
    default:
	fprintf(stderr, "FillWinBuf: unrecog window type %d\n", (int)type);
	abort();
	break;
	}
    }

TWINDOW	*NewTwindow(size, fsize, gain, type)
    long	size;		/* number of samples to hold of window */
    FLOATARG	fsize;		/* actual length - real, so can be frac */
    FLOATARG 	gain;		/* scaling away from unity (power? peak?) */
    TWINTYPE	type;		/* indicates type */
    {
    long	twByteSize;
    TWINDOW	*newTwdw;

    /* constrain length to be odd ? */
    if( (size & 1) == 0 )	/* passed size is even */
	++size;
    twByteSize = sizeof(TWINDOW) + size*sizeof(float);
    if( (newTwdw = (TWINDOW *)malloc(twByteSize)) == NULL)
	{  fprintf(stderr, "NewTwdw: malloc fail (wanted %ld)\n", twByteSize);
	   abort();	}
    newTwdw->kern = (float *)(newTwdw+1);	/* space is after struct */
    newTwdw->size = size;
    newTwdw->type = type;
    FillWindowBuffer(newTwdw->kern, size, fsize, gain, type);
    return newTwdw;
    }

void ApplyTwindow(twdw, in, inlen, out, outlen)
    TWINDOW	*twdw;	/* window to apply - padded with zeros if too short */
    float	*in;		/* what to window */
    long	inlen;		/* how long it is */
    float	*out;		/* where to put answer */
    long 	outlen;		/* how much answer to put */
    {
    long 	l, len;
    float	*win = twdw->kern;
    long	wlen = twdw->size;

    if(out == NULL)	out = in;
    if(outlen == 0)	outlen = inlen;
    /* if the input is even length, we align the odd length window one
       point later */
    if(inlen > outlen)	/* can only propogate some to output - skip input */
	{
	/* if inlen is one larger than outlen, drop last point */
	inlen += (inlen-outlen)/2;
	inlen = outlen;
	}
    if(inlen > wlen)	/* window shorter than input - skip input */
	{
	/* twdw->size is assumed odd.  If insize is even but one greater
	   than twdw, we want the twdw to start at (in+1) (gap at front) */
	in += (inlen + 1 - wlen)/2;
	inlen = wlen;
	}
    if(inlen < wlen)	/* input shorter than window - skip window */
	{
	/* very often, window size is one more than input data.
	   In this case, we discard the last point in the window */
	win += (wlen - inlen)/2;
	wlen = inlen;
	}
    if(inlen < outlen)
	{
	/* if inlen is even, tend to put 'extra' zero after it (outlen odd)
	   if inlen is odd, put extra zero in front (outlen even) */
	len = (outlen-(inlen&-2))/2;
	for(l = 0; l<len; ++l)
	    {
	    *out++ = 0.0;
	    --outlen;
	    }
	}
    /* apply window to input */
    for(l = 0; l<inlen; ++l)
	{
	*out++ = *in++ * *win++;
	--outlen;
	}
    /* zero rest of output buffer */
    while(outlen>0)
	{
	*out++ = 0.0;
	--outlen;
	}
    }

#ifdef DEBUG	/* test code */

#define ISIZE		16
#define OSIZE 		16
#define WSIZE		16

main()
    {
    TWINDOW *twdw;
    float   in[ISIZE], out[OSIZE];
    long    i;

    for(i=0; i<ISIZE; ++i)
	in[i] = 1.0;
    twdw = NewTwindow(WSIZE, 0.0, 1.0, TT_HAMM);
    ApplyTwindow(twdw, in, ISIZE, out, OSIZE);
    for(i=0; i<OSIZE; ++i)
	fprintf(stderr, "%3d: %8.5f\n", i, out[i]);
    FreeTwindow(twdw);
    }


#endif /* DEBUG */
