/***************************************************************\
* wndw.c							*
* Moderately device independent drawing primitives		*
* Essentially, allows XWindow draws to be diverted to   	*
* PostScript, rather than endless screendumps.			*
* Based on stripk		 				*
* Generalized dpwe 1992oct24					*
\***************************************************************/

#include "wndw.h"
#include <math.h>	/* log/exp in WAddLogTime */
#include <idrawop.h>	/* hand-offs to idraw output */

/* chars are promoted to ints when args */
#define CHARARG int

/* static fn protos */
static float WNearestRoundNum PARG((FLOATARG num, float *sub));
static void WReplaceOrAppend PARG((char *s, CHARARG c1, CHARARG c2));
static int WDecPoints PARG((FLOATARG step));
static void SetupColours PARG((Display *disp, int scrnNum));

static XColor *colours[WGMD_NCOLORS];

void WHide(WIND *wind)
{   /* make this window invisible */
    if(wind->hidden)  {
	DBGFPRINTF((stderr, "WHide: wind %lx already hidden\n", wind));
    } else {
	XUnmapWindow(wind->disp, wind->win);
	wind->hidden = TRUE;
    }
/*    XDestroyWindow(wind->disp, wind->win);  */
}

void WUnhide(WIND *wind)
{   /* make a hidden window visible again */ 
    if(!wind->hidden)  {
	DBGFPRINTF((stderr, "WUnhide: wind %lx already visible\n", wind));
    } else {
	XMapWindow(wind->disp, wind->win);
	wind->hidden = FALSE;
    }
}

void WSwapGCs(WIND *wind)
{	/* two gc's are provided, so can perform limited context switch */
    GC	gc;

    gc = wind->gc;
    wind->gc = wind->altgc;
    wind->altgc = gc;
}

void WPushGC(WIND *wind)
{	/* save current graphics context for later restore (actually swaps) */
    WSwapGCs(wind);
}

void WPopGC(WIND *wind)
{	/* restore graphics context earlier saved (actually swaps) */
    WSwapGCs(wind);
}

static void WNewGC(WIND *wind)
{
    XGCValues	gcVals;
    GC		gc;	

    SetupColours(wind->disp, wind->scrn);

    gcVals.function   = GXcopy;  
    gcVals.line_width = 0;
    wind->fg = WGMD_BLACK;
    wind->bg = WGMD_WHITE;
    gcVals.foreground = colours[wind->fg]->pixel;
    gcVals.background = colours[wind->bg]->pixel;
    gc = XCreateGC(wind->disp, wind->win, 
		   GCFunction|GCLineWidth|GCForeground|GCBackground, &gcVals);
    wind->gc = gc;
}

void WSetGMode(WIND *wind, int fg, int bg, int mode, int lwidth, int lstyle)
{
    Display	*disp = wind->disp;
    int 	scrnNum = wind->scrn;
    GC  	gc = wind->gc;
    char	dashList[1];
    unsigned long fgpixel, bgpixel;

    wind->fg = fg;	fgpixel = colours[wind->fg]->pixel;
    wind->bg = bg;	bgpixel = colours[wind->bg]->pixel;
    XSetForeground(disp, gc, fgpixel);
    XSetBackground(disp, gc, bgpixel);

    if(mode == WGMD_XOR)  {
	XSetState(disp, gc, fgpixel^bgpixel, bgpixel, GXxor, -1L);
    }
    else {
	XSetState(disp, gc, fgpixel, bgpixel, GXcopy, -1L);
    }
    if(lstyle == WGMD_DOTS)  {
	XSetLineAttributes(disp, gc, lwidth, LineOnOffDash, 
			   CapNotLast, JoinMiter);
	dashList[0] = '\001';
	XSetDashes(disp, gc, 0, dashList, 1); 
    }
    else {
	XSetLineAttributes(disp, gc, lwidth, LineSolid, CapButt, JoinMiter);
    }
}

static WIND *windRoot = NULL;	/* (keep list of all winds we make) */

WIND *NewWind(disp, scrn, win, idFname)
    Display	*disp;
    int 	scrn;
    Window	win;
    char        *idFname;
    {
    WIND	*wind;

    wind = TMMALLOC(WIND, 1, "NewWind");
    if(wind != NULL)
	{
	wind->disp = disp;
	wind->scrn = scrn;
	wind->win  = win;
	wind->gc   = NULL;
	wind->cur_x = 0;
	wind->cur_y = 0;
	wind->hidden = FALSE;	/* well, don't call hidden till after Hide */
	wind->linew = 1.0;
	wind->opf   = NULL;	/* by default */
	wind->evhs  = NULL;	/* no event handlers yet */
	if(idFname != NULL && idFname[0] != '\0')
	    idOpen(wind, idFname, 0, 0, 0, 0);
	WGetWidthHeight(wind);
	WNewGC(wind);
	WSwapGCs(wind);
	WNewGC(wind);	/* setup both gcs */
	/* insert in linked list */
	wind->next = windRoot;
	windRoot = wind;
	}
    return wind;
    }

static WIND **windFindParent(WIND *child)
{	/* find address of pointer to this window in our private linked list */
    WIND **rslt = &windRoot;
    int  found = 0;

    while(!found && *rslt != NULL)  {
	if(*rslt == child)  {
	    found = 1;
	} else {
	    rslt = &( (*rslt)->next );
	}
    }
    if(found)
	return rslt;
    else
	return NULL;
}

void FreeWind(wind)
    WIND	*wind;
{
    EVHDLR_NODE *n1, *n2 = wind->evhs;
    WIND	**ppar;

    if( ppar = windFindParent(wind) )  {
	/* unlink from our linked list */
	*ppar = wind->next;
    } else {
	fprintf(stderr, "FreeWind: unable to unlink %lx\n");
	/* but, hell, we'll just do it anyways */
    }
    idClose(wind);
    XDestroyWindow(wind->disp, wind->win);
    while(n2)  {	/* destroy linked list of event handler nodes */
	n1 = n2;	n2 = n2->next;		free(n1);
    }
    free(wind);
}

static Display 	    *xdisp = NULL;
static int     	    xscrn;
static GC           xdfgc;
static GContext	    xgcon;
static Font         xfont;
static XFontStruct  *xfsp;

void WSetXdisp(Display *disp)
{	/* short cut to prevent multiple XOpenDisplays */
    xdisp = disp;
}

int WInitX(void)
{	/* intialize the X server */
    char 	*disp, *getenv(const char *);
    int	  	rc = 0;		/* default : don't use X, use tty ascii */
    char	*fontname;

    if (xdisp == NULL)  {
	disp = getenv("DISPLAY");
	if(disp == NULL) return 0;
	xdisp = XOpenDisplay(disp);
    }
    if (xdisp != NULL) { /* if this OK, X is possible */
	xscrn = XDefaultScreen(xdisp);
	xdfgc = XDefaultGC(xdisp, xscrn);
	xgcon = XGContextFromGC(xdfgc);
	fontname = "6x13";                      /* set up font info... */
	xfsp  = XLoadQueryFont(xdisp,fontname);
        if(xfsp == NULL)  {
            fprintf(stderr, "cannot load font %s, trying fixed\n", fontname);
            xfsp = XLoadQueryFont(xdisp, "fixed");
        }
	xfont = xfsp->fid;
	rc = 1; 			        /* so set return code  */
	}
    return(rc);
}

void WFlush(void)
{	/* flush the x events */
    XFlush(xdisp);
}

/* active events for our windows */
#define GWINEVMSK (ExposureMask | ButtonPressMask \
		   | ButtonReleaseMask | ButtonMotionMask)
/* tried ResizeRedirect - never figured it out */
#define BDR      2  	/* window border width (?) */

Window MakeXWindow(name, x, y, w, h)
    char	*name;
    int  	x, y, w, h;
{
    Window		 xwin;
    XSetWindowAttributes watts;
    unsigned long        wmask;

    watts.background_pixel = XWhitePixel(xdisp, xscrn);
    watts.backing_store    = WhenMapped;
    /* prechosen events .. */
    watts.event_mask       = GWINEVMSK;
    wmask = CWBackingStore | CWEventMask | CWBackPixel;
    xwin = XCreateWindow(xdisp, XDefaultRootWindow(xdisp),
			 x, y, w, h, BDR, 
			 CopyFromParent, InputOutput, CopyFromParent, 
			 wmask, &watts);
    XStoreName(xdisp, xwin, name);
    XMapWindow(xdisp, xwin);   /* map the window to the screen */
    return(xwin);
}

void WNewName(WIND *wndw, char *name)
{	/* change the name of the window */
    XStoreName(wndw->disp, wndw->win, name);
}

WIND *NewXWind(x, y, w, h, name, idFname)
    int 	x, y, w, h;
    char	*name;
    char        *idFname;
{
    Window xwin;

    if(xdisp == NULL)
	if(WInitX() == 0)
	    return NULL;

    xwin = MakeXWindow(name, x, y, w, h);
    return NewWind(xdisp, xscrn, xwin, idFname);
}

void WBlitArea(WIND *wndw, RECT *pr, int right, int down)
{	/* use blitting to move the contents of the specified rectangle 
	   by the specified amounts.  Negative args give left and 
	   up shifts respectively.  The result will not move outside 
	   of <*pr>, so the region shifted is reduced accordingly.  
	   If <pr> is NULL, the entire window is shifted. */
    int x, y, w, h, destx, desty;

    if(pr) {
	x = pr->x;		y = pr->y;
	w = pr->width;		h = pr->height;
    } else {
	x = 0;			y = 0;
	w = wndw->wWidth;	h = wndw->wHeight;
    }
    destx = x; 			desty = y;
    if(right > 0)  {
	w -= right;		destx = x + right;
    } else if(right < 0)  {
	w += right;		x -= right;
    }
    if(down > 0)  {
	h -= down;		desty = y + down;
    } else if (down < 0)  {
	h += down;		y -= down;
    }
    XCopyArea(wndw->disp, wndw->win, wndw->win, wndw->gc, x, y, 
	      w, h, destx, desty);
}

void WSetBlack(wind)
    WIND	*wind;
    {
    XSetForeground(wind->disp, wind->gc, XBlackPixel(wind->disp, wind->scrn));
    }

void SetLineWidth(wind, width)
    WIND	*wind;
    FLOATARG	width;
    {
    wind->linew = width;
    }

void WGetWidthHeight(WIND *wdw)
{
    XWindowAttributes info;

    XGetWindowAttributes(wdw->disp, wdw->win, &info);
    wdw->wWidth = info.width;  wdw->wHeight = info.height;
}

void WResizeWindow(WIND *wind, int width, int height)
{	/* does this work properly?  problems with resizeRedirect */
    fprintf(stderr, "WRsz w=%d, h=%d\n", width, height);
    XResizeWindow(wind->disp, wind->win, width, height);
    wind->wWidth = width;  wind->wHeight = height;
}

void WMoveTo(wind, x, y)
    WIND    *wind;
    int     x, y;
    {
    wind->cur_x = x;  wind->cur_y = y;
    }

void WLineTo(wind, x, y)
    WIND    *wind;
    int     x, y;
    {
    if( wind->opf != NULL )
	idLineTo(wind, x, y);
    XDrawLine(wind->disp, wind->win, wind->gc, 
	      wind->cur_x, wind->cur_y, x, y);
    WMoveTo(wind, x, y);
    }

void WDrawLine(WIND *wind, int x1, int y1, int x2, int y2)
{
    WMoveTo(wind, x1, y1);
    WLineTo(wind, x2, y2);
}

void WDrawX(WIND *wind, RECT *pr)
{
    WClearRect(wind, pr);
    WFrameRect(wind, pr);
    WDrawLine(wind, pr->x, pr->y, pr->x+pr->width, pr->y+pr->height);
    WDrawLine(wind, pr->x, pr->y+pr->height, pr->x+pr->width, pr->y);
}

#define MAXPTS	200

void WPolyLine(WIND *wind, int buf[][2], long cnt)
{	/* poly line independent of line from/to */
    /*    XDrawLines(disp,thisWin,gc,(XPoint *)pts, npts, CoordModeOrigin); */
static XPoint	pts[MAXPTS];
    int i, npts;

    if(wind->opf != NULL)
	idPolyLine(wind,  buf, cnt);
    pts[0].x = buf[0][0];    pts[0].y = buf[0][1];
 top:
    npts = MIN(cnt, MAXPTS);
    for(i = 1; i<npts; ++i)  {
	pts[i].x = buf[i][0];
	pts[i].y = buf[i][1];
    }
    XDrawLines(wind->disp, wind->win, wind->gc, pts, npts, CoordModeOrigin);
    cnt -= npts - 1;
    if(cnt > 1)  {
	pts[0] = pts[npts-1];
	buf += npts - 1;
	goto top;
    }
    return;
}

void WClearExpose(WIND *wind, RECT *pr)
{	/* clear rect; generate exposes */
    XClearArea(wind->disp, wind->win, pr->x, pr->y, pr->width, pr->height, 
	       True);
}

void WClearRect(wind, pr)
    WIND    *wind;
    RECT    *pr;
{
    if(pr)
	XClearArea(wind->disp, wind->win, pr->x, pr->y, pr->width, pr->height, 
		   False);
    else
	XClearArea(wind->disp, wind->win, 0, 0, wind->wWidth, wind->wHeight, 
		   False);
}

void WFillRect(wind, pr)
    WIND    *wind;
    RECT    *pr;
    {
    if( wind->opf != NULL )
	idFill(wind, pr);
    XFillRectangle(wind->disp, wind->win, wind->gc, 
		   pr->x, pr->y, pr->width, pr->height);
    }

void WColorRect(WIND *wind, RECT *pr, unsigned long color)
{	/* I'm sorry.  Color a rectangle in a color gotten from 
	   the pseudocolor lookup table - NOT from the other color method */
    XSetForeground(wind->disp, wind->gc, color);
    XFillRectangle(wind->disp, wind->win, wind->gc, pr->x, pr->y, 
		   pr->width, pr->height);
}

void WFrameRect(wind, pr)
    WIND    *wind;
    RECT    *pr;
    {
    if( wind->opf != NULL )
	idFrame(wind, pr);
    XDrawRectangle(wind->disp, wind->win, wind->gc, 
		   pr->x, pr->y, pr->width, pr->height);
    }

void WSetClip(WIND *wind, RECT *pr)
{   /* Clip graphics to within rect, or no clipping if <pr> NULL */
    XRectangle r, ra[1];

    if(pr==NULL)  {
	XSetClipMask(wind->disp, wind->gc, None);
    } else {
	r.x = pr->x; r.y = pr->y; r.width = pr->width; r.height = pr->height;
	ra[0] = r;
	XSetClipRectangles(wind->disp, wind->gc, 0, 0, ra, 1, Unsorted);
    }
}

void WWriteAt(wind, x, y, str)
    WIND    *wind;
    int     x, y;
    char    *str;
    {
static XTextItem txitem[1];
static int init = 0;
    int     nc, dir, fa, fd;
    XCharStruct	xcs;

    if(init==0)
	{
	txitem[0].delta  = 1;
	txitem[0].font   = None;
	init = 1;
	}
    if( wind->opf != NULL )
	idWriteAt(wind, x, y, str);
    nc = strlen(str);
    XQueryTextExtents(wind->disp, XGContextFromGC(wind->gc),
		      str, nc, &dir, &fa, &fd, &xcs);
    txitem[0].chars  = str;
    txitem[0].nchars = nc; 
    XDrawText(wind->disp, wind->win, wind->gc, x, y+fa, txitem, 1);
    }

int WTextWidth(WIND *wind, char *str)
{
    XCharStruct xcs;
    int         nc, dir, fa, fd;

    nc = strlen(str);
    XQueryTextExtents(wind->disp, XGContextFromGC(wind->gc),
		      str, nc, &dir, &fa, &fd, &xcs);
    return xcs.width;
}

int WTextHeight(WIND *wind, char *str)
{
    XCharStruct xcs;
    int         nc, dir, fa, fd;

    nc = strlen(str);
    XQueryTextExtents(wind->disp, XGContextFromGC(wind->gc),
		      str, nc, &dir, &fa, &fd, &xcs);
    return fa+fd;
}

void WWriteIn(wind, pr, str, lcrAlign)	/* write string in rectangle */
    WIND    *wind;
    RECT    *pr;
    char    *str;
    int     lcrAlign;	/* -1 -> left align; 0 -> ctr, 1 -> right */
    {
    int     strwidth;
    int     strheight;
    int     acty;

    strwidth = WTextWidth(wind, str);
    strheight = WTextHeight(wind, str);
    /* always centrally aligned in vert */
    acty = pr->y+(pr->height - strheight)/2;	/* acty is top of text */
    if(lcrAlign == LALIGN)		/* left align */
	WWriteAt(wind, pr->x, acty, str);
    else if(lcrAlign == RALIGN)	/* right align */
	WWriteAt(wind, pr->x + pr->width - strwidth, acty, str);
    else			/* CALIGN i.e. centred */
	WWriteAt(wind, pr->x + (pr->width - strwidth)/2, acty, str);
    }

#define     TICKSPC     50	/* ideal # pixels between ticks */
#define     HTICKSPC    70	/* ideal # pixels between horiz ticks */
#define     VTICKSPC    20	/* ideal # pixels between vert ticks */
#define     TKPX	2	/* pixels for the tick mark */
#define     VOFFS       3	/* Border to offset text from ticks */
#define     HOFFS       5	/* Border to offset text from ticks */
#define     LINEH	12	/* Guess at character height */

void WAddTimeLine(wind,pr,t1,t2,vnh,sns,und,tix,pntx) 
    /* write linear scale into a box */
    WIND    *wind;
    RECT    *pr;	    /* where to put it (& how long it is) */
    FLOATARG   t1,t2;	    /* start and end times of box */
    int     vnh;	    /* 0 => horiz timeline, 1 => vert timeline */
    int	    sns;	    /* 1 => low to high; 0 => hi to low */
    int     und;	    /* flag to put ticks : 1= top of box, 0 = bottom */
    int	    tix[];	    /* array of indices which are labelled */
    int	    *pntx;
    {
    float   dt;
    short   dx,xb,xi,tkh,guardx,tkspc;
    float   xPerT, tPerTk, xPerTk;
    int     tkPerIx, tkix;
    long    n0;
    float   nrst, pwr, subdiv;
    float   t0, x0;
    int     dp,nc;
    char    s[32];
    RECT    trec;
    float   rescale = 1.0;
    char    rescCh = '.';
    int     cwidth = WTextWidth(wind, " ");
    int     lineh  = WTextHeight(wind, "H");

    trec  = *pr;
    *pntx = 0;

    dt = t2-t1;
    if(dt==0.0) return;
    if(dt<0)	dt = -dt;   /* args in wrong order ? */
    /* choose SI range for units & labels */
    if(dt>5000)                       { rescale = 0.001;      rescCh = 'k'; }
    else if(dt>5000000)               { rescale = 0.000001;   rescCh = 'M'; }
    else if(dt<0.5 && t2 <1.0)        { rescale = 1000.0;     rescCh = 'm'; }
    else if(dt<0.0005 && t2 <0.001)   { rescale = 1000000.0;  rescCh = 'u'; }
    
    if(vnh)
	{
	dx = pr->height;
	xb = pr->y;
	trec.height = 0;
	trec.width -= cwidth;
	trec.x += cwidth/2;
	tkspc = 1.5*lineh;
	guardx = tkspc/2;
	}
    else{
	dx = pr->width;
	xb = pr->x;
	trec.width = 0;
	trec.height -= 2*VOFFS;
	trec.y += VOFFS + (und?0:-(TKPX-1));
	tkspc = WTextWidth(wind, "0.0000   ");
	guardx = WTextWidth(wind, "0.0000");
	}
    xPerT = ((float)dx)/dt;
    tPerTk  = ((float)tkspc)/xPerT;	/* ideal time spacing */
    /* find 1, 2 or 5 x 10^n closest to tPerTk */
    tPerTk = WNearestRoundNum(tPerTk, &subdiv);
    tkPerIx = (int)(.5+tPerTk/subdiv);	/* had a rounding problem here */
    dp = WDecPoints(tPerTk*rescale);
    tPerTk = subdiv;	/* actually refer to period of ticks */
    xPerTk = tPerTk * xPerT;
    n0 = (long)(.99+(t1/tPerTk)); 
    t0 = tPerTk*(float)n0;         /* earliest time exact multiple of tPerTk */
    tkix = n0%tkPerIx;

    x0 = (t0-t1)*xPerT;
    while(x0 < dx)
	{
	if(sns)
	    xi = xb+(short)x0;
	else
	    xi = xb + dx - (short)x0;
	tkh = TKPX;
	if(tkix==0 && x0<dx-guardx)
	    {
	    FltToString(t0*rescale,s,dp);
	    if(rescale != 1.0)
		WReplaceOrAppend(s, '.', rescCh); /* replc dp with e.g. 'k' */
	    if(vnh)
		{  trec.y = xi-lineh/2; 
		   WWriteIn(wind, &trec, s, RALIGN);	}
	    else
		{  trec.x = xi;
		   WWriteIn(wind, &trec, s, LALIGN); }

	    tkh = 2*TKPX;
	    *tix++ = xi;	/* remember where we put these ticks */
	    ++(*pntx);
	    }
	if(vnh)
	    {  WMoveTo(wind, und?pr->x:pr->x+pr->width, xi);
	       WLineTo(wind, und?(pr->x+tkh):(pr->x+pr->width-tkh), xi);  }
	else
	    {  WMoveTo(wind, xi, und?pr->y:pr->y+pr->height);
	       WLineTo(wind, xi, und?(pr->y+tkh):(pr->y+pr->height-tkh));  }
	t0 += tPerTk;
	x0 += xPerTk;
	tkix = (tkix +1 )%tkPerIx;
	}
    }

static float WNearestRoundNum(num, sub)
    FLOATARG	num;
    float	*sub;	/* return next size down too as sub-step */
    {
static int	possVals = 3;
static float    rndNums[3][2] = {{ .707, 1.0 },{ 1.414, 2.0 },{ 3.163, 5.0}};
static float	oom = 10.0;	/* round number configuration */
    float	scale = 1.0;
    float	ref;
    int 	i;

    ref = rndNums[0][0];	/* lower bound for lowest round number */
    if(num < 0)		{ num *= -1.0; 	scale *= -1.0; }
    while(num < ref)
	{
	num *= oom;	/* next order of magnitude */
	scale /= oom;	/* keep track of order of magnitude */
	}
    while(num >= oom*ref)
	{
	num /= oom;	/* next order of magnitude */
	scale *= oom;	/* keep track of order of magnitude */
	}
    /* new num is in range ref <= num < oom*ref */
    /* quantize.. */
    i = possVals - 1;
    while(num < rndNums[i][0] && i>0)
	--i;
    num = scale * rndNums[i][1];
    if(i < 2)
	*sub = scale * rndNums[i-2+possVals][1] / oom;
    else
	*sub = scale * rndNums[i-2][1];
    return num;
    }

static void WReplaceOrAppend(s, c1, c2)
    char *s;		/* target string */
    CHARARG c1;		/* char to find */
    CHARARG c2;		/* char to replace it with - or stick on end */
    {
    char *t;

    if( (t = strchr(s, c1)) == NULL)
	{ t = s + strlen(s);	*(t+1) = '\0';	} /* hope for space */
    *t = c2;
    }

static int WDecPoints(step) /* return num of decimal points to code a step */
    FLOATARG step;
    {
    float ref = 1.0;
    int   pts = 0;

    if(step<0) step = -step;
    if(step==0)	return 0;
    while(ref>step)
	{ ref /= 10.0; ++pts;  }
    return pts;
    }

void WAddLogTimeLine(wind,pr,t1,t2,vnh,sns,und,tix,pntx) 
    /* write logarithmic scale into a box */
    WIND    *wind;
    RECT    *pr;	    /* where to put it (& how long it is) */
    FLOATARG   t1,t2;	    /* start and end times of box */
    int     vnh;	    /* 0 => horiz timeline, 1 => vert timeline */
    int	    sns;	    /* 1 => low to high; 0 => hi to low */
    int     und;	    /* flag to put ticks : 1= top of box, 0 = bottom */
    int	    tix[];	    /* array of indices which are labelled */
    int	    *pntx;
    {
    float   dt, pixScale, idealStep;
    short   dx,xb,xi,tkh;
    float   t0, nextt0, nt0, nextdt, x0, tkdt, nexttkdt;
    int     dp,nc,tkspc;
    char    s[32];
    int     cwidth = WTextWidth(wind, " ");
    int     lineh  = WTextHeight(wind, "H");

    RECT   trec;

    trec = *pr;
    *pntx = 0;

    if(t1*t2 <= 0.0)	{ fprintf(stderr,"LogLine: invalid %f..%f\n", t1,t2);
			  abort(); }
    dt = t2-t1;
    if(dt==0.0) return;
/*    if(dt<0)	dt = -dt;   /* args in wrong order ? */
    if(vnh)
	{
	dx = pr->height - 1;
	xb = pr->y+1;
	trec.height = 0;
	trec.x += cwidth/2;	trec.width -= cwidth;
	tkspc = 1.5*lineh;
	}
    else{
	dx = pr->width - 1;
	xb = pr->x + 1;
	trec.width = 0;
	trec.y += VOFFS + (und?0:-(TKPX-1));
	trec.height -= 2*VOFFS + (und?0:-(TKPX-1));
	tkspc = WTextWidth(wind, "0.0000   ");
	}
    pixScale = log(t2/t1)/dx;	/* so t(x) = t1 * exp(pixscale*(x-left)) */
    idealStep = exp(pixScale*tkspc);
    if(idealStep>2.0)	idealStep = 2.0;	/* else lose orders of mag */
    /* 2.0 is the smallest ratio between adjacent 'RoundNumbers' - I know */
/*    fprintf(stderr,"idealStep %f\n",idealStep);	*/

    dt = WNearestRoundNum(t1 * (idealStep-1.0), &tkdt);
    t0 = dt * MAX(1, (int)(.0 + t1/dt));	/* hacked- blocking 1993apr */
    dt = WNearestRoundNum(t0 * (idealStep-1.0), &tkdt);
    t0 = dt * (int)(.5 + t0/dt);

    while( (x0 = log(t0/t1)/pixScale) < dx)
	{
/*	fprintf(stderr, "t0 %.0f dt %.0f tkdt %.0f\n", t0, dt, tkdt);	/* */
/*	x0 = log(t0/t1)/pixScale; */
	if(x0>=0)
	    {
	    if(sns)
		xi = xb+(short)x0;
	    else
		xi = xb + dx - (short)x0;
	    dp = WDecPoints(dt);
	    FltToString(t0,s,dp);
	    if(vnh)
		{
		trec.y = xi /* - lineh/2 */;  /* dpwe 1993mar14: need lower */
		WWriteIn(wind, &trec, s, RALIGN);
		}
	    else{
		trec.x = xi;
		WWriteIn(wind, &trec, s, LALIGN);
		}

	    tkh = 2*TKPX;
	    *tix++ = xi;	/* remember where we put these ticks */
	    ++(*pntx);
	    if(vnh)
		{
		WMoveTo(wind, und?pr->x:pr->x+pr->width, xi);
		WLineTo(wind, und?(pr->x+tkh):(pr->x+pr->width-tkh), xi);
		}
	    else
		{
		WMoveTo(wind, xi, und?pr->y:pr->y+pr->height);
		WLineTo(wind, xi, und?(pr->y+tkh):(pr->y+pr->height-tkh));
		}
	    }
	nextt0 = t0+dt;
	nextdt = WNearestRoundNum(nextt0 * (idealStep-1.0), &nexttkdt);
	nextt0 = dt * (int)(.5 + nextt0/dt);
	nextdt = WNearestRoundNum(nextt0 * (idealStep-1.0), &nexttkdt);
	nt0    = nextdt * (int)(.51 + nextt0/nextdt);
	/* had to make it .51 to make .35 round to .4 when 350 made 400 */ 
	if(nt0 > nextt0)	nextt0 = nt0;	/* poke & hope for 1, 2, 5 */
	tkh = TKPX;
	while(((t0 += tkdt) < (.99*nextt0)) && (x0 = log(t0/t1)/pixScale)<dx )
	    if(x0>0) 
		{
		if(sns)
		    xi = xb+(short)x0;
		else
		    xi = xb + dx - (short)x0;
		if(vnh)
		    {
		    WMoveTo(wind, und?pr->x:pr->x+pr->width, xi);
		    WLineTo(wind, und?(pr->x+tkh):(pr->x+pr->width-tkh), xi);
		    }
		else
		    {
		    WMoveTo(wind, xi, und?pr->y:pr->y+pr->height);
		    WLineTo(wind, xi, und?(pr->y+tkh):(pr->y+pr->height-tkh));
		    }
		}
	t0 = nextt0;
	dt = nextdt;
	tkdt = nexttkdt;
	}
    }

void FltToString(f,s,dp)	/* convert a float into an ascii string */
	FLOATARG	f;
        char	*s;
	int 	dp;	        /* number of decimal places */
	{
	long	l;
	char	t[32];
	float	k = 1.0;
	int 	i;
	float	rup = 0.5;		/* roundup */

	for(i = dp; i>0; --i, rup /= 10.0)
		;	        /* form rup as 1/2 smallest digit (per DP) */
	if(f > 0)
	    f += rup;		/* Add on, so truncation gives right result */
	else 			/* (shift other way to round -ve nums) */
	    f -= rup;
	l = (long)f;
	NmToString(l,s);		/* convert int part */
	if(dp > 0)
		{
		for(i=0; i++ < dp; k *= 10 )	/* K scales for dp */
			;
		l = (long)(k*(1.0+f - (float)l));
		NmToString(l,t);
		t[0] = '.';	      /* overwrite place holding 1, add dp */
		strcat(s,t);
		}
	}

void NmToString(n,s)	/* was itoa from K&R via sndf.c */
    long n;
    char *s;
    {
    int     i,j;
    long    sign;
    char    c;

    if( (sign = n) < 0)     n = -n;
    i = 0;
    do	{
	s[i++] = n%10 + '0';
	} while( (n /= 10) > 0);
    if(sign < 0)
	s[i++] = '-';
    s[i--] = '\0';
    for(j=0; j<i; ++j,--i)	/* reverse the string */
	{ c = s[i]; s[i] = s[j]; s[j] = c; }
    }

void SetRct(RECT *pr, int x, int y, int width, int height)
{
    pr->x = x; pr->y = y; pr->width = width; pr->height = height;
}

void InsetRct(RECT *pr, int xin, int yin)
{
    pr->x += xin;	pr->width -= 2*xin;
    pr->y += yin;	pr->height -= 2*yin;
}

void OffsetRct(RECT *pr, int xoff, int yoff)
{
    pr->x += xoff;
    pr->y += yoff;
}

/* straight from macalike.c */
static XColor dummy;
static XColor blackCol = {0,0,0,0,0,0}, whiteCol = {0,0,0,0,0,0};
static XColor redCol = {0,0,0,0,0,0}, greenCol = {0,0,0,0,0,0};
static XColor blueCol = {0,0,0,0,0,0}, magentaCol = {0,0,0,0,0,0};
static XColor yellowCol = {0,0,0,0,0,0}, cyanCol = {0,0,0,0,0,0};
/* order must match #define BLACK etc in dspBench.h */
/* static XColor *colours[WGMD_NCOLORS];  */ /* at top */

static void SetupColours(Display *disp, int scrnNum)
    {
    if(redCol.pixel == 0)  {
	XAllocNamedColor(disp, DefaultColormap(disp, scrnNum), "red",
			 &redCol, &dummy);
	XAllocNamedColor(disp, DefaultColormap(disp, scrnNum), "green3",
			 &greenCol, &dummy);
	XAllocNamedColor(disp, DefaultColormap(disp, scrnNum), "royalblue1",
			 &blueCol, &dummy);
	XAllocNamedColor(disp, DefaultColormap(disp, scrnNum), "magenta",
			 &magentaCol, &dummy);
	XAllocNamedColor(disp, DefaultColormap(disp, scrnNum), "cyan",
			 &cyanCol, &dummy);
	XAllocNamedColor(disp, DefaultColormap(disp, scrnNum), "yellow",
			 &yellowCol, &dummy);
	XAllocNamedColor(disp, DefaultColormap(disp, scrnNum), "black",
			 &blackCol, &dummy);
	XAllocNamedColor(disp, DefaultColormap(disp, scrnNum), "white",
			 &whiteCol, &dummy);
	colours[WGMD_BLACK]  = &blackCol;
	colours[WGMD_WHITE]  = &whiteCol;
	colours[WGMD_RED]    = &redCol;
	colours[WGMD_GREEN]  = &greenCol;
	colours[WGMD_BLUE]   = &blueCol;
	colours[WGMD_YELLOW] = &yellowCol;
	colours[WGMD_MAGENTA]= &magentaCol;
	colours[WGMD_CYAN]   = &cyanCol;
	}
    }

/**** from db.xmain.c, via cqgram.c - (pseudo) colormap setup code ****/
static unsigned long 	*colrs;
static int		grays;

typedef struct {	/* one anchor point in RGB space */
    int red;
    int green;
    int blue;
} RGB_TRIPLE;

static int SetColorCells(Display *display, int screen, 
			 unsigned long gray[], int grays, 
			 RGB_TRIPLE *rgb1, RGB_TRIPLE *rgb2)
{	/* fill in a range of color cells linearly between rgb pairs */
    Colormap color_map;
    XColor exact_def,col_def;

    unsigned long plane_masks[1];
    unsigned long i;
    long	su = 256/grays;
    int 	r0, g0, b0;
    int 	dr, dg, db;

    r0 = rgb1->red   << 8;	dr = rgb2->red   - rgb1->red;	
    g0 = rgb1->green << 8;	dg = rgb2->green - rgb1->green;	
    b0 = rgb1->blue  << 8;	db = rgb2->blue  - rgb1->blue;	

    color_map = DefaultColormap(display, screen);
    if(XAllocColorCells(display,color_map,	
			False,plane_masks,0,gray,grays)==0) {
	fprintf(stderr,"SetColorCells fail (XAllocColorCells)\n");
	return 0;
    }
    else {
	for(i=0; i<grays; ++i)  {
	    col_def.pixel = gray[i];
	    col_def.red =   r0+(dr*i*su);
	    col_def.green = g0+(dg*i*su);
	    col_def.blue =  b0+(db*i*su);
	    col_def.flags = 7;
	    XStoreColor(display,color_map,&col_def);
	}
    }
    XFlush(display);
    return 1;
}

/* ------------------------- colormap code from dspB --------------------- */

#define MAXGRAYS 128
#define GRAYS 64		/* <= MAXGRAYS */

static int graycolors = 0;	/* control flags - gray not pseudo color */
static int bicolore   = 0;	/* - one bit rather than many */

typedef struct {	/* an anchor point plus how many steps before next */
    RGB_TRIPLE 	rgb;
    int		npts;
} RGB_SEGMENT;

#define RGB_POINTS	8
RGB_SEGMENT rgb_table[RGB_POINTS + 1] = 
{  { {   0,   0,   0 }, 8 },
   { {  16,   8,  96 }, 8 },
   { {  32,  16, 160 }, 8 }, 
   { { 128,  32, 144 }, 8 }, 
   { { 192,  32, 100 }, 8 }, 
   { { 224, 120,  82 }, 8 }, 
   { { 255, 208,  64 }, 8 },
   { { 255, 232, 160 }, 8 },
   { { 255, 255, 255 }, 0 } };


#define BW_POINTS	1
RGB_SEGMENT bw_table[BW_POINTS + 1] = 
{  { { 255, 255, 255 }, 32 }, 
   { {   0,   0,   0 }, 0 } };

/*
#define RGB_POINTS	4	/* how many line segments in our table * /
RGB_SEGMENT rgb_table[RGB_POINTS + 1] = 
{  { {   0,   0,   0 }, 8 }, 
   { {  32,  32, 255 }, 8 }, 
   { {  32, 128,  32 }, 8 }, 
   { { 220, 200,  96 }, 8 },
   { { 255, 255, 192 }, 0 } };	/* N+1 points for N segs * /   
 */

static unsigned long *def_gray(display,screen,red,green,blue)
    Display *display; 
    int screen,red,green,blue;
{
    static unsigned long ucolors[GRAYS];
    int	   depth, i;
    RGB_SEGMENT *table;
    int	   tabsize;
    int	   ok;

    if(graycolors)  {	/* monochrome (grayscale) colormap */
	table = bw_table;
	tabsize = BW_POINTS;
    } else {
	table = rgb_table;
	tabsize = RGB_POINTS;
    }

    depth = DisplayPlanes(display,screen);
    /* <bicolore> is command line '-b' flag to use just 2 colors */
    if(bicolore || depth < 3)	/* less than eight colours - use B&W */
	{
	grays = 2;
	ucolors[0] = XBlackPixel(display,screen);
	ucolors[1] = XWhitePixel(display,screen);
	return(ucolors);
	}
    else
	{
	grays = 1<<(depth - 1);		/* up to half of the avai grays.. */
	if(grays > GRAYS)	grays = GRAYS;	/* but at most GRAYS */
/*	SetColorCells(display, screen, ucolors, grays, 
		      0, 0, 0, red, green, blue);    */
	/* try a funky color map */
	grays = 0;
	for(i = 0; i<tabsize; ++i)  {
	    ok = SetColorCells(display, screen, &ucolors[grays], 
			       table[i].npts, &table[i].rgb, &table[i+1].rgb);
	    grays += table[i].npts;
	} 
	if(ok)
	    return ucolors;
	else 
	    return NULL;
    }
}

/* ----------------- end of colormap code from dspB --------------------- */

static unsigned long *cmap = NULL;

int WInitColormap(int gray, int onebit)
{	/* setup the pseudocolor colormap.
	   If <gray>, make it true shades of gray, not pseudocolor.
	   If <onebit>, reduce to a single color. 
	   Return zero on failure (typically run out of colormap space). */
    bicolore = onebit;
    graycolors = gray;	/* (uglily controlled by statics) */
    if(xdisp == NULL)
	assert(WInitX());
    if(cmap == NULL)
	cmap = def_gray(xdisp, xscrn, 255, 255, 255);
    if(cmap == NULL)
	return 0;
    else
	return 1;
}

int WGetPixelTable(unsigned long **pixels, int *ngrays)
{	/* retrieve the size and address of the (pseudo) colormap.
	   Returns zero if no colormap allocated yet. */
    if(cmap)  {
	*ngrays = grays;
	*pixels = cmap;
	return 1;
    } else {
	return 0;
    }
}

unsigned long WPixelForByte(int byte)
{	/* Map an unsiged byte (0..255) into a pixel value suitable 
	   for.. well.. XSetColor.. um.  This is inefficient for 
	   bulk use, but serves as an example.  */
    assert(cmap);
    if(byte > 255) 	byte = 255;
    if(byte < 0)	byte = 0;
    return cmap[(grays * byte)/256];
}

/******** dpwe 1993sep17 *********/
/* Take a byte array and draw it 'just so' */

#define MAXRECTS	128		/* maximum number of rects we buffer up */

static void WColorOneColor(WIND *wind, RECT *pr, int cols, int rows, 
						   BYTE *data, BYTE color)
{	/* go through the data finding all cels of <color>, draw them all */
	XRectangle rects[MAXRECTS], *ptr;
	int nrects;
	int i, j;

	XSetForeground(wind->disp, wind->gc, cmap[color*grays/256]);
	ptr = rects;	nrects = 0;
	for(i = 0; i<rows; ++i)  {
		for(j = 0; j<cols; ++j)  {
			if(*data++ == color)  {
				ptr->x = pr->x + j * pr->width / cols;
				ptr->y = pr->y + i * pr->height / rows;
				ptr->width  = pr->x + (j+1) * pr->width / cols - ptr->x;
				ptr->height = pr->y + (i+1) * pr->height / rows - ptr->y;
				++ptr;	++nrects;
				if(nrects == MAXRECTS)  {
					XFillRectangles(wind->disp, wind->win, wind->gc, 
									rects, nrects);
					ptr = rects;	nrects = 0;
				}
			}
		}
	}
	if(nrects)  {
		XFillRectangles(wind->disp, wind->win, wind->gc, 
						rects, nrects);
	}
}

#include <arrayops.h>

void WColorArray(WIND *wind, RECT *pr, ByArray *array)
{	/* Convert the array <*data> (<rows> rows of <cols> bytes) into 
	   a pseudocolored region filling <*pr>, using the prevailing colormap. 
	   ACTUALLY QUANTIZES THE MATRIX IN-PLACE TO THE NUMBER OF AVAI GRAYS */
	int x, y, w, h;
	int i, j, c;
	int donecolors[MAXGRAYS];	/* flag when each level done */
	int ncolors;
	int cols = array->cols;
	int rows = array->rows;
	BYTE *data = array->data;
	BYTE *dcopy = data;

	if( wind->opf != NULL )		/* (write to ps file too) */
	    idRectRleGraymap(wind, pr, array);

	for(i = 0; i<grays; ++i)
		donecolors[i] = 0;
	ncolors = 0;
	/* first pass counts the number of diff colors and quantizes input */
	for(i = 0; i<rows; ++i)  {
		for(j= 0; j<cols; ++j)  {
			c = *dcopy * grays/256;
			*dcopy++ = c * 256 / grays;		/* write back quantized */
			if(donecolors[c] == 0)
				++ncolors;		/* tally number of distinct colors */
			++donecolors[c];	/* count instances of this color */
		}
	}
	/* second pass draws up elements of each color */
	i = 0;
	while(ncolors && i<grays)  {
		if(donecolors[i] > 0)  {
			WColorOneColor(wind, pr, cols, rows, data, i * 256 / grays);
			--ncolors;
		}
		++i;
	}
}

/**** end of pseudocolormap code *************************/

int RangeOverlap(int l1, int w1, int l2, int w2)
{    /* is there overlap between ranges l1..l1+w1 and l2..l2+w2 ? */
    int u1 = l1+w1;
    int u2 = l2+w2;

    return !((u1<l2) || (u2<l1));
}

int PtInRct(int x, int y, RECT *pr)
{
/*    if( (x >= pr->x) && (x < pr->x + pr->width)
       && (y >= pr->y) && (y < pr->y + pr->height)) */
    if(RangeOverlap(pr->x, pr->width, x, 0)
       && RangeOverlap(pr->y, pr->height, y, 0))
	return(1);
    else
	return(0);
    }

int RectOverlap(RECT *pr1, RECT *pr2)
{	/* rects do not overlap if *either* of their axes do not */
    if(RangeOverlap(pr1->x, pr1->width, pr2->x, pr2->width)
       && RangeOverlap(pr1->y, pr1->height, pr2->y, pr2->height))
	return(1);
    else
	return(0);
}

static EVHDLR_NODE *rootEvhn = NULL;	/* freestanding event handlers */

static WIND *windFindWin(Window win)
{
    WIND *rslt = windRoot;
    int found = 0;

    while(!found && rslt)  {
	if(rslt->win == win)
	    found = 1;
	else
	    rslt = rslt->next;
    }
    return rslt;
}

static int windApplyEvhs(WIND *wndw, XEvent *event)
{	/* apply the event handlers on this window */
    EVHDLR_NODE *evhn;
    int taken = 0;

    if(wndw)	
	evhn = wndw->evhs;
    else
	evhn = rootEvhn;	/* applyEvhs on NULL means `other' events */
    while(!taken && evhn)  {
	taken = (*(evhn->evhFn))(event, evhn->evhData);
	evhn = evhn->next;
    }
    return taken;
}

#define MYX_ANY_EVENT (-1L)

int WEventHandler(XEvent *event, void *data)
{	/* just try and handle a random XEvent */
    int rc = 0;
    WIND *wndw;
    wndw = windFindWin(event->xany.window);
    /* if not our window, wndw returns NULL, 
       which is OK, because wndwApplyEv uses `other' ev handlers 
       if called with NULL */
    if( (rc = windApplyEvhs(wndw, event)) == 0)
	DBGFPRINTF((stderr, 
		    "WEventHdlr: event not handled on wndw %lx\n", wndw));
    return rc;
}

static int pauseEvents = FALSE;		/* global pausing */

int WSetEventPause(int newPause)
{	/* set the event loop to return control or not */
    int	oldPause = pauseEvents;

    pauseEvents = newPause;
    return oldPause;
}

int WToggleEventPause(void)
{	/* just pause <=> no pause */
    return WSetEventPause(!pauseEvents);
}

int WEventLoop(void)
{	/* handle another event; returns after each event, 
	   you should call it again. */
    XEvent	event;
    int 	rc = 1;		/* if returns straight away, that's OK */

    do  {
/*	while(XCheckMaskEvent(xdisp, MYX_ANY_EVENT, &event)) {	*/
	if(1)  {
	    XNextEvent(xdisp, &event);
	    WEventHandler(&event, NULL);
	}
    } while(pauseEvents);
    return rc;
}

int WEventLoop2(void)
{	/* handle another event; returns after each event, 
	   you should call it again. */
    XEvent	event;
    int 	rc = 1;		/* if returns straight away, that's OK */

    do  {
	while(XCheckMaskEvent(xdisp, MYX_ANY_EVENT, &event)) {
/*	if(1)  {	*/
	    XNextEvent(xdisp, &event);
	    WEventHandler(&event, NULL);
	}
    } while(pauseEvents);
    return rc;
}

void WAddEventHandler(WIND *wndw, eventHandlerFn evhFn, void *evhData)
{      	/* add an event handler for this window */
    EVHDLR_NODE *evhn;
    EVHDLR_NODE **pparevhn;

    if(wndw)	
	pparevhn = &wndw->evhs;
    else
	pparevhn = &rootEvhn;	/* EventHandler for NULL gets `unrec' events */
    evhn = TMMALLOC(EVHDLR_NODE, 1, "WAddEventHandler");
    evhn->evhFn = evhFn;
    evhn->evhData = evhData;
    /* splice into some event handler list */
    evhn->next = *pparevhn;
    *pparevhn = evhn;
}

int WGetShiftEtc(WIND *wndw)
{   /* return a bit field including the bottom bit as the shift key */
    Window	root, child;
    int 	rx, ry, wx, wy;
    unsigned int keys;

    XQueryPointer(wndw->disp, wndw->win, &root, &child, &rx, &ry, 
		  &wx, &wy, &keys);
    return ((keys & ShiftMask)!= 0);
}
