/************************************************************************/
/*  stfta.c	    (after pva.c)					*/
/*  Frequency Domain Analysis						*/
/*  Takes a time domain soundfile and converts it into a file of STFT	*/
/*  This version just calculates pure FFT rather than PVOC's phi-dot	*/
/*  More importantly, it reads from disk rather than using so much core */
/*  dpwe  14feb91							*/
/************************************************************************/

/* #define DBGWDWS	1	/* graphic debug */

#include <stdio.h>

#ifdef THINK_C
#include <tc_main.h>
#include <stdlib.h>
char sfdir_path[] = "Frida:dpwe:sound:";
#else
#include <malloc.h>
#endif

#include <fft.h>
#include <dsputil.h>
#include <sndf.h>
#include <pvoc.h>
#ifdef DBGWDWS
#include <window.h>
#endif

/*** Static function prototypes ***/
#ifdef __STDC__			/* prototype arguments */
static FILE *CreatePVfile(char *pvPath, PVSTRUCT **oPVH, SFSTRUCT *iSnd);
static void takeFFTs(SFSTRUCT *inputSound, PVSTRUCT  *outputPVH,
		     FILE *sndFile, FILE *ftFile);
static void PrintBuf(float *buf, int size, char *msg);
static void convertToMono(SFSTRUCT **theSound);
static void CopyShorts(short *sce, int sinc, short *dst, int dinc, 
		       long len, int fwdNbck);
#else				/* no argument prototypes */
static FILE *CreatePVfile();
static void takeFFTs();
static void PrintBuf();
static void convertToMono();
static void CopyShorts();
#endif

#define MINFRMMS	20	/* frame defaults to at least this many ms */
#define MAXFRMPTS	65536
#define MINFRMPTS	16	/* limits on fft size */
#define DFLT_OVLP	4	/* default frame overlap factor */
#define DFLT_EXTN	".aft"	/* default output file extension */

/* Global variables */
int	    sfErr;
char	    *programName;
int	    debug = 0;		/* tweaked inside ! */
int	    pvoccy = 0;		/* flag to write phi-dot, not just phi */
float	    winTime = 0.0;	/* window duration */
int 	    winPts = 0;

/* Static global variables */
static	long frameSize	= 0;	    /* size of FFT frames */
static	long frameIncr	= 0;	    /* step between successive frames */

complex *basis;     /* LUTable for FFT */

main(argc,argv)
    int argc;
    char *argv[];
    {
    SFSTRUCT	*inputSound;
    PVSTRUCT	*outputPVH;
    FILE	*sndFile;
    FILE	*ftFile;
    char    *inputSoundFile = NULL;
    char    outputPvocFile[80];
    int     i,j,ovlp = 0;	/* number of overlapping windows to have */
    int     err = 0;
    double	atof();

    outputPvocFile[0] = '\0';
/*    float   a = 3.9;
    fprintf(stderr,"(int)(%5.2f) = %d (int)(%5.2f) = %d\n", 
	    -a, (int)(-a), a, (int)(a) );	*/
    /* Check arguments */
    programName = argv[0];
    frameIncr = frameSize/DFLT_OVLP /* ovlp must be zero to check if set */;
    for(i=1; i<argc && err==0; ++i)
	{
	if(argv[j=i][0]=='-')
	    switch(argv[i][1])
		{
	    case 'n':
		if(i<argc-1)
		    {
		    frameSize = atol(argv[++i]);
		    if(frameSize<MINFRMPTS || frameSize>MAXFRMPTS
		       || !IsPowerOfTwo(frameSize))
			{
			fprintf(stderr,"%s:n (%s) must be 2^r and %d..%d\n",
				programName, argv[i], MINFRMPTS, MAXFRMPTS);
			err = 1;
			}
		    }
		else
		    err = 1;
		break;
	    case 'o':
		if(frameIncr == 0 && i<argc-1)
		    ovlp = atol(argv[++i]);
		else
		    err = 1;
		break;
	    case 'i':
		if(ovlp == 0 && i<argc-1)
		    frameIncr = atoi(argv[++i]);
		else
		    err = 1;
		break;
	    case 'w':
		if(i<argc-1)
		    winTime = atof(argv[++i]);
		else
		    err = 1;
		break;
	    case 'p':
		pvoccy = 1;	break;
	    default:
		fprintf(stderr,"unrecog option: %s \n",argv[j]);
		err = 1;
		break;
		}
	else{
	    if(inputSoundFile==NULL)
		inputSoundFile = argv[i];
	    else
		strcpy(outputPvocFile, argv[i]);
	    }
	}
    if(err || inputSoundFile==NULL)
	{
	fprintf(stderr,
	"Usage: %s [-n n][-w t][-o n | -i n][-p] <inputSnd> [<outFFTfile>]\n",
		programName);
	fprintf(stderr,"where \n");
	fprintf(stderr," -w t   sets the window length (dflt 20ms)\n");
	fprintf(stderr," -n n   sets the fft size (power of 2) (dflt 20ms)\n");
	fprintf(stderr," -o f   sets the overlap factor (dflt 4)\n");
	fprintf(stderr," -i h   sets the hopsize (increment) (dflt n/4)\n");
	fprintf(stderr," -p     write a PVOC file (phi-dot) not pure STFT\n");
	fprintf(stderr," inputSound   is a SNDF-style soundfile\n");
	fprintf(stderr," outputFFT    is a PVOC-style file (dflt stem+'%s')\n",
		DFLT_EXTN);
	exit(1);
	}
    if(outputPvocFile[0] == '\0')
	SubstExtn(inputSoundFile, DFLT_EXTN, outputPvocFile);
    /* Open sound file */
    if( (sndFile = SFOpenAllocRdHdr(inputSoundFile, &inputSound))==NULL)
	{
	printf("%s: error opening %s as input soundfile\n", programName,
	       inputSoundFile);
	exit(1);
	}
    /* Abort if unsupported format */
    if(debug)
	printf("Read sound at %lx; magic = %lx, srate = %6.1f\n",
	       inputSound, inputSound->magic, inputSound->samplingRate);
    if (inputSound->format != SFMT_SHORT)
	{
	fprintf(stderr, "%s: unsupported sound format: %d\n",
		programName, inputSound->format);
	exit(1);
	}

    /* setup frame size etc according to sampling rate */
    if(winTime == 0.0)
	winTime = ((float)MINFRMMS)/1000;
    else
	{
	if(winTime<0.001 || winTime > 0.100)
	    {  fprintf(stderr, "%s: bad win time of %.3f must be .001 to .1\n",
		       programName, winTime);	exit(0);	}
	}
    winPts = inputSound->samplingRate * winTime;
    if( (winPts & 1)==0)	++winPts;	/* make it odd - kicks */

    if(frameSize==0)		/* not specified on command line */
	{
	/* default frame size is > MINFRMMS milliseconds */
	frameSize = MAXFRMPTS;
	while( (frameSize>>1) >= winPts && frameSize > MINFRMPTS)
	    frameSize >>= 1;	/* divide down until just larger */
	}
    if (ovlp==0 && frameIncr==0)
	{
	ovlp = DFLT_OVLP;	/* default overlap */
	frameIncr = winPts / ovlp;
	}
    else if(ovlp == 0)
	ovlp = winPts/frameIncr;
    else
	frameIncr = winPts/ovlp;

    if ( (ovlp < 2) || (ovlp > 16) )
	{
	fprintf(stderr, "%s: %d is a bad window overlap index\n",
		programName, ovlp);
	exit(1);
	}
    fprintf(stderr, "%s: frSiz %d wSiz %d hop %d\n", programName,
	   frameSize, winPts, frameIncr);
    if(frameSize < winPts)
	{
	fprintf(stderr, "%s: frameSize %d < winPts %d : stand by for crash\n",
		programName, frameSize, winPts);
	}

    basis = AssignBasis(NULL,frameSize); /* set up FFT tables */
    if( (ftFile=CreatePVfile(outputPvocFile,&outputPVH,inputSound))==NULL)
	{
	fprintf(stderr,"%s: error opening %s as output FT file\n",
		programName, outputPvocFile);
	exit(1);
	}

    takeFFTs(inputSound, outputPVH, sndFile, ftFile);

    /* Close files */
    PVCloseWrHdr(ftFile, outputPVH);
    fprintf(stderr, "%s: wrote %s\n", programName, outputPvocFile);
    fclose(sndFile);
    }

static FILE *CreatePVfile(pvPath,ppPVH,inputSound)
    char	    *pvPath;
    PVSTRUCT	    **ppPVH;
    SFSTRUCT	    *inputSound;
    {
    float   rate     = (float)inputSound->samplingRate;
    int     chans    = (int)inputSound->channels;
    int     dsize    = SFSampleSizeOf(inputSound->format);
    long    frameWords;
    int     err;
    FILE    *opFile;

    if(debug)
	printf("Framesiz %ld, framInc %ld\n",frameSize, frameIncr);
    /* If we dealt with frames that hit the ends properly, we'd have */
    /*	 (size/sizeof(short))/frameIncr frames (>= 1/2 inside file) */
    frameWords = frameSize + 2L; /* each frame has Mag & phase for n/2 +1 */
    if(inputSound->dataBsize != SF_UNK_LEN)
	printf("Est. frames %ld\n", 
	       inputSound->dataBsize/(chans*dsize*frameIncr));
    if(chans > 1)
	{
	printf("Sorry: using just 1st channel of %d-channel sound\n", chans);
/*	chans = 1;	/* will write # chans into PV header tho' only mono */
	}
    printf("Output: %d chan%s, %ld frsize, %ld frInc\n",
       chans, (chans==1)?"":"s", frameSize, frameIncr);
    /* Allocate a new PV block */
    if (err = PVAlloc(ppPVH, PV_UNK_LEN, PVFLOAT, rate, chans,
	      frameSize, frameIncr, sizeof(float)*2*(frameSize/2 + 1),
	      (pvoccy)?PVPVOC:PVPOLAR, 0.0, rate/2.0, PVLIN, 4))
    { fprintf(stderr, "%s: %s\n", programName, PVErrMsg(err));
      exit(1);	}
    opFile = PVOpenWrHdr(pvPath, *ppPVH);
    return opFile;
    }

/*
 * takeFFTs
 *  Go through the (mono) input sound frame by frame and find the
 *  magnitude and phase change for a string of FFT bins
 */
static void takeFFTs(inputSound,outputPVH, sndFile, pvFile)
    SFSTRUCT	    *inputSound;
    PVSTRUCT	    *outputPVH;
    FILE	    *sndFile;
    FILE	    *pvFile;
    {
    long    i,j;
    float   *tmpBuf, *oldInPh, *winBuf;
    float   sampRate = inputSound->samplingRate;
    int     chans    = inputSound->channels;
    int     not_done = 1;
    long    read_in  = -1;
    short   *shBuf;
    long    fsIndepVals = (frameSize/2L)+1L;
#ifdef DBGWDWS
    WINDAT  *windo1, *windo2;

    windo1 = NewWin("prerot",1);
    windo2 = NewWin("stfta",1);
#endif 

    shBuf   = (short *)MakeBuf(chans*frameSize);
    tmpBuf  = MakeBuf(frameSize * 2L);
    oldInPh = MakeBuf(frameSize);
    winBuf  = MakeHalfWin(winPts,(float)1.0,1);

    /* initially, clear first half of buffer .. */
    for(i = 0; i<(chans*frameSize/2); ++i)	shBuf[i] = 0;
    /* .. and read in second half from file */
    fread(shBuf+frameSize/2, sizeof(short), (chans*frameSize/2), sndFile);

    printf("frame: ");
    for (i = 0; not_done; i++)
	{
	if((i%20)==0)   printf("%ld ", i); fflush(stdout);
	
/*	FloatAndCopy(shBuf,tmpBuf,frameSize);	*/
	Short2Float(shBuf, chans, tmpBuf, 1, winPts);
	/*  PrintBuf(tmpBuf, frameSize, "floated"); /* */

	CopyShorts(shBuf+(frameIncr*chans),1,shBuf,1,
		   (winPts-frameIncr)*chans,0);
	if(read_in)		/* if it worked last time */
	    read_in = fread(shBuf+(winPts-frameIncr)*chans, sizeof(short), 
			    frameIncr*chans, sndFile);
	else
	    not_done = 0;	/* stop after this run-thru */
	/* clear end if not all read in */
	if(read_in < frameIncr)
	    for(j = winPts-frameIncr+read_in; j<winPts; ++j)
		shBuf[j] = 0;
	ApplyHalfWin(tmpBuf,winBuf,winPts /* frameSize */ );
#ifdef DBGWDWS
	DoDisp(windo1, tmpBuf, winPts);
#endif
	HalfRotate(tmpBuf, winPts);
	PadInMiddle(tmpBuf, winPts, frameSize);
	/*  PrintBuf(tmpBuf, frameSize, "windo'd"); /* */
#ifdef DBGWDWS
	DoDisp(windo2, tmpBuf, frameSize);
#endif
	UnpackReals(tmpBuf, frameSize);
	/*  PrintBuf(tmpBuf, frameSize, "unpacked"); /* */
	FFT2real((complex *)tmpBuf, frameSize, 1,basis);
	/*  PrintBuf(tmpBuf, frameSize, "fft'd"); /* */
	Rect2Polar(tmpBuf, fsIndepVals);
	/*  PrintBuf(tmpBuf, frameSize, "toPolar"); /* */
	if(pvoccy)
	    {
	    UnwrapPhase(tmpBuf, fsIndepVals, oldInPh); /* */
	    /*  PrintBuf(tmpBuf, frameSize, "unWrapped"); /* */
	    PhaseToFrq(tmpBuf, fsIndepVals, (float)frameIncr, sampRate); /* */
	    /*  PrintBuf(tmpBuf, frameSize, "toFrq"); /* */
	    }
	/* write straight out, just the indep vals */
	fwrite(tmpBuf, 2*sizeof(float), fsIndepVals, pvFile);
	/* debug = 0;	/* */
	}
    printf("%ld\n",i-1);
    }

#define DBGPTS 8
static void PrintBuf(buf,size,msg)
    float *buf;
    int   size;
    char  *msg;
    {
    int   i;

/*    if(!debug)
      return;
 */
    size = DBGPTS;
    printf("%s:",msg);
    for(i=0; i<size; ++i)
    printf("%7.2f ",buf[i]);
    printf("\n");
    }

/* Convert multi-channel sound to mono by tossing extra channels */
static 	void	convertToMono(theSound)
    SFSTRUCT  **theSound;
    {
    SFSTRUCT *newSound;
    short   *oldPtr, *newPtr;
    int     size     = (*theSound)->dataBsize;
    int     format   = (*theSound)->format;
    int     rate     = (*theSound)->samplingRate;
    int     channels = (*theSound)->channels;
    int     newSize, i, j;

    newSize = size/channels;
    if (sfErr = SFAlloc(&newSound, newSize, format, rate, 1, 4))
	{
	printf("%s: sound error: %s\n", programName,
		    SFErrMsg(sfErr));
	exit(1);
	}
    oldPtr  = (short *)SFDataLoc(*theSound);
    newPtr  = (short *)SFDataLoc(newSound);
    for (i = 0, j = 0; i < newSize/sizeof(short); i++, j+=channels)
	newPtr[i] = oldPtr[j];
    SFFree(*theSound);
    *theSound = newSound;
    }

static void CopyShorts(sce, sinc, dst, dinc, len, fwdNbck)
    short *sce;
    int sinc;
    short *dst;
    int dinc;
    long len;
    int fwdNbck;
    {
    if(fwdNbck)		/* fwdNbck => copying forward, so start with last */
	{
	sce += len * sinc;
	dst += len * dinc;
	while(len--)
	    {
	    dst -= dinc;	sce -= sinc;
	    *dst = *sce;
	    }
	}
    else		/* copying backward, so start with first */
	while(len--)
	    {
	    *dst = *sce;
	    dst += dinc;	sce += sinc;
	    }
    }

