/***************************************************************\
*	exspec.c		  				*
*   Generate a best-estimate spectrum for the next time instant	*
*   based on the current tracks					*
*   dpwe 07oct91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#include <dpwelib.h>
#include <math.h>
/* personal libs */
#include <genutils.h>	/* maskPhs */
#include <samplops.h>	/* ClearFloats */
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include "peaks.h"
#include "tracks.h"	/* for TRACK *  */
#include "exspec.h"

/* types */

/* constants */

/* static function prototype */

/* static data */

/* functions */

void AddCpxPolar(cp1, cp2)
    cpxpolar	*cp1;
    cpxpolar	*cp2;
    {
    float	re, im;

    re = cp1->mag*cos(cp1->pha);
    im = cp1->mag*sin(cp1->pha);
    re += cp2->mag*cos(cp2->pha);
    im += cp2->mag*sin(cp2->pha);
    cp1->mag = hypot(im, re);
    if(cp1->mag != 0.0)
	cp1->pha = atan2(im, re);
    else
	cp1->pha = 0.0;
#if !(__sgi || THINK_C)	/* no isnan on indigo or mac */
    if(isnan(cp1->mag) || isnan(cp1->pha))	abort();
#endif /* !__sgi||THINK_C) */
    }

void ExtrapTrack(trk, t0, peak)
    TRACK	*trk;		/* a current track */
    FLOATARG	t0;		/* time - presu. in future */
    PeakPos	*peak;		/* return predicted peak in here */
    {
    PeakPos	*botPk;

    /* could use fm & am if had them.  Current models are 0-order */
    botPk = trk->tkbot->peak;
    peak->tim = t0;
    peak->frq = botPk->frq;
    peak->mag = botPk->mag;
    peak->phs = maskPhs(botPk->phs + (t0 - botPk->tim)*2.0*PI*peak->frq);
    /* wind the phase forward to the appropriate time */
    }

void SuperposeHump(peak, spec, bins, octs, topfrq, Q)
    PeakPos	*peak;		/* return predicted peak in here */
    cpxpolar	*spec;		/* complex spectrum output array */
    int 	bins;		/* bins per oct for o/p spec */
    int 	octs;		/* total number of 8ves in o/p */
    FLOATARG	topfrq;		/* frq of highest bin */
    FLOATARG	Q;		/* Q factor of peaks to synthesise */
    {
    int 	totbins, i;
    float	thisFrq, df, fscale, mscale;
    cpxpolar	thisVal;

    /* 1992oct26 - *measured* best-fit Q is 19.0984 when 
       measured Q of these ones is 15.996 (when 'Q' = 16.0). 
       So for the while, we will just scale up Q */
    Q *= 1.19394;

    totbins = bins*octs;
    mscale = exp(peak->mag/LOGN2DB);	/* plus some fixup.. */
    /* expect the Q=16.0 to have 1/e^2 points at fc +/- fc/8 */
    /* i.e. exp(-Q*Q/2/peak->frq/peak->frq * df*df) */
    /* our reference point is that "Q=8.0" centred on 4000 Hz has 
       1/e (-8.686 dB) points at +/-750 Hz (3250, 4750) = fc*3/16 */
/*    fscale = -16.0*256.0/(Q*peak->frq*peak->frq*9.0);	*/
    /* see p77 of dpwe SM thesis.  Q is called k there */
    fscale = -Q*Q/2/peak->frq/peak->frq;
    thisVal.pha = peak->phs;
/*    fprintf(stderr,"Sup pk: f %7.1f m %7.4f topf %7.1f Q %7.4f\n",
	    peak->frq, mscale, topfrq, Q);	*/
    for(i=0; i<totbins; ++i)
	{
	thisFrq = pow(2.0, ((float)(i-totbins))/((float)bins)) * topfrq;
	df = peak->frq - thisFrq;
/*	thisVal.mag = exp(df*df*fscale)*mscale; 	*/
	/* Has to be relative to bw *AT THAT FRQ* */
	thisVal.mag = exp(-df*df*Q*Q/2/thisFrq/thisFrq)*mscale;
	AddCpxPolar(&spec[i], &thisVal);
	}
    }

void PrintCpxPolar(cpxp, n, s)
    cpxpolar	*cpxp;
    long 	n;
    char 	*s;
    {
    long	i;

    fprintf(stdout, "Cpxpolar: %s\n", s);
    for(i = 0; i<n; ++i)
	fprintf(stdout, "%4d:m %8.3f p %8.3f\n", 
		i, cpxp[i].mag, cpxp[i].pha);
    }

PKLIST ExtrapTrkList(TKLIST curTrax, FLOATARG t0)
{	/* take a LIST of TRACK 
	   extrapolate each botpk to time <t0>, make a new peak.
	   return a PKLIST of these new peaks 
	   (trk history??? maybe tack onto botpk?) */
    TRACK	*trk;
    PeakPos	*peak;
    PKLIST	root = NULL;
    PKLIST  	*plist;

    plist = &root;
    while(curTrax && (trk = (TRACK *)car(curTrax)))  { /* loop on track cpts */
	curTrax = cdr(curTrax);
	peak = NewPeak(0.0,0.0,0.0,t0);
	ExtrapTrack(trk, t0, peak);	    	   /* figure likely peak */
	plist = (PKLIST *)append_ptr((LIST *)plist, 
				     new_node(NULL, (LIST)peak, T_PEAK));
    }
    return root;
}

void PkListToSpec(PKLIST pks, cpxpolar *spec, int bins, int octs, 
		  FLOATARG topfrq, FLOATARG Q)
{	/* 'convolve' the peaks in the list with the assumed filter 
	   shape of our bank */
    int totbins;

    totbins = octs*bins;
    ClearFloats((float *)spec, 1, 2*totbins);	   /* set to zero first */
    while(pks)	{
	SuperposeHump(pks->peak, spec, bins, octs, topfrq, Q);
	/* add in peak centred on this peak */
	pks = pks->next;
    }
}
    
PKLIST ExpectSpect(curTrax, t0, spec, bins, octs, topfrq, Q)
    TKLIST	curTrax;	/* list of TRACK * */
    FLOATARG	t0;		/* time to calc spectra for */
    cpxpolar	*spec;		/* complex spectrum output array */
    int 	bins;		/* bins per oct for o/p spec */
    int 	octs;		/* total number of 8ves in o/p */
    FLOATARG	topfrq;		/* frq of highest bin */
    FLOATARG	Q;		/* Q factor of peaks to synthesise */
{
    PKLIST	pklist;

    pklist = ExtrapTrkList(curTrax, t0);
    PkListToSpec(pklist, spec, bins, octs, topfrq, Q);

    return pklist;
}
	

PKLIST oldExpectSpect(curTrax, t0, spec, bins, octs, topfrq, Q)
    TKLIST	curTrax;	/* list of TRACK * */
    FLOATARG	t0;		/* time to calc spectra for */
    cpxpolar	*spec;		/* complex spectrum output array */
    int 	bins;		/* bins per oct for o/p spec */
    int 	octs;		/* total number of 8ves in o/p */
    FLOATARG	topfrq;		/* frq of highest bin */
    FLOATARG	Q;		/* Q factor of peaks to synthesise */
    {
    int 	totbins;
    TRACK	*trk;
    PeakPos	*peak;		/* NOT a pointer ! */
    PKLIST	root = NULL;
    PKLIST  	*plist;

    plist = &root;
    totbins = octs*bins;
    ClearFloats((float *)spec, 1, 2*totbins);	   /* set to zero first */
    while(curTrax && (trk = (TRACK *)car(curTrax)))  /* loop on track cpts */
	{
	curTrax = cdr(curTrax);
	peak = NewPeak(0.0,0.0,0.0,t0);
	ExtrapTrack(trk, t0, peak);	    	   /* figure likely peak */
/*	if(fabs(peak->frq-2000)<5.0)	PrintPeak(peak, "EX");  */
	/* could maybe predict fm as well for phase, broadening */
	SuperposeHump(peak, spec, bins, octs, topfrq, Q);
	/* add in peak centred on this peak */
	plist = (PKLIST *)append_ptr((LIST *)plist, 
				     new_node(NULL, (LIST)peak, T_PEAK));
	}
/*    PrintCpxPolar(spec, (long)totbins, "espec");	*/
    return root;
    }

