/***************************************************************\
*	optpath.c		  				*
*   It's a little difficult to figure out which path to take	*
*   through the choices of phase at each data point.  This 	*
*   routine hopes to find the least-cost phase set, based on 	*
*   expected phase increment (from frequency) using dynamic 	*
*   programming type algorithm					*
*   dpwe 19sep91						*
\***************************************************************/

/* std libs */
#include "optpath.h"	/* includes <dpwelib.h> */
#include <math.h>
/* personal libs */
#include <genutils.h>	/* for maskPhs() */
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include "peaks.h"	/* includes <llists.h> */

/* constants */

/* types */
/* typedef struct 
/*     {
/*     LIST	track;
/*     LIST	phases;		/* list of phase indexes : 0 or 1 */
/*     float	cost;		/* overall cost of this path */
/*     int 	costLen;	/* number of nodes this path is defined for */
/*     } PATH;
/* PATH defined in optpath.h now */

/* static function prototype */
static void SwapPtrs PARG((char **ptr1, char **ptr2));
static float PhaseCostFn PARG((PeakPos *pk1, PeakPos *pk2, 
			       int phx1, int phx2));
static void AppendPathStep PARG((PATH *path, int phase, FLOATARG cost));
static void CopyPathPhases PARG((PATH *srcPath, PATH *dstPath));
static void SwapPhases PARG((PeakPos *peak));

/* static data */

/* functions */

/* How to store a possible path?
   We have the <track> passed in as a list
   We should keep a vector of choices at each stage as our hypothesis,
   as well as its total cost (sum(abs(phase-expected_phase))) 
 */

static void SwapPtrs(ptr1, ptr2)
    char **ptr1;
    char **ptr2;
    {
    char *tmpPtr;

    tmpPtr = *ptr1; *ptr1 = *ptr2; *ptr2 = tmpPtr;
    }

static float PhaseCostFn(pk1, pk2, phx1, phx2)
    PeakPos	*pk1;	/* successive peak descriptions */
    PeakPos	*pk2;
    int 	phx1;	/* corresponding choices; 0 => phs, 1 => ophs */
    int 	phx2;
    {	/* work out cost of this phase step */
    float	edph, ph1, ph2, cost;

    if(pk1 == NULL || pk2 == NULL)
	return 0;	/* can't calculate : return no cost (?) */
    edph = (pk1->frq + pk2->frq)*PI;		/* AVG(f1,f2)*radspercycle */
    edph = edph * (pk2->tim - pk1->tim);	/* dph = w * dt */
    ph1 = (phx1==0)?pk1->phs:pk1->ophs;
    ph2 = (phx2==0)?pk2->phs:pk2->ophs;
    cost = fabs(maskPhs(ph2 - (ph1 + edph)));
    return cost*cost;	/* squared error -> avoid a few big jumps, even-out */
    }

float CostOfPath(path)
    PATH	*path;
    {	/* calculate the total cost of the path defined in <path> */
    LIST 	trkNode;
    LIST	phsNode;
    int 	nodeCnt = 0;
    float 	cost = 0;

    /* we will use the <length> field in the path structure to avoid recalc */
    trkNode = (LIST)path->track->tktop;
    phsNode = path->phases;
    if(path->costLen)
	cost = path->cost;
    while(nodeCnt < path->costLen)
	{
	trkNode = cdr(trkNode);
	phsNode = cdr(phsNode);
	++nodeCnt;
	}	/* will get buserr if nodeCnt > size(lists) */
    while(cdr(phsNode))	/* phsNode had better be at least 1 longer than p->l */
	{
	cost += PhaseCostFn((PeakPos *)car(trkNode), 
			    (PeakPos *)car(cdr(trkNode)), 
			    (int)car(phsNode), (int)car(cdr(phsNode)) );
	trkNode = cdr(trkNode);
	phsNode = cdr(phsNode);
	++nodeCnt;
	}
    return cost;
    }

PATH *NewPath(track)
    TRACK	*track;
    {
    PATH	*newPath;

/*    newPath = (PATH *)malloc(sizeof(PATH)); 	*/
    newPath = TMMALLOC(PATH, 1, "NewPath"); 
    newPath->track  = track;
    newPath->phases = NULL;
    newPath->cost   = 0.0;
    newPath->costLen= 0;
    return(newPath);
    }

void FreePath(path)
    PATH	*path;
    {
    free_list(path->phases);		/* list of phase choice nodes */
    free(path);
    return;
    }

static void AppendPathStep(path, phase, cost)
    PATH	*path;
    int 	phase;
    FLOATARG	cost;
    {	/* add one more step onto the phase list */
    LIST	*lastPhsPtr;

    lastPhsPtr = find_last_ptr(&path->phases);
    *lastPhsPtr = new_node(NULL, (LIST)phase, T_INT);
    path->cost = cost;
    ++path->costLen;
    }

static void CopyPathPhases(srcPath, dstPath)	/* copy phase ixs in paths */
    PATH	*srcPath;
    PATH	*dstPath;
    {
    free_list(dstPath->phases);		/* dispose of existing path */
    dstPath->phases = copy_top(srcPath->phases);
    }

void PrintPath(path, name)
    PATH	*path;
    char	*name;
    {
    NODE	*php, *pkp;
    PeakPos	*peak;
    int 	i = 0, p;
    float	phs, ophs;

    fprintf(stderr, "%s\n", name);
    fprintf(stderr, "TrackLen %d, PhsLen %d, CostLen %d, cost %f\n",
	    top_size((LIST)path->track->tktop), top_size(path->phases), 
	    path->costLen, path->cost);
    php = path->phases;
    pkp = (LIST)path->track->tktop;
    while(php)
	{
	peak = (PeakPos *)car(pkp);
	p = (int)car(php);
	if(p == 0)
	    {
	    phs = peak->phs;	ophs = peak->ophs;
	    }
	else
	    {
	    phs = peak->ophs;	ophs = peak->phs;
	    }
	fprintf(stderr, "node %2d time %5.3f phs %d = %6.3f not %6.3f\n",
		i, peak->tim, p, phs, ophs);
	pkp = cdr(pkp);
	php = cdr(php);
	++i;
	}
    }

PATH *BuildBestPath(track)
    TRACK	*track;
    { /* return optimal set of phase choices */
    /* need to hold several possible hypotheses since at each point, we
       don't know which of the two possible phases will be the best 
       intermediate point.  However, we _do_ know the cheapest path to 
       both of these points, so only need (at most) two hypotheses live 
       at once */
    PATH	*paths[2];
    LIST	trkNode;
    PeakPos	*thisPk, *nextPk;
    float	costAA, costAB, costBA, costBB, costA, costB;
    int 	debug = 0;

    paths[0] = NewPath(track);
    paths[1] = NewPath(track);

    AppendPathStep(paths[0], 0, 0);	/* free first step in both paths */
    AppendPathStep(paths[1], 1, 0);

    trkNode = (LIST)track->tktop;
    while(cdr(trkNode))
	{
	thisPk = (PeakPos *)car(trkNode);
	nextPk = (PeakPos *)car(cdr(trkNode));
	costAA = paths[0]->cost + PhaseCostFn(thisPk, nextPk, 0, 0);
	costAB = paths[0]->cost + PhaseCostFn(thisPk, nextPk, 0, 1);
	costBA = paths[1]->cost + PhaseCostFn(thisPk, nextPk, 1, 0);
	costBB = paths[1]->cost + PhaseCostFn(thisPk, nextPk, 1, 1);
	if(debug)    fprintf(stderr, "AA %6.3f BA %6.3f BB %6.3f AB %6.3f \n",
			     costAA, costBA, costBB, costAB);
	if( (costAA <= costBA) && (costBB > costAB) )
	    {	/* both next phases cheaper via this A phase */
	    if(debug)	fprintf(stderr,"type 1: both A\n");
	    CopyPathPhases(paths[0], paths[1]);	/* copy A path to B */
	    costA = costAA;
	    costB = costAB;
	    }
	else if( (costAA > costBA) && (costBB <= costAB) )
	    {	/* both next phases cheaper via this B phase */
	    if(debug)	fprintf(stderr,"type 2: both B\n");
	    CopyPathPhases(paths[1], paths[0]);	/* copy B path to A */
	    costA = costBA;
	    costB = costBB;
	    }
	else if( (costAA <= costBA) && (costBB <= costAB) )
	    {	/* both tracks stay on course */
	    if(debug)	fprintf(stderr,"type 3: straight\n");
	    costA = costAA;
	    costB = costBB;
	    }
	else /* costAA > costBA && costBB > costBA : swap A & B */
	    {
	    if(debug)	fprintf(stderr,"type 4: swap\n");
	    SwapPtrs((char **)&paths[0], (char **)&paths[1]);
	    costA = costBA;
	    costB = costAB;
	    }
	AppendPathStep(paths[0], 0, costA);
	AppendPathStep(paths[1], 1, costB);
	if(debug) {	PrintPath(paths[0], "A"); PrintPath(paths[1], "B"); }
	trkNode = cdr(trkNode);
	}
    if(paths[0]->cost <= paths[1]->cost)
	{
	FreePath(paths[1]);
	return(paths[0]);
	}
    else{
	FreePath(paths[0]);
	return(paths[1]);
	}
    }

static void SwapPhases(peak)	/* swap over <phs> and <ophs> fields */
    PeakPos	*peak;
    {
    float	tmpphs;

    tmpphs = peak->phs;
    peak->phs = peak->ophs;
    peak->ophs = tmpphs;
    }

void FixTrackPerPath(path)
    PATH	*path;
    {	/* swap phases in track so that path in <path> is always in ->phs */
    LIST	trk, phs;
    PeakPos	*peak;

    trk = (LIST)path->track->tktop;
    phs = path->phases;
    while(phs)
	{
	peak = (PeakPos *)car(trk);
	if((int)car(phs) != 0)
	    SwapPhases(peak);
	trk = cdr(trk);
	phs = cdr(phs);
	}
    }

#ifdef MAIN	/* testing code below here */

float testPhs[][2] =
    { {    0.2,	  0.1  },
      {    0.3,   0.2  },
      {    0.2,   0.45 },
      {    0.3,   0.55 },
      {    0.35,  0.40 },
      {    0.5,   0.75 }
    };

void main()
    {
    TRACK	*track;
    LIST        *n_ptr;
    PeakPos	*peak;
    PATH	*path;
    int 	i, t_len;
    float	frq = 0.0;	/* makes edph == 0 */
    float	mag = 1.0;
    float	tim = 0, dt = 1;
    int 	types[1];
    void	(*free_fns[1])();	/* array of ptrs to fns */

    track = tkNew(NULL);
    n_ptr = &(track->tktop);
    t_len = sizeof(testPhs)/(2*sizeof(float));
    for(i = 0; i<t_len; ++i)
	{
	peak = NewPeak(frq, mag, testPhs[i][0], tim);
	peak->ophs = testPhs[i][1];
	*n_ptr = new_node(NULL, (LIST)peak, T_PEAK);
	n_ptr = cdr_ptr(*n_ptr);
	tim += dt;
	}
    tkPrint(track, "Test track");
    path = BuildBestPath(track);
    PrintPath(path,"path");
    FixTrackPerPath(path);
    tkPrint(track, "Afterwards..");
    FreePath(path);
    types[0] = T_PEAK;	free_fns[0] = FreePeak;
    free_any_lists(track->tktop, 1, types, free_fns);
    tkFree(track);	/* doesn't get rid of peaklists */
    }

#endif /* MAIN */
