/***************************************************************\
*	pklist.c		  				*
*   Take a spectral frame and convert into a list of peaks	*
*   Based on routines developed within dspB, for thesis		*
*   dpwe 08jul91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#include <dpwelib.h>
/* personal libs */
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include "peaks.h"

/* constants */
#define PEAKWIN	 5	/* how many points to search for maxima over */
#define PEAKWIN2 (PEAKWIN/2)

#define MINLINMAG   (0.00001)   /* don't even bother with peaks this small */

/* static function prototype */
static void FitQuadtoPeak(float *mags, int minc, float *phas, int pinc, int pkIndex, 
				float *pFrq, float *pMag, float *pPha, float *pOph, int phaQ);

/* static data */
extern float dbgFrq;
extern float dbgFrqWin;

/* functions */

static peakIndex = 0;		/* each node has a unique id # */

void ResetPeakIndex(void)
{
    peakIndex  = 0;
}

BigPeakPos *NewBigPeak(frq, mag, phs, tim)
    FLOATARG	frq;
    FLOATARG	mag;
    FLOATARG	phs;
    FLOATARG	tim;
    {
    BigPeakPos	*rslt;

    rslt = TMMALLOC(BigPeakPos, 1, "NewBigPeak");
    rslt->index = peakIndex++;
    rslt->frq = frq;
    rslt->mag = mag;
    rslt->phs = phs;
    rslt->tim = tim;
    rslt->x = 0;	rslt->y = 0;
    rslt->next = NULL;
    return rslt;
    }

PeakPos *NewPeak(frq, mag, phs, tim)
    FLOATARG	frq;
    FLOATARG	mag;
    FLOATARG	phs;
    FLOATARG	tim;
    {
    PeakPos	*rslt;

    rslt = TMMALLOC(PeakPos, 1, "NewPeak");
    rslt->frq = frq;
    rslt->mag = mag;
    rslt->phs = phs;
    rslt->tim = tim;
    rslt->x = 0;	rslt->y = 0;
#ifdef TRACEALLOC
    fprintf(stderr, "NP(%lx,%.6f) ", rslt, tim);	fflush(stderr);
#endif /* TRACEALLOC */
    return rslt;
    }

void FreePeak(peak)
    PeakPos	*peak;
    {
#ifdef TRACEALLOC
    fprintf(stderr, "FP(%lx,%.6f) ", peak, peak->tim);	fflush(stderr);
#endif /* TRACEALLOC */
    free(peak);
    }

static void FitQuadtoPeak(mags, minc, phas, pinc, pkIndex, 
			  pFrq, pMag, pPha, pOph, phaQ)
    float	*mags;		/* results of fft */
    int 	minc;		/* stepsize for mag array */
    float	*phas;
    int 	pinc;		/* stepsize for phase array */
    int 	pkIndex;	/* bin index of the actual peak */
    float	*pFrq;		/* return 'real' index of fitted peak */
    float	*pMag;		/* return magnitude at peak */
    float	*pPha;		/* ? interpolated phase at peak */
    float	*pOph;		/* alternate interpolated phase (other loop) */
    int 	phaQ;		/* if zero, don't bother with the phs calc */
    {
    float	frq;
    float	yn, y0, yp, x0;
    float	ph0, ph1;
    int 	mapToLog = 1;

    yn = mags[minc*(pkIndex-1)];
    y0 = mags[minc*pkIndex];
    yp = mags[minc*(pkIndex+1)]; 
    /* hope the peakpkr didn't leave us at end */
    /* Parabola should be the 'right' fit in logmag space */
    if(mapToLog)  {
	if(yn <= 0 || yp <= 0 || y0 <= 0)  {
	    fprintf(stderr, "Oops - tried to find peak in zero/negative data by log\nMags %f %f %f\n", yn, y0, yp);
	    abort();
	}
	yn = log(yn);
	y0 = log(y0);
	yp = log(yp);
#if !(__sgi || THINK_C)   /* no isnan on indigo or Mac */
	if(isnan(yn) || isnan(yp) || isnan(y0))		abort();
#endif /* !(__sgi||THINK_C) */
    }
    x0 = (yn - yp)/(2*(yp + yn - 2*y0));
    *pFrq = x0+(float)pkIndex;
    *pMag = y0 - ((yp+yn)/2-y0)*x0*x0;
    if(mapToLog)
	*pMag = exp(*pMag);
    if(phaQ)
	{
	float dpha;	/* calculate rate of change of phi */

	if(x0<0) 
	    {
	    --pkIndex;
	    x0 = x0 + 1;
	    }
	ph0 = phas[pinc*pkIndex];
	ph1 = phas[pinc*(pkIndex+1)];
	if( (ph1 - ph0)>PI )	ph1 -= TWOPI;	/* tentative ph unwrap..? */
	if( (ph1 - ph0)<-PI )	ph1 += TWOPI;
	dpha = ph1-ph0;
	*pPha = ph0 + x0*(ph1-ph0);		/* linear interpolate */
	if( ph1 > ph0 )	ph1 -= TWOPI;
	else		ph1 += TWOPI;
	*pOph = ph0 + x0*(ph1-ph0);		/* alternative wrapping */
	if( *pOph>PI )	*pOph -= TWOPI;		/* make sure PV */
	if( *pOph<-PI )	*pOph += TWOPI;

	*pOph = dpha;	/******** overwrite opha slot with dpha - for now */
	}
    }

PKLIST ListPeaks(magfrm, maginc, phsfrm, phsinc, len, sign, 
		 time, binsPOct, frqBase, frqLim, logFlg)
    float	*magfrm;
    int 	maginc;
    float	*phsfrm;
    int 	phsinc;
    long	len;
    int 	sign;	/* normally 1, or -1 to find troughs */
    FLOATARG	time;	/* what to put in time slot */
    FLOATARG	binsPOct;
    FLOATARG	frqBase;	/* to convert bin index to Hz : exp(fs*b)*fb */
    FLOATARG	frqLim;		/* don't return peaks above this frq */
    int 	logFlg;	/* flag: 1 => take log of magnitude */
    /* return a LIST structure of the Peaks found in the vector, interp'd */
    {
    float	f;
    int 	base = 0, limbase, l, m, u;
    PKLIST	root = NULL;
    PeakPos	*peak;
    int 	doPhase = 0;
    float	phsep;
    float	frqScale;
    float	frq, mag, phs1, phs2;

#define SPINSIZE 4
    static	int spin = 0;
    static      char *spinset = "-\\|/";

    static	char tchr = 'A';
    char        schr = 'A';

/*    putc('\r', stderr);	putc(spinset[spin],stderr);	fflush(stderr);
    if(++spin == SPINSIZE)	spin = 0;
    fprintf(stderr, " %.5f", time);	fflush(stderr);
*/
    if(phsfrm)		doPhase = 1; 	/* only interp. phase if some passed */
    frqScale = log(2.0)/binsPOct;
    base = len - PEAKWIN;	/* start at end, build list in reverse */
    limbase = binsPOct * log(frqLim/frqBase)/log(2.0);
    if(base > limbase)	base = limbase;
    while(base >= 1-PEAKWIN2)
	{
	m = base+PEAKWIN2;
	l = m - 1;
	u = m + 1;
	if( (f = sign * magfrm[maginc * m]) 
	   > sign * magfrm[maginc * l]
	   && f >= sign * magfrm[maginc * u]
	   && (logFlg && f > MINLINMAG) )	/* no peaks blw MINLINMAG */
	    { /* centre of frame has max value in that frame */
	    peak = NewPeak(0.0,0.0,0.0,time);

/*	    if(fabs(time-0.0681450143) < 0.0001)
		fprintf(stderr,"sawit\n");
 */
	    FitQuadtoPeak(magfrm, maginc, phsfrm, phsinc, m, 
			  &frq, &mag, &phs1, &phs2, doPhase);
	    if(logFlg)
		{
		if(mag>MINLINMAG)
		    peak->mag = LOGN2DB*log(mag);
		else
		    peak->mag = -100.0;
		}
	    else
		peak->mag = mag;
	    /* convert frq from filterbank units to Hz */
	    peak->frq = exp(frqScale*frq)*frqBase;
	    peak->phs = phs1;
	    peak->ophs= phs2;	/* alternate phase wrap */

	    if(frq > (float)m)
		phsep = phsfrm[phsinc*m] - phsfrm[phsinc*u];
	    else
		phsep = phsfrm[phsinc*l] - phsfrm[phsinc*m];
/* 	    if(fabs(phsep)> .8*PI && fabs(phsep)<1.2*PI)
		fprintf(stderr,
	  "LstPk: ambiguous phase: t%5f f%6.2f mg%6.2f dPh%6.3f\n",
			time, peak->frq, peak->mag, phsep); */
	    
	    /* Crazy debug for rogue phase */
/*	    if(time>=0.1432 && time <0.1438 
/*	      && frq >= 40 && frq < 44  )
/*		{	/* new formatting for EMACS sort & xgraph */
/*		if(frq > (float)m)
/*		    fprintf(stderr,
/*	    "%c%c %.5f %6.3f\n%c%c %.5f %6.3f\n%c%c %.5f %6.3f\n",
/*	    schr,tchr,time,phsfrm[phsinc*m],
/*	    schr+1,tchr,time,peak->phs,
/*	    schr+2,tchr,time,phsfrm[phsinc*u]);
/* /*  "^ m %6.2f:%6.2f:%6.2f f%4.0f:%6.2f:%4.0f p%6.3f:%6.3f:%6.3f\n",
/*			    magfrm[maginc*m], peak->mag, magfrm[maginc*u],
/*			    (float)m, peak->frq, (float)u,
/*			    phsfrm[phsinc*m], peak->phs, phsfrm[phsinc*u]); 
/* */
/*		else
/*		    fprintf(stderr,
/*	    "%c%c %.5f %6.3f\n%c%c %.5f %6.3f\n%c%c %.5f %6.3f\n",
/*	    schr,tchr,time,phsfrm[phsinc*l],
/*	    schr+1,tchr,time,peak->phs,
/*	    schr+2,tchr,time,phsfrm[phsinc*m]);
/* /*  "v m %6.2f:%6.2f:%6.2f f%4.0f:%6.2f:%4.0ff p%6.3f:%6.3f:%6.3f\n",
/*			    magfrm[maginc*l], peak->mag, magfrm[maginc*m],
/*			    (float)l, peak->frq, (float)m,
/*			    phsfrm[phsinc*l], peak->phs, phsfrm[phsinc*m]); 
/* */
/*		++tchr;	/* static lexical time index advanced */
/*		}
 */
	    if(fabs(peak->frq - dbgFrq)<dbgFrqWin)	
		PrintPeak(peak, "AC");
#if !(__sgi || THINK_C)   /* no isnan on indigo or mac */
	    if(isnan(peak->frq))
	       abort();
#endif /* !(__sgi || THINK_C) */
	    root = (PKLIST)add_before((LIST)root, (NODE *)peak, T_PEAK);
	    }
	--base;
	}
    return root;
    }

void pklistFree(PKLIST pl)
{	/* call to list-free that can handle peaknodes */
    int types[1];
    void (*freeFns[1])();
    
    types[0] = T_PEAK;
    freeFns[0] = FreePeak;
    free_any_lists((LIST)pl, 1, types, (NodeFunc *)freeFns);
}

void PrintPeak(PeakPos *peak, char *label)
{
    fprintf(stderr, "<Pk'%s':t=%.6f,f=%.2f,m=%7.3f,p=%9.6fpi,dp=%9.6fpi/b>\n",
	    label, peak->tim, peak->frq, peak->mag, 
	    peak->phs, peak->ophs);
}

/* static function prototype */
static void PrintPeakListSubs(PKLIST list, int level);
static void PrintPeakNode(PKLIST node, int level);

static void PrintPeakListSubs(list, level)
    PKLIST	list;
    int 	level;
    {
    PKNODE	*node;

    node = list;
    while(node)
	{
	PrintPeakNode(node, level);
	node = (PKNODE *)cdr((LIST)node);
	}
    }

static void PrintPeakNode(node, level)
    PKLIST	node;
    int 	level;
    {
    int 	t,l;
    PeakPos	*pk;

    if(node == NULL)	return;
    t = cartype((LIST)node);
    l = 0;
    while(l++<level)
	printf("    ");
    if(t == T_PEAK)
	{
	pk = (PeakPos *)car((LIST)node);
	printf("PeakNode(%lx,%lx): f %6.2f, m %6.1f, p %6.2f, t %6.4f\n",
	       node, pk, pk->frq, pk->mag, pk->phs, pk->tim);
	}
    else if(t == T_LIST)	/* is this necessary?  No nested pklists now */
	{
	printf("Node(%lx):\n", node);
	PrintPeakListSubs((PKLIST)car((LIST)node), level+1);
	}
    }

void PrintPeakList(pklist, name)
    PKLIST	pklist;
    char	*name;
{
    LIST nearLast;
    LIST list = (LIST)pklist;	/* convenient typecast */
    long tpsiz = top_size(list);

    if(name)	printf("%s\n",name);
    if(tpsiz < 20)  {	/* only print manageably-sized lists in entirety */
	PrintPeakListSubs(pklist, 0);
    } else {	/* print first and last nodes */
	PrintPeakNode(pklist, 0);
	PrintPeakNode((PKLIST)cdr(list), 0);
	PrintPeakNode((PKLIST)cdr(cdr(list)), 0);
	PrintPeakNode((PKLIST)cdr(cdr(cdr(list))), 0);
	printf("...\n");
	nearLast = find_prec(list, 
			     find_prec(list, 
				       find_prec(list, 
						 find_last(list))));
	PrintPeakListSubs((PKLIST)nearLast, 0);
    }
}

