/***************************************************************\
*	tkgroups.c		  				*
*   Basic operations on structures which are groups of tracks 	*
*   dpwe 1993may02						*
\***************************************************************/

#include "tkgroups.h"
#include <genutils.h>	/* alloc */

static int tkgr_extend_stats(NODE *data, int type, void *arg)
{	/* MAPpable function gathers data from all TrackStats */
    TRACK *trk = (TRACK *)data;
    TrackStats *st = (TrackStats *)arg;

    assert(type == T_TRACK);
    tkStStatUpdate(st, &(trk->st));
    return 1;
}

TKGROUP *tkgrNew(TKLIST trax, int type)
{   /* instance a new tkgroup containing <trax>, of <type> */
    TKGROUP *tkg;

    tkg = TMMALLOC(TKGROUP, 1, "tkgrNew");
    tkg->parts = trax;
    /* calculate overall stats */
    tkStReset(&(tkg->st), NULL);
    maparg((LIST)trax, tkgr_extend_stats, &(tkg->st));
    tkStBoildown(&(tkg->st));	/* nonsense for sd's, works for avgs */
    tkg->flags = 0;
    tkg->type  = type;
    tkg->score = -1.0;	/* not a valid score */
    tkg->label = NULL;
    return tkg;
}

void tkgrFree(TKGROUP *tkg)
{   /* destroy an instance */
    if(tkg->label)	free(tkg->label);	/* (a little risky) */
    free(tkg);
}

int tkgrAddTracks(TKGROUP *tkg, TKLIST tks)
{	/* append to <tkg->parts> TRACKs from <tks> not currently there.
	   Update stats.  Returns the actual number of tracks added. */
    TKLIST newTrax;

    newTrax = (TKLIST)remove_list((LIST)tks, (LIST)tkg->parts);
    tkStBoilup(&(tkg->st));
    maparg((LIST)newTrax, tkgr_extend_stats, &(tkg->st));
    tkStBoildown(&(tkg->st));
    /* just append to parts list - no sorting imposed */
    *(find_last_ptr((LIST *)&(tkg->parts))) = newTrax;
    return top_size(newTrax);
}

TKGROUP *tkgrDup(TKGROUP *tkgr)
{   /* allocate a copy of the existing <tkgr> */
    TKGROUP	*tkg;
    TKLIST	trx;

    trx = (TKLIST)copy_top((LIST)tkgr->parts);	/* new copy of track list */
    tkg = tkgrNew(trx, tkgr->type);
    /* bit-for-bit copy of stats str (despite recalc in tkgrNew) ? */
/*    memcopy(&(tkg->st), &(tkgr->st), sizeof(TKGROUP)-sizeof(TKLIST));  */
    return tkg;
}

/* currently in db.peaks.c ??? */
extern int TimeFreqComp(NODE **pnode1, NODE **pnode2);

void tkgrAddTrack(TKGROUP *tkgr, TRACK *trk)
{   /* add a new track to existing group */
    TrackStats *st = &(tkgr->st);
    LIST	newNode;

    newNode = new_node(NULL, (LIST)trk, T_TRACK);
    insert_by_comp((LIST *)&(tkgr->parts), newNode, TimeFreqComp);
    tkStBoilup(st);
    tkStStatUpdate(st, &(trk->st));
    tkStBoildown(st);
}

static char *tkgrNames[] = {	/* ascii names for type field */
    /* must match enum in tkgroups.h */
    "NON", "HRM", "ONS", "CNT", "AMP", "NSE", "RCS", "OTH"
};

void tkgrTitle(TKGROUP *tkgr, char *str)
{   /* print some descriptive title for this tkgr into <str>, at 
       least 80 chr long */
    sprintf(str, "g%s:t=%8.6f..%8.6f, f=%6.1f..%6.1f, m=%6.2f..%6.2f", 
	    tkgrNames[tkgr->type], 
	    tkgr->st.tmin, tkgr->st.tmax, tkgr->st.fmin, tkgr->st.fmax, 
	    tkgr->st.mmin, tkgr->st.mmax);
}

void tkgrPrint(TKGROUP *tkgr, char *label)
{   /* print some brief summary of the tkgroup */
    char str[128];
    fprintf(stderr, "trk grp '%s':\n", label);
    tklPrintOnsets(tkgr->parts);
    tkgrTitle(tkgr, str);
    fprintf(stderr, "%s\n", str);
}

static int tk_print_id(NODE *data, int type, void *arg)
{	/* MAPpable function prints track ID to stream */
    TRACK *trk = (TRACK *)data;
    FILE  *file = (FILE *)arg;

    assert(type == T_TRACK);
    fprintf(file, "%d ", trk->id);
    return 1;
}

void tkgrSave(TKGROUP *tkgr, char *label, FILE *file)
{   /* save the group to disk with a descripton.
       Just write the track IDs. */
    fprintf(file, "'%s' %d ", label, tkgr->type);
    maparg((LIST)tkgr->parts, tk_print_id, file);
    fprintf(file, "\n");
}

static int tg_save(NODE *node, int type, void *data)
{	/* mappable fn to save each tk group in a list */
    TKGROUP *tkg = (TKGROUP *)node;
    FILE *stream = (FILE *)data;
    char *name;

    assert(type==T_TKGROUP);
    if(tkg->label)	name = tkg->label;
    else		name = "(noname)";
    tkgrSave(tkg, name, stream);
    return 1;
}

#define TGLF_MAGIC "TGLF"	/* 4 byte magic ID for saved TG files */

int tglSave(TGLIST tgl, char *name, char *title)
{   /* save a whole list of track groups */
    FILE *file;

    if((file = fopen(name, "w")) == NULL)  {
	fprintf(stderr, "tglSave: could not create '%s'\n", name);
	return 0;
    }
    /* first line is magic number ID */
    fprintf(file, "%s\n", TGLF_MAGIC);
    /* next line is group title */
    fprintf(file, "GroupListTitle: %s\n", title);
    /* following lines are group specifications */
    maparg((LIST)tgl, tg_save, (void *)file);
    fprintf(stderr, "TrackGrp '%s' saved to file '%s' (%d grps, %d bytes)\n", 
	    title, name, top_size(tgl), ftell(file));
    fclose(file);
    return 1;
}

static int GobbleInt(FILE *file, int *ans)
{   /* take chrs from file one by one, make into an int.
       Return 0 if no number found. */
    char c;
    int  rslt = 0;
    int  started = 0;
    int  positive = 1;
    int	 done = 0;

    while(!done)  {
	c = fgetc(file);
	if(feof(file))  {
	    done = 1;	/* hit eof */
	} else {
	    if(!started && (c == ' ' || c == '\t'))  {
		;		/* skip over leading white space */
	    } else if(!started && c == '-')  { /* leading minus */
		positive = -1;
		started = 1;
	    } else if(c >= '0' && c <= '9')  {
		started = 1;
		rslt = (10*rslt)+(c - '0');
	    } else {		/* could not parse - stop */
		done = 1;
		/* should maybe re-push last chr */
		/* also, maybe push back '-' if ate one */
	    }
	}
    }
    if(started)  {	/* (this makes '-' read as '0') */
	*ans = positive * rslt;
	return 1;
    } else {		/* nothing found */
	return 0;
    }
}

static void GobbleToNL(FILE *file)
{   /* pull chrs off file until '\n' */
    char c;
    int  done = 0;
    int  reportingerror = 0;

    while(!done)  {
	c = fgetc(file);
	if(feof(file))  {
	    done = 1;
	} else {
	    if(c == '\n' || c == '\r')
		done = 1;
	    else if(reportingerror || (c != ' ' && c != '\t'))  {
		/* warn user if gobbling non-WS */
		if(!reportingerror)  {
		    fprintf(stderr, "GobbleNL ate '%c", c);
		    reportingerror = 1;
		} else {
		    fprintf(stderr, "%c", c);
		}
	    }
	}
    }
    if(reportingerror)
	fprintf(stderr, "'\n");
}

/* used also by db.trkio.c, where it comes from */
int tk_check_id(LIST car, int type, void *arg)
{	/* a NodeFunc to find a node with a particular ID */
    assert(type == T_TRACK);
    if(((TRACK *)car)->id == (long)arg)
	return 1;
    else
	return 0;
}

TKLIST tklLoad(FILE *file, TKLIST allTrx)
{   /* build a track list from a sequence of track IDs on a single line.
       Find the tracks that match the trackIDs from allTrx. */
    TKLIST	tkl = NULL, *runner;
    TKLIST	top = allTrx;
    TRACK 	*trk;
    int 	id;
    int 	warned = 0;

    runner = &tkl;
    while(GobbleInt(file, &id))  {
	top = (TKLIST)search_by_fn((LIST)top, tk_check_id, (void *)id);
	if(top == NULL)  {
	    if(warned == 0)  {
		fprintf(stderr, "tklLoad: could not find id %d\n", id);
		warned = 1;
	    }
	    /* abort(); */
	} else {
	    trk = (TRACK *)CAR((LIST)top);
	    runner = append_ptr(runner, new_node(NULL, (LIST)trk, T_TRACK));
	}
    }
    return tkl;
}

static int GobbleQuotedString(FILE *file, char *buf)
{	/* read a string chr by chr into the (pre-allocd) buf.
	   String must begin and end with single-quote (?) */
    char c;
    int done = 0;
    int ok = 0;
    int i = -1;

    while(!done)  {
	c = fgetc(file);
	if(feof(file))  {
	    done = 1;	/* hit eof */
	} else {
	    if(i == -1)  {
		assert(c == '\'');
		++i;
	    } else if(c=='\'')  {
		buf[i++] = '\0';
		done = 1;
		ok = 1;		/* correct termination */
	    } else {
		buf[i++] = c;
	    }
	}
    }
    return ok;
}

TKGROUP *tkgrLoad(FILE *file, TKLIST allTrx)
{   /* Read back a tkgroup def from file, advance file ptr.
       Need to have allTrx passed in, since we search through it
       for the tkIds specified in the file.  */
    char 	buf[256];
    int 	type, id, quotedStringOK;
    TKGROUP	*rslt = NULL;
    TKLIST	tkl;

/*  fscanf(file, "'%s' ", buf);	/* does not work well - gets second ' */
    quotedStringOK = GobbleQuotedString(file, buf);
    if(!feof(file))  {
	assert(quotedStringOK);
	/* fprintf(stderr, "group '%s'\n", buf); */
	/* next comes type number */
	assert(GobbleInt(file, &type));
	/* remaining numbers are tracks */
	if(tkl = tklLoad(file, allTrx)) {
	    rslt = tkgrNew(tkl, type);
	    /* read through to end of line */
	    /* GobbleToNL(file); */ /* tklLoad already gobbles NL */
	}
    }
    return rslt;
}

TGLIST tglLoad(char *name, char **ptitle, TKLIST allTrax)
{   /* load a list of track groups from file <name>.  
       If tkgs read, return the list (else NULL), and 
       write the specified title into <*ptitle>.
       Needs <allTrx> context to be able to find tracks 
       by ID number as they are referenced in the file. */
    FILE *file;
    char buf[256];
    TGLIST tgl = NULL;
    TKGROUP *tg;
    LIST *runner;

    if((file = fopen(name, "r")) == NULL)  {
	fprintf(stderr, "tglLoad: could not open '%s'\n", name);
	return NULL;
    }
    /* first line is magic number ID */
    fscanf(file, "%s\n", buf);
    if(strcmp(buf, TGLF_MAGIC)!= 0)  {
	fprintf(stderr, "tglLoad: '%s' is not a track file ('%s')\n",
		name, buf);
	goto end;
    }
    /* next line is group title */
    fscanf(file, "GroupListTitle: %s\n", buf);
    if(ptitle)
	*ptitle = strdup(buf);
    /* following lines are group specifications */
    runner = (LIST *)&tgl;
    while( tg = tkgrLoad(file, allTrax) )  {
	runner = append_ptr(runner, new_node(NULL, (LIST)tg, T_TKGROUP));
    }
    fprintf(stderr, "TrackGrp '%s' loaded from '%s' (%d grps, %d bytes)\n", 
	    buf, name, top_size(tgl), ftell(file));
 end:
    fclose(file);
    return tgl;
}

