/***************************************************************\
*	tkgroups.h		  				*
*   Definition of structures comprising *groups* of tracks 	*
*   dpwe 1993may02						*
\***************************************************************/

#ifndef _TKGRP_H_
#define _TKGRP_H_

#include <dpwelib.h>
#include "tracks.h"	/* brings in all the other stuff too */

/* first, define specific innard parts for different groups */

enum {	/* possible values of `type' field */
    TG_HARM = 1, 	/* a harmonic grouping */
    TG_ONSET, 		/* an onset-aligned grouping */
    TG_CONT,		/* a t-f continuity grouping (metatrack) */
    TG_AM,		/* common amplitude modulation */
    TG_NOISE, 		/* region of uncorrelated noise energy */
    TG_NEST,		/* a NESTED set of groupings */
    TG_OTHER		/* ? undefined */
};

typedef struct {	/* special field for harmonic groups */
    TKLIST	fund;	/* one track/metatrack holds parent frq contour */
    float	ratio;	/* actual fundamental bears this relation to trk */
} HARM_INFO;

typedef struct {	/* special field for onset groups */
    TRACK      *latest;	/* track that was most recently added */
    float	otime;	/* the nominal onset time */
} ONSET_INFO;

typedef struct {	/* special field for continuity groups */
    TRACK      *latest;	/* track that was most recently added */
    float	durav;	/* average duration of component tracks */
    float	dursd;	/* std devn of duration of component tracks */
    float	magav;	/* average track magnitude */
    float	ttot;	/* sum of durations of all tracks */
} CONT_INFO;

typedef struct {	/* special field for common-am groups */
    float	prdav;	/* average period between successive am peaks */
    float	prdsd;	/* std.dev. of period between am peaks */
} AM_INFO;

typedef struct {	/* special field for noise regions */
    float	tdecay;	/* decay time of overall ampl. env */
} NOISE_INFO;

typedef struct {	/* special field for nested groupings */
    /* TGLIST */ LIST	support;	/* TKGROUPs this is based on */
} NEST_INFO;

typedef struct {	/* descriptive wrapper around list of tracks */
    TKLIST	parts;	/* the component tracks, could also be TKGROUPs */
    TrackStats	st;	/* description of support and stats of this set */
    int 	flags;	/* various undefined bits */
    int 	type;	/* indicates character of this group */
    union {		/* special fields for different types */
	HARM_INFO	harm;
	ONSET_INFO	onset;
	CONT_INFO	cont;
	AM_INFO		am;
	NOISE_INFO	noise;
	NEST_INFO	nest;
    } 		v;
    float	score;	/* `goodness of fit' score for this set */
    char 	*label;	/* text description for selpanel entries (free'd) */
} TKGROUP;

typedef LIST TGLIST;

#define T_TKGROUP	105	/* <llists> type flag for these ones */

TKGROUP *tkgrNew(TKLIST trax, int type);
    /* instance a new tkgroup containing <trax>, of <type> */
void tkgrFree(TKGROUP *tkg);
    /* destroy an instance */
int tkgrAddTracks(TKGROUP *tkg, TKLIST tks);
    /* append to <tkg->parts> TRACKs from <tks> not currently there.
       Update stats.  Returns the actual number of tracks added. */
TKGROUP *tkgrDup(TKGROUP *tkgr);
    /* allocate a copy of the existing <tkgr> */
void tkgrAddTrack(TKGROUP *tkgr, TRACK *trk);
    /* add a new track to existing group */
void tkgrTitle(TKGROUP *tkgr, char *str);
    /* print some descriptive title for this tkgr into <str>, at 
       least 80 chr long */
void tkgrPrint(TKGROUP *tkgr, char *label);
    /* print some brief summary of the tkgroup */
void tkgrSave(TKGROUP *tkgr, char *label, FILE *file);
    /* save the group to disk with a descripton. Just write the track IDs. */
int tglSave(TGLIST tgl, char *name, char *title);
    /* save a whole list of track groups */
int tk_check_id(LIST car, int type, void *arg);
    /* a NodeFunc to find a node with a particular ID */
TKLIST tklLoad(FILE *file, TKLIST allTrx);
    /* build a track list from a sequence of track IDs on a single line.
       Find the tracks that match the trackIDs from allTrx. */
TKGROUP *tkgrLoad(FILE *file, TKLIST allTrx);
    /* Read back a tkgroup def from file, advance file ptr.
       Need to have allTrx passed in, since we search through it
       for the tkIds specified in the file.  */
TGLIST tglLoad(char *name, char **ptitle, TKLIST allTrx);
    /* load a list of track groups from file <name>.  
       If tkgs read, return the list (else NULL), and 
       write the specified title into <*ptitle>.
       Needs <allTrx> context to be able to find tracks 
       by ID number as they are referenced in the file. */

#endif /* _TKGRP_H_ */

