/***************************************************************\
*	tkplay.c		  				*
*   Takes 'track' file as written by trk and uses the tracks 	*
*   to drive a bank of sinewave oscillators, superimposed into 	*
*   a single soundfile, the resynthesised sound.		*
*   1993mar03  dpwe after tksyn.c				*
\***************************************************************/

/**** flags ****/
#define CONST_FREQ	/* selects between zero-order and 1st-order
			   frequency interpolation.  Problem is, first 
			   order frequency fixup method has mundo 
			   wiggles at half the track sampling rate 
			   (see notebook, 1993mar08) */
/* #define USE_TSIN	/* use table lookup cosine function 
			   (a little faster than cos() on DECstn) */
 
/**** std libs ****/
#include <dpwelib.h>
#include <assert.h>
#include <math.h>	/* rint() at least */
double rint(double);	/* WHY not in math.h??? */
 
/**** personal libs ****/
#include <genutils.h>   /* for maskPhs */
#include <vecops.h>	/* for VZero */
#include <byteswap.h>	/* byteswap trk files in load */

#include "tkplay.h"	/* prototypes and tracks.h/trkheadio.h/snd.h */

#ifndef FOR_MATLAB
#include <audIO.h>	/* for real-time playing output */
#else /* is FOR_MATLAB - need some dumb debug variables to link */
float dbgFrq = 2000.0;
float dbgFrqWin = 5.0;
#endif /* FOR_MATLAB */

#ifdef USE_TSIN
#include "tsin.h"	/* table lookup sine function a little faster? */
#endif /* USE_TSIN */

/**** constants ****/
#define DFLT_SRATE	44100.0		/* if none other spec'd */
#define DBOFST    (-29.7)	/* knock mags down by 30dB for approx scale */

/**** static function prototype ****/
 
/**** static data ****/
       double srate = DFLT_SRATE;
       double speriod = 1.0/DFLT_SRATE;
double dbgtime = 999999.9999;
TRACK  *dbgtrk = NULL;
 
/**** macros ****/
/* way-cool PHMSK, PHMSKSETUP macros now in peaks.h  - 1993apr18 */ 
 
/**** functions ****/

static double tkBestDphi(PeakPos *p0, PeakPos *p1)
{	/* estimate the actual phase advance between two peaks 
	   i.e. resolve the 2pi ambiguity based on the frq estimates */
    double	twopi = 2.0*PI;
    double	dphi, efrq, dt, expDphi, phierr;
 
    /* We now have the measured phase points for two sample 
       points, ph0 at t0 and ph1 at t1.  We also have frequency 
       'estimates' i.e. the interpolated peak positions at both 
       times, ef0 and ef1.  However, ph0 and ph1 are of course 
       two-pi ambiguous.  We want to know the 'true' phase 
       increment between the two samples, so we will assume 
       an average frequency midway between ef0 and ef1, then 
       find the phase advance consistent with ph0 and ph1 that 
       is closest to that average frq estimate, by adding 2pi's 
       to the raw difference
     */
    dphi = p1->phs - p0->phs;		/* raw frq diff'ce */
    efrq = (p0->frq + p1->frq)/2.0;	/* avg estimated frq */
    dt   = p1->tim - p0->tim;		/* time step */
    expDphi = twopi * efrq * dt;	/* expected total phase advance */
    phierr  = expDphi - dphi;		/* total error from raw */
 
    return dphi + twopi * rint(phierr/twopi);	/* add 2nPi closest to err */
}
 
static void tkFixupFrqs(TRACK *trk)
{	/* adjust the frq data to exactly match phase data */
    PKLIST pkl = trk->tktop;
     /* Here, we calculate a new frequency value so that linear 
	phase interpolation in resynthesis will regenerate the 
	exact phase curve (given a single initial phase position).
	   
	Consider two sample times, t0 and t1, with *actual* phase 
	measurements ph0 and ph1 which will be resynthesized 
	from frequency samples f0 and f1 which will be linearly 
	interpolated.  Thus on resynth, 
 
	   f(t) = f0 + (f1-f0)/(t1-t0) * (t - t0)   (so f(t0) = f0 etc)
 
	This results in a phase curve the integral of above i.e. 
 
	   ph(t) = ph0 + 2pi * [ f0.(t-t0) + (df/dt)/2 * (t-t0)^2 ]
 
	where ph0 is the measured phase at t0, df = f1 - f0, 
	and dt = t1 - t0.
 
	Now what we want is for ph(t1) = ph1, the measured phase 
	at t1.  If we can set f1 to any value to achieve this 
	(the premise of this routine), then we have
 
	   dph/2pi = f0.dt + (f1 - f0)/(2.dt)*(dt^2)
 
	where dph = ph1 - ph0.  Solving for f1, 
 
	   f1 = f0 + (2/dt)*(dph/2pi - f0.dt)
 
	This gives us the frequency for the end of the interval, 
	which of course goes on to be f0 in the next time slice.  
	For the initial f0, we can use the linear slope between 
	the two phases.
      */
    PeakPos	*pk0, *pk1;
    double	dphi, dt, f0, f1;
    double	twopi = 2.0 * PI;
 
    /* Need at least two points to operate */
    if(pkl == NULL || pkl->next == NULL)	return;
    pk0 = pkl->peak;
    pkl = pkl->next;
 
    /* initial f0 is just phase slope in first segment */
    pk1  = pkl->peak;
    dphi = tkBestDphi(pk0, pk1);
    dt   = pk1->tim - pk0->tim;
    f0   = (dphi / twopi) / dt;
 
    while(pkl)  {
	pk1 = pkl->peak;
	dphi = tkBestDphi(pk0, pk1);
	dt   = pk1->tim - pk0->tim;

#ifndef CONST_FREQ	 /* NOT const_frq -> linear frq interp */
 
	f1 = f0 + 2/dt*(dphi/twopi-f0*dt);
 
	/* write back new frq values, move on */
	pk0->frq = f0;
#else /* CONST_FREQ */	 /* think again ! Constant frq over interval */
	/* if the sines have constant frequency in the interval, 
	   we want that frequency to be dphi/dt .. */
	f1 = pk0->frq = (dphi/dt)/twopi;
#endif /* CONST_FREQ */
	f0 = f1;
	pk0 = pk1;
 
	pkl = pkl->next;
    }
    /* make sure to store last value too */
    pk0->frq = f0;
}
 
static void tkPrintLinFrqInterp(TRACK *track)
{	/* To check tkFixupFrqs, print out the phases that will occur 
	   under linear frq interp using the given frqs */
    PeakPos 	*pk;
    PKLIST	pkl = track->tktop;
    double 	t0, t1, f0, f1, ph0, ph1, dt, df;
    PHMSKSETUP;	/* macro sets up for phase masking below */
    
    if(pkl == NULL || pkl->next == NULL)	return;
    pk = pkl->peak;
    t0 = pk->tim;
    f0 = pk->frq;
    ph0 = pk->phs;	/* just this used as starting point */
    pkl = pkl->next;
    while(pkl)  {
	pk = pkl->peak;
	t1 = pk->tim;
	f1 = pk->frq;
	dt = t1 - t0;
	df = f1 - f0;
 
	/* dphi = w.dt + (dw/dt) * (dt^2)/2   like taylor */
 
	ph1 = ph0 + 2.0*PI*( f0*(t1-t0) + df/dt*(t1-t0)*(t1-t0)/2);
	ph1 = PHMSK(ph1);
	fprintf(stderr, "t %8.6lf, phi %6.2lf\n", t1, ph0);
 
	t0 = t1;
	f0 = f1;
	ph0 = ph1;
	pkl = pkl->next;
    }
}

#define FADE_CYCS (3.0)         /* how many cycles of ctr frq for fade I/O */
#define V_SMALL_MAG (-100.0)    /* special dBmag value indicates zero lin */

static PKLIST tkFadeTrk(TRACK *track)
{       /* insert fabricated peaks with exceedingly small amplitude 
	   at the start and end of a track in order to allow 
	   fading in and out during synthesis.  This will 
	   mess up the ordering of the tracks in their parent 
	   list, since their onset time will now change, so 
	   must re-sort tracks after applying this. */
    PeakPos *pk;
    PKLIST  pkl, tail;
    double  frq, tim, phs, mag;

    assert(track != NULL);
    pkl = track->tktop;
    assert(pkl != NULL);
    /* build a fade-in node, insert in front of track */
    pk  = pkl->peak;
    frq = pk->frq;
    tim = pk->tim - FADE_CYCS * (1.0/frq);
    phs = maskPhs(track->startPhase - (FADE_CYCS*2*PI));
    mag = V_SMALL_MAG;
    track->tktop = (PKLIST)new_node((LIST)pkl, 
				    (LIST)NewPeak(frq, mag, phs, tim), 
				    T_PEAK);
    track->st.tmin = tim;
    track->startPhase = phs;
    ++track->st.npts;
    /* now add fade_out at end */
    tail = track->tkbot;
    pk = tail->peak;
    frq = pk->frq;
    tim = pk->tim + FADE_CYCS * (1.0/frq);
    phs = maskPhs(pk->phs + (FADE_CYCS*2*PI));
    mag = V_SMALL_MAG;
    set_next((LIST)tail, 
	     new_node(NULL, (LIST)NewPeak(frq, mag, phs, tim), T_PEAK));
    track->tkbot = tail->next;
    track->st.tmax = tim;
    ++track->st.npts;

    return pkl;
}

int tkCmpPklsByT(NODE **pn1, NODE **pn2)
{   /* qsort comparison function for pklists */
    PKLIST pl1, pl2;
    double t1, t2;

    pl1 = (PKLIST)(*pn1);
    pl2 = (PKLIST)(*pn2);

    t1 = pl1->peak->tim;
    t2 = pl2->peak->tim;

    if(t1 > t2)  return 1;
    else if(t1 == t2)  return 0;
    else return -1;
}

int tkCmpTrxByT(NODE **pn1, NODE **pn2)
{       /* qsort comparison function for TRACK list nodes - 
       both pn1 and pn2 are treated as (TRACK **)
       returns 1 if trk1 starts after trk2; 
       if they have the same tmin, then returns 1 if 
       trk1 has a lower frequency than trk2 
       - so result is tracks sorted in ascending start time, descending frq */
    TRACK *tk1, *tk2;
    double t1, t2, f1, f2;

    tk1 = (TRACK *)(*pn1);
    tk2 = (TRACK *)(*pn2);
    t1 = tk1->st.tmin;
    t2 = tk2->st.tmin;

    if(t1 > t2)		return 1;
    else if(t1 < t2) 	return -1;
    else {
	/* but if times are =, high frequency precedes low (yuk - for now) */
	/* (this is because tracks come out of pkmatch this way) */
	f1 = tk1->tktop->peak->frq;
	f2 = tk2->tktop->peak->frq;
	if(f1 < f2)		return 1;
	else if(f1 == f2)	return 0;	/* equal - same track!? */
	else			return -1;
    }
}

void tkFreeAll(TRACK *trk)
{	/* free a track, assuming it was read in from old-format file */
    free(trk->tktop);	/* all pks alloc'd as one */
    tkFree(trk);
}

int tk_free_all(NODE *trknode, int type, void *data)
{	/* nodefunc for freeing a track from a tracklist */
    TRACK *trk = (TRACK *)trknode;

    assert(type == T_TRACK);
    tkFreeAll(trk);
    return 1;
}

void tklFreeAll(TKLIST tkl)
{	/* free all the nodes and tracks and peaklists in a list */
    free_homo_list(tkl, tk_free_all);
}

TKLIST tkPruneFix(TKLIST *ptrax, double srate, double magthresh, int tkfmt)
{	/* go through *all* tracks, discard those we don't want
	   fixup the frequency tracks of those we do.
	   Need to know tkfmt in order to know how to free the track.. */
    TRACK 	*trk;
    TrackStats 	*tks;
    TKLIST	*tktp = ptrax;
    int		complained = 0;

    while(*tktp)  {
	trk = (TRACK *)car((LIST)*tktp);
	tks = &(trk->st);
	if(tks->favg > srate || tks->mavg < magthresh)  {
	    /* we don't want it */
	    free_next_ptr((LIST *)tktp);
	    if(tkfmt == TKFMT_NOPHS)
		free(trk->tktop);	/* all pks alloc'd as one */
	    else
		pklistFree(trk->tktop);
	    tkFree(trk);
	} else {
	    /* step over this track, it's OK */
	    /* add pseudo-points, generally munge */
	    if(trk->tktop->peak->mag == V_SMALL_MAG)  {
		/* special `zero' magnitude for first peak indicates 
		   tracks already post-processed */
		if(complained == 0)  {
		    fprintf(stderr, "Yo man, these tracks *already* fixed\n");
		    complained = 1;
		}
	    } else {
		tkFixupFrqs(trk);
		tkFadeTrk(trk);
	    }
	    advance_next(&tktp);
	}
    }
    return *ptrax = qsort_list(*ptrax, tkCmpTrxByT);
}

#ifndef FOR_MATLAB     /* no synth for matlab - don't want to link sndf/play */

static AUSTRUCT *au = NULL;

void tkSynthTrk(TRACK *track, float *buf, long n, double tbase)
{	/* overlay synthesized audio from the track 
	   into the float buf of length n starting at time tbase.
	   We abuse the track->tkbot field to point the 'current' 
	   source peak i.e. that immediately preceding tbase,
	   else assert error.  Also store 'current' phase in track->startPhase.
	   Returns with <track> setup for next call. */
    PeakPos	*pk0, *pk1;
    double	t, t0, t1, dt, w0, w1, dw, m0, m1, ph0, phi, mag;
    double	twopi = 2.0*PI;
    double	speriod = 1.0/srate;
    long 	done;
    PKLIST	pkl;
    PHMSKSETUP;
 
    pkl = track->tkbot;
    assert(pkl);
    /* assume that 'pkl' is already wound on to the right spot */
    if(pkl->next == NULL)  {	/* hit end of this track - no action */
    	return;
    }
    /* check that tbase really is in current interval */    	
    assert(pkl->peak->tim <= tbase && tbase < pkl->next->peak->tim);
    
    /* preset lower bound of first interval for loop */
    pk0 = pkl->peak;
    t0  = pk0->tim;   w0 = twopi*pk0->frq;  m0 = LINOFDB(DBOFST+pk0->mag);
    ph0 = track->startPhase;
    pk1 = pkl->next->peak;
    t1 = pk1->tim;  w1 = twopi*pk1->frq;  m1 = LINOFDB(DBOFST+pk1->mag);

    done = 0;
    t = tbase;
    while(done < n && pkl->next)  {	/* loop over tk segment intervals */
	dt = t1 - t0;
	dw = w1 - w0;
#ifdef CONST_FREQ
	dw = 0;		/* CONSTANT FREQUENCY */
#endif /* CONST_FREQ */
	while(t < t1 && done < n)  {	/* loop over samples within interval */
	    phi = ph0 + w0*(t-t0) + dw/dt*(t-t0)*(t-t0)/2;
	    mag = m0 + (m1-m0)/dt*(t-t0);
#ifdef USE_TSIN
	    *buf += mag*tabCos(phi);
#else /* !USE_TSIN */
	    *buf += mag*cos(phi);
#endif /* !USE_TSIN */
	    ++buf;  ++done;
	    t = tbase + done*speriod;
	}
	/* changed  if(t >= t1)   to  while(t >= t1)  
	   to deal with case when span of certain segments may be less 
	   than 1 sample point(!!) i.e. the time advance of a single sample 
	   may no only move <t> outside the current (tn, tn+1) interval, but 
	   may even go beyond tn+2 to necessitate stepping through to 
	   (tn+2, tn+3)  - doesn't happen normally, but can 
	   with warped tracks as generated by Stan's stuff */
	while(pkl->next && t >= t1) {	/* hit end of seg, not end of buf */
	    ph0 += w0*dt + dw*dt/2;	/* should save ph0 somewhere */
	    ph0 = PHMSK(ph0);
	    t0 = t1;
	    w0 = w1;
	    m0 = m1;
	    pkl = pkl->next;
	    if( (pkl->next) != NULL)  {
		pk1 = pkl->next->peak;
		t1 = pk1->tim;  
		w1 = twopi*pk1->frq;  
		m1 = LINOFDB(DBOFST+pk1->mag);
	    }
/*
	    pkl->peak->phs = PHMSK(pkl->peak->phs);
	    if( fabs(ph0 - pkl->peak->phs) > .01)  {
		fprintf(stderr, "phase deviation %6.3f != %6.3f\n",
			ph0, pkl->peak->phs);
	    }
 */	/* never fired for gould1.trx (13917 trax) */
	    track->startPhase = ph0;	/* NEED to remember .. */
	    track->tkbot = pkl;
	}
    }
    return;
}
 
void tkRunOscs(TKLIST going, double *pt0, double t1, 
	       float *buf, long *fill, long size)
{	/* synthesize the tracks listed in 'going'.
	   Keep running until time (*pt0) is within one sample of t1
	   or until buffer pos (*fill) is at end (size) */
    TRACK	*trk;
    long 	n = 1+(t1 - *pt0)*srate;	/* n steps is just beyond t1 */
    long 	m = size - *fill;	/* m steps to fill buf */
    long	z = MIN(m,n);
 
    while(going)  {
	trk = (TRACK *)CAR(going);
	/* run this track across the appropriate interval */
	tkSynthTrk(trk, buf+ *fill, z, *pt0);
	going = cdr(going);
    }
    *pt0 += z*speriod;
    *fill += z;
}
 
#define BUFSIZE 1024

int tkCheckPkl(LIST pkl, int type, void *arg)
{	/* a debug routine I once used */
/*    PrintPeak((PeakPos *)pkl, "pk");   */
    if( (((PeakPos *)pkl)->tim) > 5.034187 
       && (((PeakPos *)pkl)->tim) < 5.034189) {
	PrintPeak((PeakPos *)pkl, "");
	fprintf(stderr, "%lx\n", pkl);
    }
    if(type == T_PEAK)
	return 1;
    else {
	fprintf(stderr, "tkCheckPkl: car %lx type %d\n", pkl, type);
	return 0;
    }
}

int tk_print_onset(TRACK *trk, int type, void *arg)
{	/* MAPpable function prints start time */
    fprintf(stderr, "trk%6ld: t %9.6f f %8.1f\n", trk->id, trk->st.tmin, trk->st.favg);
    return 1;
}

int tklPrintOnsets(TKLIST trkl)
{	/* debug function summarizes start times of tracks */
    map((LIST)trkl, (NodeFunc)tk_print_onset);
    return 0;
}

long tkSynthTrax(TKLIST intrax, double samplingrate, SOUND *snd, void *audio)
{	/* synth from a track list, writing to a SOUND file and/or 
	   real-time audio output */
    LIST	going = NULL, *gotail, *ago, tmp;
    TRACK	*track;
    TKLIST 	trax = intrax;	/* (just want to preserve intrax for debug) */
    double	t0 = 0.0, t1;
static float	*buf = NULL;
static short	*sbuf;
    long	bufsize, fill = 0, tpsiz, frames = 0;

    if(samplingrate == 0.0)	samplingrate = DFLT_SRATE;
    srate = samplingrate;
    speriod = 1.0/samplingrate;

    bufsize = BUFSIZE;
    if(buf == NULL)  {
	buf = TMMALLOC(float, bufsize, "tkSynthTrax:float");
	sbuf = TMMALLOC(short, bufsize, "tkSynthTrax:short");
    }
    VZero(buf, 1, bufsize);

#ifdef USE_TSIN		/* make sure lookup table is set up */
    tabInit(0);
#endif /* USE_TSIN */

    /* what to do with start time? 
       Some tracks get 'pushed back' into negative time
       Some track sets have nothing for a while - why synth it?
       Sometimes, want output aligned to partic. time. 
       - Should find 'tracks and tails' for start time & synth from 
         them;  for now, start at first track's starttime */
 
/*    tklPrintOnsets(trax);	*/
    fprintf(stderr, "tkSynthTrax: sorting trax..\n");
    trax = qsort_list(trax, tkCmpTrxByT);

    t0 = ((TRACK *)CAR(trax))->st.tmin - 0.1;	/* 0.1s lead-in */
    fprintf(stderr, "tkSynthTrax: starting at t=%.6f\n", t0);
    gotail = &going;
    while(trax)  {
	track = (TRACK *)CAR(trax);
	trax = CDR(trax);
	t1 = track->tktop->peak->tim;		/* start time for next track */
	assert(t1>t0-speriod);	/* i.e. tracks passed in were sorted in time */
#ifdef DEBUG
	if(dbgtrk && t1 > dbgtime)  {
	    fprintf(stderr, "at %.6f dbg_ts = %ld:\n", t1, 
		    top_size((LIST)dbgtrk->tktop));
	    /* tkPrint(dbgtrk, "dbg"); */
	    /* map_with_abort(dbgtrk->tktop, tkCheckPkl); */
	    dbgtime += 1.0;
	}
#endif /* DEBUG */
	while(t0 < t1)  {
	    tkRunOscs(going, &t0, t1, buf, &fill, bufsize);
	    if(fill == bufsize)  {
	      if(snd)  			sndWriteFrames(snd, buf, bufsize);
	      Float2Short(buf, 1, sbuf, 1, bufsize);	/* copy & convert */
	      if(audio)			AUWrite(au, sbuf, bufsize);
	      VZero(buf, 1, bufsize);
	      frames += bufsize;
	      fill = 0;
	    }
	}	    /* hit t1 i.e. time to add new track */
	/* initialize track so that tkbot field holds 'current' peak ival */
	track->tkbot = track->tktop;
	/* add next track onto front of 'going' list */
	going = new_node(going, (LIST)track, T_TRACK);
	/* strip out finished tracks from 'going' list */
	ago = &going;
	while((*ago))  {
	    if( ((TRACK *)CAR(*ago))->tkbot->next == NULL )  {	
		/* finished track, lose it, close up 'going' list */
		tmp  = *ago;
		*ago = cdr(tmp);
	/*	free_node(tmp);	*/
		/* *ago now points to a new node, no need to advance it */
	    } else {	/* this node passed, so step on *ago */
		advance_next(&ago);
	    }
	}
    }
    /* no new trax, but some still going */
    while(going)  {
	tkRunOscs(going, &t0, t0+60.0 /* say */, buf, &fill, bufsize);
	assert(fill == bufsize);	/* no time end, just fill buf */
	if(snd)				sndWriteFrames(snd, buf, bufsize);
	Float2Short(buf, 1, sbuf, 1, bufsize);	/* copy & convert */
	if(audio)			AUWrite(au, sbuf, bufsize);
	VZero(buf, 1, bufsize);
	frames += bufsize;
	fill = 0;
	/* strip out finished tracks from 'going' list */
	ago = &going;
	while((*ago))  {
	    if( ((TRACK *)CAR(*ago))->tkbot->next == NULL )  {	
		/* finished track, lose it, close up 'going' list */
		tmp  = *ago;
		*ago = cdr(tmp);
		free_node(tmp);
		/* *ago now points to a new node, no need to advance it */
	    } else {	/* this node passed, so step on *ago */
		ago = next_ptr(*ago);
	    }
	}
    }
    /* flush output */
    if(snd)  {
      sndWriteFrames(snd, buf, fill);
      sndClose(snd);
    }
    Float2Short(buf, 1, sbuf, 1, fill);	/* copy & convert */
    if(audio)  {
      AUWrite(au, sbuf, fill);
      AUClose(au, 0);
    }
    frames += fill;
    return frames;
}
#endif /* !FOR_MATLAB */

TKLIST tklFromPkls(LIST pkls)
{	/* convert entire list of peaklists to list of tracks */
    PKLIST pkl;
    TKLIST root = NULL, *tklTail;

    tklTail = &root;
    while(pkls)  {
	pkl = (PKLIST)CAR(pkls);
	pkls = CDR(pkls);
	assert(CARTYPE(pkl) == T_PEAK);
	*tklTail = (TKLIST)new_node(NULL, (LIST)tkNew(pkl), T_TRACK);
	advance_next((LIST **)&tklTail);
    }
    return root;
}

int tkOldSave(TKHDSTRUCT *trkh, TKLIST trax, char *filename)
{     	/* write out the trax to the file so named - old format */
    FILE *outFile;
    
    trkh->format = TKFMT_NOPHS;		/* this old format */
    if( (outFile = TKOpenWrHdr(filename, trkh)) == NULL)  {
	fprintf(stderr, "tkSave: could not create '%s' for trx output\n", 
		filename);
	return 0;
    } else {
	WriteTrackFile(trax, outFile);
	TKCloseWr(outFile, trkh);
	printf("tkSave: wrote tracks to '%s'\n", filename);
	return 1;
    }
}

int pkWriteSmall(PeakPos *pk, int type, FILE *file)
{	/* map-able function writes out (small) peak records */
    assert(type == T_PEAK);
    /* kinda naughty - just *know* that 1st three floats in 
       struct are all we need - write explicitly */
    return ( fwrite(pk, sizeof(float), 3, file) > 0 );
}

int tkWriteTrack(TRACK *track, int type, FILE *file)
{	/* map-able function writes out entire track to file */
    assert(type == T_TRACK);
    /* paranoid, expensive check: (but can't load unless true) */
    assert( top_size((LIST)track->tktop) == track->st.npts);	
    fwrite(track, sizeof(TRACK), 1, file);
    maparg((LIST)track->tktop, (NodeFunc)pkWriteSmall, file);
}

int tkSave(TKHDSTRUCT *trkh, TKLIST trax, char *filename)
{     	/* write out the trax to the file so named - new format */
    FILE *outFile;
    
    trkh->format = TKFMT_NOPHS;		/* the new format */
    if( (outFile = TKOpenWrHdr(filename, trkh)) == NULL)  {
	fprintf(stderr, "tkSave: could not create '%s' for trx output\n", 
		filename);
	return 0;
    } else {
	maparg(trax, (NodeFunc)tkWriteTrack, outFile);
	TKCloseWr(outFile, trkh);
	printf("tkSave: wrote tracks to '%s'\n", filename);
	return 1;
    }
}

void tkLoadOldSub(TKLIST *ptrax, FILE *inFile)
{	/* load rest of old-style trax file after header */
    LIST	pkls;

    ReadTrackFile(&pkls, inFile);
    fclose(inFile);
    *ptrax = tklFromPkls(pkls);
    free_list(pkls);
    printf("tkLoadOldSub: Read %ld tracks\n", 
	   top_size(*ptrax));
}

extern int bytemode;	/* global set up by 'TKOpen..' indicates source */

PKLIST pkLoadSmallsBM(FILE *file, long n, PKLIST *plast, int bytemode)
{	/* load <n> small peaknodes from <file>, allocate new pklist.
	   File treated as in <byteorder>. */
    PKLIST 	rslt = NULL, this, last = NULL, *run;
    MinPeakPos	*mpk;
    long	i;

    /* this one rewritten for speed: no calls in inner loop, one malloc */
    run = &rslt;
    this = (PKLIST)TMMALLOC(char, n * (sizeof(NODE)+sizeof(MinPeakPos)), 
			    "pkLoadSmallsQuick");
    for(i = 0; i<n; ++i)  {
	*run = this;
	mpk = (MinPeakPos *)(this+1);
 	this->peak = (PeakPos *)mpk;	
	this->type = T_PEAK;	
	run = &(this->next);
	assert(fread(mpk, sizeof(float), 3, file));
	if(bytemode != INORDER)  {
	    ConvertBufferBytes((char *)mpk, sizeof(float)*3, 
			       sizeof(float), bytemode);
	    /* hope floats are just byte manipulated .. ? */
	}
	last = this;
	this = (PKLIST)(mpk+1);
    }
    last->next = NULL;
    *plast = last;	/* store pointer to last node in chain (why?) */
    return rslt;
}

PKLIST pkLoadSmalls(FILE *file, long n, PKLIST *plast)
{	/* load <n> small peaknodes from <file>, allocate new pklist.
	   Assumes file is in correct byteorder. */
    PKLIST 	rslt = NULL, this, last = NULL, *run;
    MinPeakPos	*mpk;
    long	i;

    /* this one rewritten for speed: no calls in inner loop, one malloc */
    run = &rslt;
    this = (PKLIST)TMMALLOC(char, n * (sizeof(NODE)+sizeof(MinPeakPos)), 
			    "pkLoadSmallsQuick");
    for(i = 0; i<n; ++i)  {
	*run = this;
	mpk = (MinPeakPos *)(this+1);
 	this->peak = (PeakPos *)mpk;	
	this->type = T_PEAK;	
	run = &(this->next);
	assert(fread(mpk, sizeof(float), 3, file));
	last = this;
	this = (PKLIST)(mpk+1);
    }
    last->next = NULL;
    *plast = last;	/* store pointer to last node in chain (why?) */
    return rslt;
}

TRACK *tkLoadTrack(FILE *file)
{	/* try and load a track off disk; return NULL if run out */
    TRACK *trk;

    trk = TMMALLOC(TRACK, 1, "tkLoadTrack");
    if( fread(trk, sizeof(TRACK), 1, file) == 0) {	/* hit eof */
	/* free(trk);	*/
	return NULL;
    }
    if(bytemode != INORDER)  {
	ConvertBufferBytes((char *)trk, sizeof(TRACK), 
			   sizeof(float), bytemode);
	/* hope floats are just byte manipulated .. ? */
	trk->tktop = pkLoadSmallsBM(file, trk->st.npts, &trk->tkbot, bytemode);
    } else {
	trk->tktop = pkLoadSmalls(file, trk->st.npts, &trk->tkbot);
    }
#ifdef DEBUG 
    if(trk->id == 1731)  {
	dbgtrk = trk;
	tkPrint(dbgtrk, "Debug");
    }
#endif /* DEBUG */
    return trk;
}

void tkLoadSub(TKLIST *ptrax, FILE *inFile)
{	/* load rest of new-style trax file after header */
    TKLIST	*ptrx = ptrax;
    int		done = 0;
    TRACK	*trk;

    fprintf(stderr, "(BM%d)", bytemode);	fflush(stderr);
    while(!done)  {
	if( (trk = tkLoadTrack(inFile)) == NULL)  {	/* must be EOF */
	    done = 1;
	} else {	/* managed to load another track */
	    *ptrx = (TKLIST)new_node(NULL, (LIST)trk, T_TRACK);
	    advance_next(&ptrx);
	}
    }
    fclose(inFile);
    printf("tkLoadSub: Read %ld tracks\n", 
	   top_size(*ptrax));
}

TKHDSTRUCT *tkLoad(TKLIST *ptrax, char *filename)
{	/* load the specified file into a trackstruct in memory, 
	   return head of list or NULL if failed.  */
    FILE	*inFile;
    TKHDSTRUCT 	*trkh;

    if( (inFile = TKOpenAllocRdHdr(filename, &trkh)) == NULL)  {
	fprintf(stderr, "tkLoad:can't open trax file '%s'\n", filename);
	return NULL;
    }
    fprintf(stderr, "tkLoad: '%s' open..", filename);	fflush(stderr);
    if(trkh->format == TKFMT_NOPHS)
	tkLoadSub(ptrax, inFile);
    else
	tkLoadOldSub(ptrax, inFile);
    if(srate == 0)  {
	srate = trkh->samplingRate;	/* global used in synth */
    }
    speriod = 1.0/srate;
    return trkh;
}

TKHDSTRUCT *tkLoadPrune(LIST *ptrax, char *filename, double magthresh)
{	/* load the specified file into a trackstruct in memory, 
	   apply our pruning and frequency fixup factors, 
	   return head of list or NULL if failed.  */
    TKHDSTRUCT 	*trkh;

    if(trkh = tkLoad(ptrax, filename))  {
	tkPruneFix(ptrax, trkh->samplingRate, magthresh, trkh->format);
	printf("%ld tracks after prune & fixup\n", top_size(*ptrax));
    }
    return trkh;
}

#ifndef FOR_MATLAB 	/* no synth for matlab, so don't link to sndf, play */
long tkWriteSynth(TKLIST trax, double srate, int playq)
{	/* synthesize a soundfile from the passed list of tracks, 
	   output at the specified sampling rate. 
	   Will try to play it too if <playq> */
    SOUND 	*snd = sndNew(NULL);
    char	fname[32];
    long 	frames;
    char	syscmd[64];

    sndSetSrate(snd, srate);
    sndSetChans(snd, 1);
    sprintf(fname, "%s%s", "tmp.rs", sndfExtn);
    sndOpen(snd, fname, "w");
    frames = tkSynthTrax(trax, srate, snd, NULL);
    sndClose(snd);
    fprintf(stderr, "Wrote sound '%s' (%ld frames)\n", fname, frames);
    if(playq)  {
	sprintf(syscmd, "play %s", fname);
	system(syscmd);
    }
    return frames;
}
#endif /* !FOR_MATLAB */

#ifdef TKMAIN	/****************** standalone program version **********/

#define  DEF_EXT	".rs"	/* default output extension */
#define  DFLT_MAGTHRESH	(-20.0)	/* default floor of usable track mag (dB) */

char defExt[32];
char defTxExt[32];

       char	*programName;
static int	verb = 0;		/* -v flag */
       int	graphs = 0;		/* use debugging graphs */
static float    gain  = 1.0;		/* scale output by this */
static char	*inPath = NULL;
static char	*outPath = NULL;
static char	*txOutPath = NULL;

static int      doplay = 0;
static int      dosave = 0;
static int 	dotxwr = 0;	/* flag to rewrite trax file */
static int 	profdbg = 0;	/* flag to exit after reading - for profile */

static double	magthresh = DFLT_MAGTHRESH;

/* to placate pklist.c */
       float dbgFrq;
       float dbgFrqWin;

/* static function prototype */
static void tkPlay(char *inPath, char *outPath);
static void usage(void);

static void tkPlay(char *inPath, char *outPath)  
{	/* load in track file, write out pruned tracks 
	   and play or write out synthesized sound */
    SOUND	*snd = NULL;
    TKHDSTRUCT	*trkh;
    LIST	trax, tmp;

    trkh = tkLoadPrune(&trax, inPath, magthresh);

    if(profdbg)
	exit(0);	/* (for profiling load fns) */

    if(top_size(trax) < 10)  {
	fprintf(stderr, "And they are...\n");
	tmp = trax;
	while(tmp)  {
	    PrintPeakList((PKLIST)car(tmp), "pklist");
	    tmp = cdr(tmp);
	}
    }
    if(dotxwr) {	/* write out revised tracks */
	tkSave(trkh, trax, txOutPath);
    }
    if(dosave)  {	/* open up output sound file */
      snd = sndNew(NULL);
      sndSetChans(snd, 1 /* trkh->channels */);
      sndSetSrate(snd, srate);
      sndOpen(snd, outPath, "w");
    }
    if(doplay)
	au = AUOpen(SFMT_SHORT, srate, 1, 0, AIO_OUTPUT);

    tkSynthTrax(trax, srate, snd, (void *)doplay);
}

main(argc, argv)
    int 	argc;
    char	**argv;
{	
    int 	err;

    srate = 0.0;	/* indicates use default */
    programName = argv[0];
    sprintf(defExt, "%s%s", DEF_EXT, sndfExtn);
    sprintf(defTxExt, "%s%s", DEF_EXT, ".trx");
    err = HandleArgs(argc, argv);	/* extract options and filenames */
 
    /* other tests on args */
    if(err || inPath == NULL)
	usage();		/* never returns */
 
    if(dosave && outPath == NULL)	{
	outPath=
	    TMMALLOC(char, (1+strlen(inPath)+strlen(defExt)), "outPath");
	strcpy(outPath, inPath);
	SubstExtn(inPath, defExt, outPath);
    }
 
    if(dotxwr && txOutPath == NULL)	{
	txOutPath=
	    TMMALLOC(char, (1+strlen(inPath)+strlen(defTxExt)), "txPath");
	strcpy(txOutPath, inPath);
	SubstExtn(inPath, defTxExt, txOutPath);
    }
 
/*    fprintf(stderr,"%s: verbose = %s\n",programName,verb?"yes":"no");
    fprintf(stderr,"%s: infile %s, outfile %s\n",programName,inPath,outPath);
 */
    /* args all dealt with.  On to the real business */
    tkPlay(inPath, outPath);
}
 
int HandleArgs(argc, argv)
    int 	argc;
    char	**argv;
{	/* go through command line args and set up global variables.
	   Return != 0 if error. */	
    int 	i = 0;
    int 	err = 0;
 
    /**** %% DPWE STD ARG HANDLER %% ****/
    i = 0;
    while(++i<argc && err == 0)	{
	char c, *token, *arg, *nextArg;
	int  argUsed;
 
	token = argv[i];
	if(*token++ == '-')  {
	    if(i+1 < argc)	nextArg = argv[i+1];
	    else		nextArg = "";
	    argUsed = 0;
	    while(c = *token++)  {
		if(NumericQ(token))	arg = token;
		else			arg = nextArg;
		switch(c)  {
	        case 'r':		srate = atof(arg);	argUsed = 1;
		    if(srate<4000||srate>100000)  {  
			fprintf(stderr,"%s: sr of %s = %.2f not 4k..100k\n",
				programName, arg, srate);
			err = 1;
		    }
		    break;
	        case 'g':		gain = atof(arg);	argUsed = 1;
		    if(gain>100000)  {
			fprintf(stderr,"%s: gain of %s = %.2f not <100k\n",
				programName, arg, gain);
			err = 1;
		    }
		    break;
	        case 'm':		magthresh = atof(arg);	argUsed = 1;
		    if(magthresh > 100.0 || magthresh < -100.0)  {
			fprintf(stderr,
				"%s: mag thresh of %s = %.2f not -100..100\n",
				programName, arg, magthresh);
			err = 1;
		    }
		    break;
	        case 't':		dbgtime = atof(arg);	argUsed = 1;
		    break;
		case 'v': 	verb = 1; break;
		case 'p': 	doplay = 1; break;
		case 'w': 	dosave = 1; break;
		case 'x': 	dotxwr = 1; break;
		case 'd': 	profdbg = 1; break;
		default:	fprintf(stderr,"%s: unrec option %c\n",
					programName, c);
		                err = 1; break;
		    }
		if(argUsed)  {
		    if(arg == token)	token = "";   /* no more from token */
		    else		++i;	      /* skip arg we used */
		    arg = ""; argUsed = 0;
		}
	    }
	} else {
	    if(inPath == NULL)		inPath = argv[i];
	    else if(outPath == NULL)	{  outPath = argv[i];  dosave = 1;  }
	    else { fprintf(stderr,"%s: excess arg %s\n", programName, argv[i]);
		   err = 1;  }
	}
    }
    return err;
}
 
static void usage(void)		/* print syntax & exit */
    {
    fprintf(stderr,
 "Usage: %s [-v][-r srate] inTrx [outSNDF]\n",
	    programName);
    fprintf(stderr,"where  (defaults in brackets)\n");
    fprintf(stderr,
 " -v         verbose mode\n");
    fprintf(stderr,
 " -d         debug mode - just load the trax and exit (for profile)\n");
    fprintf(stderr,
 " -p         play sound in real time\n");
    fprintf(stderr,
 " -w         force write of sound to disk\n");
    fprintf(stderr,
 " -m mag     dB threshold below which tracks are supressed (%.1f)\n", 
	    DFLT_MAGTHRESH);
    fprintf(stderr,
 " -t time    debug interruption time\n");
    fprintf(stderr,
 " -x         rewrite modified track file (as instem+'%s')\n", defTxExt);
    fprintf(stderr, 
 " -r srate   sampling rate for output file (from file)\n");
    fprintf(stderr,
 " inTrx      is a Track file for input\n");
    fprintf(stderr,
 " outSNDF    is the output soundfile       (inStem+'%s')\n", defExt);
    exit(1);
    }
#endif /* TKMAIN */
