/***************************************************************\
*	trk.c			  				*
*   convert a constant-Q spectrogram file into a structured 	*
*   list of 'tracks' (mag/frq contours in time)			*
*   Based on routines developed within dspB, for thesis		*
*   dpwe 08jul91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#include <dpwelib.h>
#include <assert.h>	/* ansi debugs */
/* personal libs */
#include <genutils.h>	/* for NumericQ() */
#include <readaqt.h>
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include <vecops.h>
#include "peaks.h"
#include "tracks.h"
#include "trkmaint.h"
#include "trkheadio.h"
#include "exspec.h"

#ifdef __sgi
/* fancy FPE traps */
#include <limits.h>
#include <sigfpe.h>
#endif /* __sgi */

#ifdef THINK_C
#include "tc_main.h"
char sfdir_path[] = "behemoth:Users:dpwe:Sound";
#endif /* THINK_C */

/* constants */
#define DFLT_EXTN 	".trx"	/* default output file extension */
#define DFLT_FUDG	1.0	/* default for generic fudge factor */
#define MYINFOSIZE	32	/* PVinfo used for "Q=.." and "octs=.." */

/* static function prototype */
static void diess PARG((char *s1, char *s2));
static void PerformTracking PARG((char *inPath, char *outPath));
static void usage PARG((void));

/* static data */
       char	*programName;
static int	verb = 0;		/* -v flag */
       int	graphs = 0;		/* use debugging graphs */
       float    maxFreq = 0.0;		/* don't do tracks above here */
       float 	maxTime = 0.0;		/* don't do tracks beyond here */
       int	expect = 0;		/* flag to do expectation stuff */
       float	fudge = DFLT_FUDG;		/* generic fudge factor */

/* print out data on trks in this window */
float dbgFrq = -2000.0;
float dbgFrqWin = 5.0;


/* functions */

main(argc,argv)
    int 	argc;
    char	**argv;
    {
    int 	i = 0;
    int 	err = 0;
    char	*inPath = NULL, *outPath = NULL;

    programName = argv[0];
    /**** %% DPWE STD ARG HANDLER %% ****/
    i = 0;
    while(++i<argc && err == 0)
	{
	char c, *token, *arg, *nextArg;
	int  argUsed;

	token = argv[i];
	if(*token++ == '-')
	    {
	    if(i+1 < argc)	nextArg = argv[i+1];
	    else		nextArg = "";
	    argUsed = 0;
	    while(c = *token++)
		{
		if(NumericQ(token))	arg = token;
		else			arg = nextArg;
		switch(c)
		    {
/*	        case 'q': case 'Q':	qFactor = atof(arg);	argUsed = 1;
/*		    if(qFactor<1)
/*			{  fprintf(stderr,"%s: Q of %s = %.2f must be > 1\n",
/*				programName, arg, qFactor);
/*			   err = 1;  	}
/*		    break;
 */
		    case 'f':	maxFreq = atof(arg);	argUsed = 1;	break;
		    case 't':	maxTime = atof(arg);	argUsed = 1;	break;
		    case 'u':	fudge   = atof(arg);	argUsed = 1;	break;
		    case 'x':	expect = 1;		break;
#ifdef DEBUGGRAPHS
		case 'd':     	graphs = 1;  break;
#endif /* DEBUGGRAPHS */
		case 'v': 	verb = 1; break;
		default:	fprintf(stderr,"%s: unrec option %c\n",
					programName, c);
		                err = 1; break;
		    }
		if(argUsed)
		    {
		    if(arg == token)	token = "";   /* no more from token */
		    else		++i;	      /* skip arg we used */
		    arg = ""; argUsed = 0;
		    }
		}
	    }
	else{
    if(inPath == NULL)		inPath = argv[i];
	    else if(outPath == NULL)	outPath = argv[i];
	    else{ fprintf(stderr,"%s: excess arg %s\n", programName, argv[i]);
		  err = 1;  }
	    }
	}
    /* other tests on args */
/*    if(data < 0)
 *	{ fprintf(stderr,"%s: bad data %d: must be >= 0\n", programName, data);
 *	  err = 1; }
 */
    if(err || inPath == NULL)
	usage();		/* never returns */

    if(outPath == NULL)
	{
	outPath=
	    TMMALLOC(char, (1+strlen(inPath)+strlen(DFLT_EXTN)), "outpath");
	strcpy(outPath, inPath);
	SubstExtn(inPath, DFLT_EXTN, outPath);
	}

/*    fprintf(stderr,"%s: data = %d\n",programName,data);
 *    fprintf(stderr,"%s: verbose = %s\n",programName,verb?"yes":"no");
 *    fprintf(stderr,"%s: infile %s, outfile %s\n",programName,inPath,outPath);
 */
    /* args all dealt with.  On to the real business */

#ifdef __notsgi
    /* set up FPE traps */
    sigfpe_[_INVALID].count = INT_MAX;
    sigfpe_[_INVALID].trace = 1;
    sigfpe_[_INVALID].abort = 1;
    handle_sigfpes(_DEBUG, _EN_UNDERFL|_EN_OVERFL|_EN_DIVZERO|_EN_INVALID, 
		   0, _ABORT_ON_ERROR, 0);
	
#endif /* __sgi */
    PerformTracking(inPath, outPath);
    }

static void diess(s1, s2)
    char *s1;
    char *s2;
    {
    char *s3 = "";

    if(s1 == NULL)	s1 = s3;
    if(s2 == NULL)	s2 = s3;
    fprintf(stderr, "%s: %s : %s\n", programName, s1, s2);
    exit(0);
}

#ifdef DEBUGGRAPHS
static float	flubber[1024];	/* big enough for any oct/bin product */

void CopyAndDisp(float *src, int sinc, long npts, WINDAT *win, char *str)
{
    CopyFloats(src, sinc, flubber, 1, npts);
    if(str && strlen(str) > 0)
	SetLabel(win, str);	
    DoDisp(win, flubber, npts);
}

void DispComplex(cpxpolar *src, int sinc, long npts, 
		 WINDAT *mw, WINDAT *pw, char *str)
{
    CopyAndDisp((float *)src, 2*sinc, npts, mw, str);
    CopyAndDisp(((float *)src)+1, 2*sinc, npts, pw, str);
}

#endif /* DEBUGGRAPHS */

extern PKLIST ExpectSpect();

static void PerformTracking(inPath, outPath)
    char *inPath;
    char *outPath;
    {
    CQTINFO 	*cqi;
    TKHDSTRUCT	*trkh;
    FILE	*inFile, *outFile;
    int 	i, err, done=0, ndone=1;
#ifdef DEBUGGRAPHS
    WINDAT	*wm1, *wp1, *wm2, *wp2, *wm3, *wp3;
#endif
    char	desc[DESCLEN];
    char 	mystr[32];
    double	time, dt, duration, timeFrame, timeFix;
    long	tick = 0;
    TKLIST	currentTails;
    LIST	allTrax;	/* final tracks just PKLISTs */
    LIST	*EOallTrax;
    PKLIST	currentPeaks, strayPeaks;
    int 	totBins;
    cpxpolar 	*ans, *espec, *dspec;
    float	*ansmag, *anspha;
    float 	binsPerOct, lowestFrq, highestFrq;
    int 	debug = 0;
    float	dbgTime = 0.1;

    allTrax = NULL;
    EOallTrax = &allTrax;	/* end of allTrax is the root */
    currentTails = NULL;
    /* open input CQT file */
    if( (inFile = OpenCQTPVfile(inPath, &cqi))== NULL)
	PVDie(PVE_NOPEN, inPath);
    /* check that we can open output file before we spend too long */
    if( TKAlloc(&trkh, TKFMT_BINARY, cqi->pvh->samplingRate, 
		cqi->duration, cqi->pvh->channels, TKHDDFLTBYTS) )
	diess("TkHead alloc fail", NULL);
    if( (outFile = TKOpenWrHdr(outPath, trkh)) == NULL)
	diess("Could not open output file", outPath);

    totBins  = cqi->nBins * cqi->nOcts;
    duration = cqi->duration;
    if(maxTime > 0.0 && maxTime < duration)  {
	/* we will not process the whole file */
	duration = maxTime;
    }
    ans = TMMALLOC(cpxpolar, totBins, "ans");
    espec = TMMALLOC(cpxpolar, totBins , "espec");
    dspec = TMMALLOC(cpxpolar, totBins, "dspec");
    ansmag = &(ans[0].mag);	anspha = &(ans[0].pha);
    if(maxFreq == 0.0 || maxFreq > cqi->pvh->maxFreq)  {
	/* our search stops at maxFreq */
	maxFreq = cqi->pvh->maxFreq;	/* .. which may be from data */
    }
    dt = 2.0/maxFreq;	/* worst Nyquist rate */
    tick = 0;
    timeFix = dt - .0000636; /* specific hack - 18sep91 */
    timeFrame = 0;
    binsPerOct  = (float)cqi->nBins;
    lowestFrq   = cqi->pvh->minFreq;
    highestFrq  = maxFreq;
    assert(highestFrq > lowestFrq);
#ifdef DEBUGGRAPHS
    if(graphs) {
	wm1 = NewWin("act mag", 1);    wp1 = NewWin("act phs", 0);
	wm2 = NewWin("est mag", 0);    wp2 = NewWin("est phs", 0);
	wm3 = NewWin("dif mag", 0);    wp3 = NewWin("dif phs", 0);
	wdLink(wm1, wp1);	wdLink(wp1, wm2);
	wdLink(wm2, wp2);	wdLink(wp2, wm3);
	wdLink(wm3, wp3);
    }
#endif /* DEBUGGRAPHS */
    while( (time = tick*dt + timeFix) < duration)
	{
	if(time>timeFrame)
	    {
	    fprintf(stderr,"%4.2f ",timeFrame);	fflush(stderr);
	    timeFrame += 0.1;
	    }
	SpecOfCQTFile(cqi, time, dt, ans);
	if(expect)  {
	    pklistFree(ExpectSpect(currentTails, time, espec, binsPerOct, 
				   cqi->nOcts, highestFrq, cqi->qFactor));
	    CVCopy(ans, 1, dspec, 1, totBins);
	    CVVSub(espec, 1, dspec, 1, totBins);	    
#ifdef DEBUGGRAPHS
	    if(graphs) {
		sprintf(mystr, "time %.6f", time);
		DispComplex(ans, 1, totBins, wm1, wp1, mystr);
		DispComplex(espec, 1, totBins, wm2, wp2, mystr);
		DispComplex(dspec, 1, totBins, wm3, wp3, mystr);
	    }
#endif /* DEBUGGRAPHS */
	}
 	currentPeaks = ListPeaks(ansmag, 2, anspha, 2, totBins, 
				 1, time, binsPerOct, lowestFrq, 
				 highestFrq, 1);
		if(debug && time > dbgTime)
		    PrintPeakList(currentPeaks, "currentPeaks");
	/* pass current tails to allow differencing ? */
		if(debug && time > dbgTime)
		    tklPrint(currentTails, "currentTrx");
	strayPeaks   = MatchPeakLists(currentTails, currentPeaks);
		if(debug && time > dbgTime)
		    tklPrint(currentTails, "matchTrx");
		if(debug && time > dbgTime)
		    PrintPeakList(strayPeaks, "strayPeaks");
	AdvanceOrTrim(&currentTails);
		if(debug && time > dbgTime)
		    tklPrint(currentTails, "trimTrx");
	EOallTrax = AddNewPeaks(&currentTails, EOallTrax, strayPeaks);
		if(debug && time > dbgTime)
		    tklPrint(currentTails, "addTrx");
		if(debug && time > dbgTime)
		    PrintPeakList((PKLIST)*EOallTrax, "EOallTrax");
		if(debug && time > dbgTime)
		    debug = 0;
	++tick;
	}
    TieUpLeftovers(&currentTails);
    WriteTrackFile(allTrax, outFile);
    TKCloseWr(outFile, trkh);
    fclose(inFile);
    }

static void usage PARG((void))		/* print syntax & exit */
    {
    fprintf(stderr,
 "Usage: %s [-v][-f maxFreq][-t maxTime][-x] inCQT [outTrx]\n",
	    programName);
    fprintf(stderr,"where  (defaults in brackets)\n");
    fprintf(stderr,
 " -v         verbose mode\n");
#ifdef DEBUGGRAPHS
    fprintf(stderr, 
 " -d         draw debug graphs\n");
#endif /* DEBUGGRAPHS */
    fprintf(stderr,
 " -x         calculated expected spectra (useless without -d)\n");
    fprintf(stderr,
 " -f maxFreq do not create tracks above this frequency\n");
    fprintf(stderr,
 " -t maxTime do not create tracks beyond this time\n");
    fprintf(stderr,
 " -u fudge   set the fudge factor (%.3f)\n", DFLT_FUDG);
    fprintf(stderr,
 " inCQT      is a CQT analysis file for input\n");
    fprintf(stderr,
 " outTrx     is the output file of trax        (inStem+'%s')\n", DFLT_EXTN);
    exit(1);
    }
