/***************************************************************\
*	trkmaint.c		  				*
*   "British rail regrets to announce that there will be a bus	*
*    link operating between Winchester and Basingstoke over the	*
*    weekend to allow for essential track maintenance."		*
*   Maintain track lists by trimming and adding			*
*   Based on routines developed within dspB, for thesis		*
*   dpwe 08jul91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */

/* std libs */
#include <dpwelib.h>
#include <assert.h>
#include <math.h>
/* personal libs */
#include <dsputil.h>
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include "peaks.h"
/* in "optpath.h" : #include "tracks.h"	/* for TRACK structure */
#include "optpath.h"	/* for finding best path through phase choices */
#include "trkmaint.h"

/* types */

/* constants */

/* static function prototype */
static void UpdateTRK PARG((TRACK *trk));

/* static data */

/* functions */

float FrqOfPeakMetric(NODE *node)	/* metric for sort functions */
{
    int  t;

    if( (t = cartype(node))!=T_PEAK)
	{  fprintf(stderr,"FOPMetric: not a peak (%d)\n",t);
	   abort();	}
    return ((PeakPos *)car(node))->frq;
}

float FrqOfTRKTailMetric(NODE *node)	/* metric for sort functions */
{
    int  t;

/*    fprintf(stderr, "FOK(%lx)", node); fflush(stderr);	*/
    if( (t = cartype(node))!=T_TRACK)	/* the parent node */
	{  fprintf(stderr,"FOKMetric: not a track (%d not %d)\n",t,T_TRACK);
	   abort();	}
    node = (LIST)((TRACK *)car(node))->tkbot;	/* the Tail ptr node */
/*    if( (t = cartype(node))!=T_LIST)
	{  fprintf(stderr,"FOKMetric: not a pknod (%d not %d)\n",t,T_LIST);
	   abort();	}	*/

    return FrqOfPeakMetric(node);
}

int CmpFrqOfTrk(NODE **newnode, NODE **cmpnode)
{	/* for insert_by_cmp, this will return zero until it sees a cmpnode 
	   which newnode should preceed */
    if(FrqOfTRKTailMetric(*newnode) < FrqOfTRKTailMetric(*cmpnode))
	return -1;
    else
	return 1;
}

void ListTrackPh(PKLIST track, char *title)
{	/* print out a track with BOTH sets of phases */
    PeakPos	*thisPk;
    PKLIST	trkNode;

    fprintf(stderr, "%s\n", title);
    trkNode = track;
    while(trkNode)  {
	thisPk = (PeakPos *)car((LIST)trkNode);
	fprintf(stderr,"f %8.2f m %6.2f p1 %6.3f p2 %6.3f t %7.5f\n",
		thisPk->frq, thisPk->mag, thisPk->phs, thisPk->ophs,
		thisPk->tim);
	trkNode = (PKLIST)cdr((LIST)trkNode);
    }
}

#ifdef DEBUGGRAPHS
static WINDAT *magWin = NULL;
static WINDAT *frqWin = NULL;
static WINDAT *favWin = NULL;
static WINDAT *fvrWin = NULL;
static float  *ddata = NULL;
#define DBGASIZE 1000
extern int graphs;	/* CL option from main */
#endif

void ProcessCompleteTrack(TRACK *track)
{
    int 	n;
    PKLIST	pnode;
    PeakPos	*peak;
    double	tmpPhs, delta;
    PATH	*path;

#ifdef DEBUGGRAPHS
    if(graphs)  {
	if(magWin == NULL)
	{
	    magWin = NewWin("Magnitude",1);
	    frqWin = NewWin("Frequency",0);
	    favWin = NewWin("Avg Frq",0);
	    fvrWin = NewWin("Frq vce",0);
	    ddata = TMMALLOC(float, DBGASIZE, "PCT:ddata");
	}
	pnode = track->tktop;	n = 0;
	while(pnode)
	{
	    ddata[n] = ((PeakPos *)car((LIST)pnode))->mag;
	    ++n;	pnode = (PKLIST)cdr((LIST)pnode);
	}
	DoDisp(magWin, ddata, n);
	pnode = track->tktop;	n = 0;
	while(pnode)
	{
	    ddata[n] = ((PeakPos *)car((LIST)pnode))->frq;
	    ++n;	pnode = (PKLIST)cdr((LIST)pnode);
	}
	DoDisp(frqWin, ddata, n);
    }
#endif /* DEBUGGRAPHS */

    path = BuildBestPath(track);
    FixTrackPerPath(path);
    FreePath(path);
}

static void UpdateTRK(TRACK *trk)
{   /* one new node is incorporated into trk stats */
    TrackStats *stats = &(trk->st);
    PKLIST newBottomPeakNode = (PKLIST)CDR((LIST)trk->tkbot);

    assert(newBottomPeakNode != NULL);	/* there is another node */
    tkStBoilup(stats);
    tkStUpdate(stats, newBottomPeakNode->peak, trk->tkbot->peak);
    tkStBoildown(stats);
    trk->tkbot = newBottomPeakNode;
}

void AdvanceOrTrim(TKLIST *pcurTrax)
{	/* just advance at moment : only trim dead tracks */
    TKLIST	*pNodePtr = pcurTrax;
    TRACK	*thisTk;
    PKNODE	*thisPkNode;
    PeakPos	*thisPk, *nextPk;
    float	fmax, dt;

    /* trimming would be _not_ extending new peaks in low freq tracks
       when they are very close to the preceding peak */
    /* also maybe should reject tracks that are very quiet -
       perhaps you need to do that retrospectively, though */
    while(*pNodePtr)  {
	thisTk = (TRACK *)car((LIST)*pNodePtr);
	thisPkNode = thisTk->tkbot;
	thisPk = (PeakPos *)car((LIST)thisPkNode);
	if( cdr((LIST)thisPkNode) )  {	/* new last element */
	    nextPk = (PeakPos *)car(cdr((LIST)thisPkNode));
	    fmax = MAX(thisPk->frq, nextPk->frq);
	    dt   = nextPk->tim - thisPk->tim;

	    if(dt < .5/(fmax/2.0) )  {	/* dep. on analysis Q */
		/* too soon : delete peak */
		FreePeak(nextPk);
		free_node(close_up_ptr(cdr_ptr((LIST)thisPkNode)));
		if(cdr((LIST)thisPkNode))	abort();
	    } else {	/* incorporate peak into track, advance tail ptr */
		UpdateTRK(thisTk);	/* avg fm, am etc */
	    }
	    advance_next(&pNodePtr);
	} else {			/* dying track */
	    ProcessCompleteTrack(thisTk);
	/*    tkFree(thisTk);	*/
	    free_node(close_up_ptr(pNodePtr));		/* remove track ptr */
	}
    }
    *pcurTrax = qsort_list2(*pcurTrax, CmpFrqOfTrk);
}

void TieUpLeftovers(TKLIST *pTails)
{	/* We have reached the end of the file: post-process live tracks */
    TKLIST	*pNodePtr = pTails;
    PKNODE	*newPeakNode;
    TRACK	*thisTk;
    PeakPos	*thisPk, *nextPk;
    double	fmax, dt;

    /* tracks will already be advanced */
    while(*pNodePtr)  {
	thisTk = (TRACK *)car((LIST)*pNodePtr);
	ProcessCompleteTrack(thisTk);
	/* tkFree(thisTk); */
	free_node(close_up_ptr(pNodePtr));		/* remove track ptr */
    }
    /* *pTails should be empty by now */
}

LIST *AddNewPeaks(TKLIST *pTails, LIST *pTrax, PKLIST strayPeakNodes)
{	/* return end of NewPeaks */
    PKLIST	newPeakNode;
    TKLIST	newTKnode;
    LIST	newTracks = NULL;

    /* WHAT FOLLOWS IS OBSOLETE AND SOON MAY NOT BE TRUE: (1993apr17)

       fact is, trk does not use TRACK structs so much; it maintains 
       them for currently growing tails, but once they're finished, 
       it just remembers their PKLISTS in a big nested list.
       pTrax is a pointer into that list, so it's a pure LIST, not 
       a TKLIST (although it is, in effect, a list of (headerless) tracks) */

/*    fprintf(stderr, "ANP: tails size %d, trax size %d, stray %d\n",
	    top_size(*pTails),	top_size(*pTrax), top_size(strayPeakNodes)); */
    while(newPeakNode = strayPeakNodes)  {
	strayPeakNodes = (PKLIST)cdr((LIST)strayPeakNodes);  
	                                             /* read _next, then.. */
	set_next((LIST)newPeakNode, NULL);	     /* unlink 1st element */
	newTKnode = (TKLIST)new_node(NULL, 
				     (LIST)tkNew(newPeakNode), 
				     T_TRACK);
	insert_by_comp((LIST *)pTails, (LIST)newTKnode, CmpFrqOfTrk);
	newTracks = cons((LIST)newPeakNode, T_LIST, (LIST)newTracks);
    }
/*    fprintf(stderr, "tails size %d, trax size %d, stray %d\n",
	    top_size(*pTails),	top_size(*pTrax), top_size(strayPeakNodes)); */
    return append_ptr(pTrax, newTracks); 
    /* somewhere near end of pTx */
}

