/***************************************************************\
*	trkstats.c		  				*
*   Functions to calculate the descriptive statistics for each	*
*   list of peaks (aka track).  TrackStats is part of the TRACK	*
*   header.  It has two states, BOILUP which means the values	*
*   are 'running totals', and BOILDOWN which means the values	*
*   have been reduced to their nominal meanings (std devn etc)	*
*   Functions here can calc stats for an entire peaklist at 1ce *
*   or add one peak at a time to BOILUP'd variables		*
*   1993apr18  dpwe 						*
\***************************************************************/

#include <dpwelib.h>
#include <assert.h>
#include <math.h>	/* some math fns here, like sqrt() */

#include "tracks.h"	/* brings in llist.h, peaks.h and TrackStats def */

static long trkIndex = 0;	/* indexes tracks on creation */

TRACK *tkNew(PKLIST pklist)
{	/* instance a new TRACK comprising the given <pklist> */
    TRACK	*trk;

    trk = TMMALLOC(TRACK, 1, "tkNew");
    trk->tktop = pklist;
    trk->tkbot = tkStatsRecalc(pklist, &(trk->st));
    if(pklist)  {
	trk->startPhase = pklist->peak->phs;
    } else {
	trk->startPhase = 0.0;
	fprintf(stderr, "tkNew(NULL)\n");	/* does this happen? */
    }
    trk->id = trkIndex++;
    return trk;
}

void tkFree(trk)
    TRACK	*trk;
{
    /* what about those PeakNodes? */
    free(trk);
}

void tkFreeIncPeaks(TRACK *trk)
{	/* free the track and apply pklistFree to its peaks too */
    pklistFree(trk->tktop);
    tkFree(trk);
}

void tkStReset(TrackStats *stats, PeakPos *pk)
{	/* initialize all trackstat fields from peak or to some 
	   `zero' equiv if <pk> is null */
    double	tim, frq, linmag;

    if(pk)  {
	stats->tmin = pk->tim;	stats->tmax = pk->tim;
	stats->fmin = pk->frq;	stats->fmax = pk->frq;
	stats->mmin = LINOFDB(pk->mag);
	stats->mmax = LINOFDB(pk->mag);
    } else {
	stats->tmin = 999999.9;	stats->tmax = -999999.9;
	stats->fmin = 999999.9;	stats->fmax = -999999.9;
	stats->mmin = 999999.9;	stats->mmax = -999999.9;
    }
    stats->tavg = 0.0;		stats->tsd  = 0.0;
    stats->favg = 0.0;		stats->fsd  = 0.0;
    stats->mavg = 0.0;		stats->msd  = 0.0;
    stats->dfdt = 0.0;
    stats->npts = 0;
    stats->state = TKST_BOILUP;	/* was TKST_NULL */
}

#define ALPHA .9	/* decay constant for df/dt - smaller-> faster */

void tkStUpdate(TrackStats *stats, PeakPos *pk, PeakPos *lastPk)
{	/* add one peak to assumed <boiled-up>ed track stats */
    double	tim, linmag, frq, dfdt;

    assert(stats->state == TKST_BOILUP);

    ++stats->npts;
    tim = pk->tim;
    frq = pk->frq;
    if(stats->tmin < 0)	stats->tmin = tim;
    stats->tmax = tim;
 
    /* calculate 1-pole smoothed df/dt, but no smoothing 1st point */
    if(lastPk && stats->npts > 1)  {
	dfdt = (frq - lastPk->frq) / (tim - lastPk->tim);
	if(stats->npts == 2)  {		/* second point gives 1st df/dt */
	    stats->dfdt = dfdt;
	} else {	/* already have some df/dt, so smooth it */
	    stats->dfdt *= ALPHA;
	    stats->dfdt += (1.0 - ALPHA) * dfdt;
	}
    }

    linmag = LINOFDB(pk->mag);
    if(linmag > stats->mmax)	stats->mmax = linmag;
    if(linmag < stats->mmin)	stats->mmin = linmag;
    if(frq > stats->fmax)	stats->fmax = frq;
    if(frq < stats->fmin)	stats->fmin = frq;

    stats->mavg += linmag;
    stats->favg += linmag * frq;
    stats->tavg += linmag * tim;
    stats->msd  += linmag * linmag;		/* bear with me */
    stats->fsd  += linmag * frq * frq;
    stats->tsd  += linmag * tim * tim;
}

void tkStStatUpdate(TrackStats *dst, TrackStats *add)
{	/* perform an update of <dst>, track stats summarizing 
	   a set of other track stats presented singly as <add> */
    /* 1993oct12: not at all clear to me that this routine does anything 
       useful.  Should at least establish boilup/boildown status of 
       <add>.  Modified <fsd> update to use favg not tavg, but 
       not at all clear why. */
    double linmag;

    ++(dst->npts);
    if(add->tmin < dst->tmin)		dst->tmin = add->tmin;
    if(add->tmax > dst->tmax)		dst->tmax = add->tmax;
    if(add->fmin < dst->fmin)		dst->fmin = add->fmin;
    if(add->fmax > dst->fmax)		dst->fmax = add->fmax;
    if(add->mmin < dst->mmin)		dst->mmin = add->mmin;
    if(add->mmax > dst->mmax)		dst->mmax = add->mmax;
    linmag = LINOFDB(add->mavg);
    dst->tavg += linmag * add->tavg;
    dst->tsd  += linmag * (/* add->tsd * add->tsd + */ add->tavg * add->tavg);
    dst->favg += linmag * add->favg;
    dst->fsd  += linmag * (/* add->fsd * add->fsd + */ add->favg * add->favg);
    dst->mavg += linmag;
    dst->msd  += linmag * linmag;	/*!!sd of mavgs, not avg of msds */
}

PKLIST tkStatsRecalc(PKLIST trk, TrackStats *stats)
{	/* calculate the stats for the passed pklist.
	   Return the last element in the peaklist (a favourite) */
    PeakPos	*pk, *lastPk = NULL;
    PKLIST	botPkN = NULL;

    pk = trk->peak;
    tkStReset(stats, pk);
    stats->state = TKST_BOILUP;
    while(trk)  {
	botPkN = trk;		/* will hold last eventually */
	pk = trk->peak;
	trk = trk->next;
	tkStUpdate(stats, pk, lastPk);
	lastPk = pk;
    }
    tkStBoildown(stats);
    return botPkN;
}

void tkStBoilup(TrackStats *stats)
{	/* Convert trackStats 'to date' into working numbers for 
	   running total */
    if(stats->npts > 0)  {
	assert(stats->state == TKST_BOILDOWN);
	/* inverse of BoilDown().. */
	/* convert magnitude fields back into linear */
	stats->msd += stats->mavg;
	stats->mavg = LINOFDB(stats->mavg);
	stats->msd  = LINOFDB(stats->msd) - stats->mavg;

	stats->mmax = LINOFDB(stats->mmax);
	stats->mmin = LINOFDB(stats->mmin);

	/* *sd fields are actually sumsq */
	stats->msd = stats->npts * (stats->msd * stats->msd
				    + stats->mavg * stats->mavg);
	/* .. and avgs are sum(x)*mag (or sum(mag) for mavg) */
	stats->mavg *= stats->npts;
	stats->tsd = stats->mavg * (stats->tsd * stats->tsd 
				    + stats->tavg * stats->tavg);
	stats->fsd = stats->mavg * (stats->fsd * stats->fsd 
				    + stats->favg * stats->favg);
	stats->tavg *= stats->mavg;
	stats->favg *= stats->mavg;
    }
    stats->state = TKST_BOILUP;
}

void tkStBoildown(TrackStats *stats)
{	/* convert the 'running total' stats into meaningful values */
    if(stats->npts > 0)  {
	assert(stats->state == TKST_BOILUP);
	stats->favg /= stats->mavg;
	stats->tavg /= stats->mavg;
	if(stats->npts == 1)  {
	    stats->fsd = 0;
	    stats->tsd = 0;
	} else {
	    double a,b,c;
	    a = stats->fsd / stats->mavg;
	    b = stats->favg * stats->favg;
	    if(a < b)  {
	/*	fprintf(stderr, "tkStBd: E[f^2] = %.2f, E[f]^2 = %.2f\n", 
			a, b);
	/*	abort();	*/
		c = a; a = b; b = c;	/* so ordering is right! */
	    }
	    stats->fsd  = sqrt( a - b );
	/*    stats->fsd  = sqrt( (stats->fsd / stats->mavg) 
			       - stats->favg * stats->favg);  */
	    stats->tsd  = sqrt( (stats->tsd / stats->mavg) 
			       - stats->tavg * stats->tavg);
	}
	stats->mavg /= (double)stats->npts;
	if(stats->npts == 1)
	  stats->msd  = 0;
	else  {
	    double a,b,c;
	    a = stats->msd / stats->npts;
	    b = stats->mavg * stats->mavg;
	    if(a < b)  {
	/*	fprintf(stderr, "tkStBd: E[m^2] = %.2f, E[m]^2 = %.2f\n", 
			a, b);
	/*	abort();	*/
		c = a; a = b; b = c;	/* so ordering is right! */
	    }
	    stats->msd  = sqrt( a - b );
	/*    stats->msd  = sqrt( (stats->msd / stats->npts)
			    - stats->mavg * stats->mavg); */
	}
	stats->mmin = DBOFLIN(stats->mmin);
	stats->mmax = DBOFLIN(stats->mmax);
	/* make magnitude `sd' in dB be the difference in dB 
	   between the average (linear) magnitude and that 
	   average plus one (linear) sd */
	stats->msd  = DBOFLIN(stats->msd + stats->mavg);
	stats->mavg = DBOFLIN(stats->mavg);
	stats->msd  -= stats->mavg;
    }
    stats->state = TKST_BOILDOWN;
}

TrackStats *tkStats(PKLIST trk)
{	/* calculate the frequency and magnitude ranges for this track */
    TrackStats	*stats;

    if(!trk)	return NULL;
    stats = TMMALLOC(TrackStats, 1, "TkStats");
    /* botPk = */
    tkStatsRecalc(trk, stats);
    return stats;
}

