/***************************************************************\
*	trksynth.c		  				*
*   Convert tracks into sound waveforms				*
*   Based on routines developed within dspB, for thesis		*
*   dpwe 08jul91						*
\***************************************************************/

/* #define DEBUGGRAPHS 1	/* only if .. */
/* #define DEBUG 1	/* */
#define CUBPHAS  1	/* comment out this flag for linear frq interp */

/* std libs */
#include <dpwelib.h>
/* personal libs */
#include <sndf.h>
#ifdef DEBUGGRAPHS
#include <window.h>
#endif
#include "peaks.h"
/* #include "tracks.h"	/* for LOGN2DB (but it's in peaks.h now) */

/* constants */

/* static function prototype */

/* static data */

/* functions */

#define MMmaskPhs(p,q,r,s) /* p is pha, q is as int, r is PI, s is 1/PI */ \
    q = (int)(s*p);	\
    p -= r*(float)((int)((q+((q>=0)?(q&1):-(q&1) ))));

void CalcCubPhasCoefs(p0, p1, pd0, pd1, dt, pa0, pa1, pa2, pa3)
    FLOATARG	p0, p1;		/* start and end phases */
    FLOATARG	pd0, pd1;	/* stt, end phase-dots (freqs) (rad/s) */
    FLOATARG	dt;		/* time in between 0 and 1 */
    float	*pa0, *pa1, *pa2, *pa3;	/* returns cubic poly params */
    /* calculate coefficients for a cubic phase function that meets 
       phase (principle) values and derivatives over a time range */
    /* same as Marc Zissman's phcoefgen, altho' I worked it from 1st princs */
    {
    float	x, y, dp;
    double	rint();

    /* figure how many 2PIs to add to p1's principle value */
    /* choose closest to average frequency */
    x = ((pd0+pd1)/2.0)*dt;	/* expected dphase */
    y = (p0+x) - p1;		/* difference from expected phase */
    y = 2.0 * PI * rint(y/(2.0*PI));	/* choose best multiple of 2 pi */
    p1 += y;			/* fixup final phase */
    dp = p1 - p0;		/* overall phase change */
    *pa0 = p0;			/* must start here */
    *pa1 = pd0;			/* initial derivative this */
    *pa2 = 1/(dt*dt)*(3*dp - dt*(pd1+2*pd0));
    *pa3 = 1/(dt*dt*dt)*(-2*dp+dt*(pd1+pd0));	/* well that's the math */
/*    fprintf(stderr,
"p0 %.4f, p1 %.4f, pd0 %.2f, pd1 %.2f, dt %.3f, a's %.3f %.3f %.3f %.3f\n",
	    p0, p1, pd0, pd1, dt, *pa0, *pa1, *pa2, *pa3);	*/
    }

#define MINUSALOT (-101.7)	/* as a flag for log(0) (FNX) */
#define DBOFST    (-29.7)	/* knock mags down by 30dB for approx scale */

float TrackOscillator(track, time, buf, num, phi0, sRate)
    LIST	*track;		/* may be stepped forward */
    float	*time;		/* will be incremented accordingly */
    float	*buf;
    long	num;		/* how many points to do & where */
    FLOATARG	phi0;		/* initial osc phase, from last call */
    FLOATARG	sRate;		/* overall sampling rate of buf */
    {	/* calculate some new points.  Return phase at end */
    float	f0,f1,t0,t1,m0,m1,mm0,mm1,phi,startTime,dt,oneOnSr;
    float	ph0, ph1, a0, a1, a2, a3, t;
    PeakPos	*thisPeak, *nextPeak;
    float	aMarg,tmCycs,frac,frq,mag;
    long 	i;
    float 	oneOnTwoPi, oneOnDt, toRadS, frqToDph;
    float	pi, oneOnPi;
    int 	q;
    long 	dnum;

/*    fprintf(stderr, "t(%.4f)n(%ld)-",*time,num);	fflush(stderr);	*/

    aMarg = 50.0;	/* fade in/out is this many dB down */ 
    tmCycs = 2*2.0;	/* ? fade in/out time is this many cycles */
    oneOnSr = 1.0/sRate;
    frqToDph = 2.0*PI/sRate;
    pi = PI;
    oneOnPi = 1.0/pi;
    oneOnTwoPi = 1.0/(2.0 * PI);
    toRadS = 2*PI;

 top:
/*    fprintf(stderr, "<%d@%.6f>", num, *time); fflush(stderr);	*/
    if(*track)		thisPeak = (PeakPos *)car(*track);
    else		thisPeak = NULL;

    if(thisPeak)
	{	m0 = thisPeak->mag;	f0 = thisPeak->frq;
		t0 = thisPeak->tim;	ph0 = thisPeak->phs;
		t1 = t0;	}
    else { 	/* asked to synth with no track info - lose */
	*time += num * oneOnSr;
	return phi0;	/* osc. not alive : output nothing */
    }

    /* t0 must be >= *time, so if t1 < *time, or even if t1 = *time, 
       shift t1 back to t0 */
    while(thisPeak && t1 <= *time)	/* current peak precedes spec time */
	{
	if(cdr(*track))	    nextPeak = (PeakPos *)car(cdr(*track));
	else		    nextPeak = NULL;
	if(nextPeak == NULL)
	    {	/* assume track has died, need to fade out gracefully */
	    f1 = f0;	    ph1 = ph0 + tmCycs*2*PI;
	    t1 = t0 + (tmCycs/f0);	m1 = MINUSALOT; /* m0-aMarg; */
	    }
	else{
	    f1 = nextPeak->frq;	    t1 = nextPeak->tim;	    
	    m1 = nextPeak->mag;	    ph1 = nextPeak->phs;
	    }
	if(t1 <= *time)
	    {			/* need to advance <track> */
	    thisPeak = nextPeak;
	    t0 = t1;	f0 = f1;	m0 = m1;	ph0 = ph1;
	    *track = cdr(*track);
	    }
	}
    	
    if(thisPeak && *time < t0)
	{	/* assume pre-start, so consider fading in */
	t1 = t0;	f1 = f0;	m1 = m0;	ph1 = ph0;
	t0 = t1 - (tmCycs/f0);
	m0 = MINUSALOT; /* m1 - aMarg; */	ph0 = ph1 - 2*PI*tmCycs;
	}

    dnum = ((t0 - *time)*sRate);		/* samples FROM *time TO t0*/
    if(t1 <= *time || num < dnum)  {
	/* either couldn't reach *time in current chain, 
	   or *time so far before t0 that no samples ever generated */
	*time += num * oneOnSr;
	return phi0;	/* osc. not alive : output nothing */
    }
    if(dnum > 0)  {
	/* here, *time <= t1, but 0 < dnum < num */
	/* wind forward to t0 .. */
	*time += dnum * oneOnSr;
	buf += dnum;	num -= dnum;
	goto top;
    }

/*     if(thisPeak == NULL || !(t0 <= *time && *time < t1))
/* 	{
/* 	long dnum = 1+((t0-*time)*sRate);	/* samples twixt time & t0 */
/* 	fprintf(stderr, "\n<t0:%.6f t:%0.6f t1:%0.6f dn:%ld", 
/* 		t0, *time, t1, dnum);  fflush(stderr);
/* 	if( dnum < 0 || dnum >= num )
/* 	    {	/* dnum < 0 implies t1 <= *time, dnum >= num => never @t0 */
/* 	    *time += num*oneOnSr;
/* 	    return phi0;	/* oscillator not alive : output nothing */
/* 	    }
/* 	else{			/* presume *t<t0 but (n*o)+*t>=t0 */
/* 	    if(dnum >= num)	/* (won't happen at present) */
/* 		{  fprintf(stderr, "OSC: wierd %ld, %ld\n", dnum, num);
/* 		   abort();	}
/* 	    *time += dnum*oneOnSr;
/* /* 	    return TrackOscillator(track,time,buf+dnum,num-dnum,phi0,sRate);
/* 	    buf = buf+dnum;	num -= dnum;
/* 	    goto top;
/* 	    }
/* 	}
 */

#ifdef DEBUG
    fprintf(stderr,"+");	fflush(stderr);
#endif

    CalcCubPhasCoefs(ph0, ph1, 
		     toRadS*f0, toRadS*f1, 
		     t1-t0, &a0, &a1, &a2, &a3);
/*    fprintf(stderr,"f %.2f-%.2f Hz ", 
		     oneOnTwoPi*toRadS*f0, 
	             oneOnTwoPi*toRadS*f1);
    fflush(stderr);	*/
/*    fprintf(stderr,
"ph0 %.4f, ph1 %.4f, pd0 %.2f, pd1 %.2f, dt %.3f, a's %.3f %.3f %.3f %.3f\n",
	    ph0, ph1, toRadS*f0, toRadS*f1,
	    t1-t0, a0, a1, a2, a3);	*/

    if(m0==MINUSALOT)	mm0 = 0; 	else	mm0 = exp((DBOFST+m0)/LOGN2DB);
    if(m1==MINUSALOT)	mm1 = 0; 	else	mm1 = exp((DBOFST+m1)/LOGN2DB);
    oneOnDt = 1.0/(t1 - t0);
    startTime = *time;
    phi = phi0;
    i = 0;
    while(i<num && *time < t1)
	{
	frac = (*time - t0)*oneOnDt;
/*	mag = -18.0+m0 + (m1-m0)*frac;		/* lin intp of log mag (dB) */
	mag = mm0 + frac*(mm1-mm0);		/* NO - lin intp of LIN mag */
	frq = f0 + (f1-f0)*frac;		/* lin intp of frq - well.. */
	t = *time - t0;
#ifdef CUBPHAS
	phi = a0 + a1*t + a2*t*t + a3*t*t*t;	/* cubic interp */
	MMmaskPhs(phi,q,pi,oneOnPi) /* phi is phase, q is an int, consts */
/*	fprintf(stderr,"ph%.4f",phi);	fflush(stderr);	*/
/*	*buf++ += exp(mag/LOGN2DB)*cos(phi);	*/
	*buf++ += mag*cos(phi);
#else /* linear frequency interpolation */
/* 	*buf++ += exp(mag/LOGN2DB)*cos(phi);	*/
	*buf++ += mag*cos(phi);
	phi += frq*frqToDph;
	if(phi > PI)	phi -= 2*PI; /* *(int)((phi-PI)*oneOnTwoPi); */
#endif /* CUBPHAS */
	++i;
	*time = startTime + (float)i*oneOnSr;
	}
/*    fprintf(stderr,"%.3f\n",phi);	*/
    if(i<num)  {	/* overran this segment - deal with it recursively */
/*	phi = TrackOscillator(track, time, buf, num-i, phi, 
			      sRate);	*/
	/* don't recurse, loop */
	num -= i;	phi0 = phi;
/*	fprintf(stderr, "+");	fflush(stderr);  */
	goto top;
    }
    return phi;		/* also modified *track, *time */
    }

#define SATBSIZE 1024
SynthATrack(LIST trax, double sRate, double duration)
{
    SFSTRUCT	*newSnd;
    int 	err = 0;
    long	samps, dBsize;
    int 	dsize  = sizeof(float);
    int 	format = SFMT_FLOAT;
    int 	chans  = 1;
/*    float	buf[SATBSIZE];	*/
    long	buflen = SATBSIZE, n, todo;
    float	ph = 0, time;
    float 	*fptr, *fp2;
    PKLIST	track;
    float	thisOnesStartTime, latestTime = 0.0;
    char 	filename[64];

    samps  = (long)(duration*sRate);
    dBsize = dsize*chans*samps;
    if( err=SFAlloc(&newSnd, dBsize, format, sRate, chans, SFDFLTBYTS) )
	SFDie(err, "SynthATrack" );
    fptr = (float *)SFDataLoc(newSnd);
    ClearFloats(fptr, 1, samps);
    while(trax) 
	{	/* for each track .. */
/*	do  {	/* scan to a track worth synthing? (poor metric) */
/*	    track = car(trax);
/*	    trax = cdr(trax);
/*	    } while(trax && ((PeakPos *)car(track))->mag < -10.0);	*/
	track = (PKLIST)car(trax);
	trax  = cdr(trax);	

	n = 0;	fp2 =fptr;	/* start at top of sound (again) */
	thisOnesStartTime = track->peak->tim;
	if(thisOnesStartTime > latestTime)	/* progress report */
	    {	    fprintf(stderr,"Synth: t %.4f\n",latestTime);
		    latestTime += .1;	    }
	while(n<samps)	/* this buflen chunking is unneeded now.. */
	    {		/* synth each track over ENTIRE LENGTH of sound ?? */
	    if(samps-n > buflen)
		todo = buflen;
	    else
		todo = samps - n;
	    time = n / sRate;
/*	    fprintf(stderr, "SAT time %.4f\n",time);	fflush(stderr);	*/
	    ph = TrackOscillator(&track, &time, fp2, todo, ph, 
				 sRate);
	    n += todo;
	    fp2 += todo;
	    }
	}
    if( err = ConvertSoundFormat(&newSnd, SFMT_SHORT))
	SFDie(err, NULL);

    sprintf(filename, "tmp%s", sndfExtn);
    SFWrite(filename, newSnd);
    }

